"""This module defines two classes which allow you to
use pickle to save SWIGged C instances of DigitalMaterial
Serializable classes.  It defines two classes:

class PickleWriter( filename, DMWriter=ASCIIWriter ):
    def dump( object )

class PickleReader( filename, DMReader=ASCIIReader ):
    def load(): returns object

Given a filename "z", it will write Python classes to
"z.pickle" and C instances inherited from Serializable
to "z.cdf".  Only those shadow classes which have a
__getstate__ and __setstate__ like that defined in
Rand.i will work.
"""

try:
    import MD3Dc as MDc
    import MD3D as MD
except ImportError:
    import MD2Dc as MDc
    import MD2D as MD

from PythonSerialMemo import new_PythonSerialMemo, PythonSerialMemoType
import pickle, copy_reg

def PythonSerialMemoSaver(object):
    """Called by Pickle to save a PythonSerialMemo.
    The PythonSerialMemo has a pointer to the C instance
    we wish to save.  This function uses a global DMWriter
    to save that C instance to a file separate from the
    pickle file, and it returns to pickle the name under
    which the C instance was saved, so pickle can find it
    again during loading."""
    owner = object.GetPythonInstance()
    creator = 'MDc.new_' + owner.GetType()
    name = id(owner)
    global _writer
    _writer.Put( owner, repr(name) )
    return PythonSerialMemoLoader, (creator,name)

def PythonSerialMemoLoader(creator,name):
    """Called by Pickle to create a PythonSerialMemo.
    The arguments are the constructor to call and the
    name of the object in the DMReader's serialization file.
    This function creates the C instance, loads it with
    information from the DMReader's file, and stores a
    pointer to it in the PythonSerialMemo, which is then
    returned to pickle."""
    createString = creator+"()"
    this = eval(createString)
    global _reader
    _reader.Fill( this, repr(name) )
    a = new_PythonSerialMemo()
    a.SetPythonInstance( this )
    return a

# This tells pickle that, when it wants to serialize or create
# a PythonSerialMemoType, it should call the Saver and Loader
# functions given above.  Note that the PythonSerialMemoType
# is an extension class only because copy_reg insists that
# it will handle only extension classes.
copy_reg.pickle( PythonSerialMemoType, PythonSerialMemoSaver,
                 PythonSerialMemoLoader )

class PickleWriter:
    def __init__(self,filename,writer = MD.ASCIIWriter):
        self.filename = filename
        self.writer = writer

    def dump( self, obj ):
        global _writer
        filename = self.filename
        _writer = self.writer( filename + ".cdf" )
        pickle.dump( obj, open(filename + ".pickle", 'w' ) )
        _writer.Close()
        del _writer

class PickleReader:
    def __init__(self,filename,reader = MD.ASCIIReader):
        self.filename = filename
        self.reader = reader

    def load( self ):
        global _reader
        filename = self.filename
        _reader = self.reader( filename+".cdf" )
        obj = pickle.load( open(filename + ".pickle", 'r') )
        del _reader
        return obj

# Only export these two symbols.
__all__ = [ "PickleWriter", "PickleReader" ]

############### Testing ################
def save():
    a = MD.Rand()
    a.SetSeed(47)
    a.uniform()
    writer = PickleWriter("z")
    writer.dump( a )
    print 'saved ',a, a.__class__, a.uniform()

def load():
    reader = PickleReader('z')
    obj = reader.load()
    print 'loaded',obj, obj.__class__, obj.uniform()
    return obj

if __name__ == "__main__":
    save()
    load()
