# DomainSubdivision.py

"""Decides which atoms belong where, and who needs to know about them.

Two related tasks:
  1: Decide on which processor an atom should reside.
  2: Decide which processors need to know about an atom that resides on
     this processor.
"""


import math
import Numeric
import LinearAlgebra

# Helper functions
def _crossproduct(a, b):
    return Numeric.array((a[1]*b[2] - a[2]*b[1],
                          a[2]*b[0] - a[0]*b[2],
                          a[0]*b[1] - a[1]*b[0]))

def _norm(a):
    return math.sqrt(Numeric.dot(a,a))

# The DomainSubdivision class is the main part of this module.
class DomainSubdivision:
    """Decides which atoms belong where, and who needs to know about them.

    Two related tasks:
      1: Decide on which processor an atom should reside.
      2: Decide which processors need to know about an atom that resides on
         this processor.

    It only works in 3D (so far).

    Bugs: Boundary conditions are ignored so far.
    """

    def __init__(self, basis, cpulayout, bufferlayer):
        """Creates the domain decomposition object."""
        self.basis = basis
        self.invbasis = LinearAlgebra.inverse(basis)
        self.ncpu = cpulayout
        self.ncpu0 = self.ncpu[0]
        self.ncpu01 = self.ncpu[0] * self.ncpu[1]
        self.ndomains = self.ncpu[0] * self.ncpu[1] * self.ncpu[2]
        self.dim = 3
        self.bufferlayer = bufferlayer

    def cpucoord_to_domain(self, cpu):
	"""Converts three coordinates in cpu space to domain number."""
	return cpu[0] + self.ncpu0 * cpu[1] + self.ncpu01 * cpu[2]

    def domain_to_cpucoord(self, domain):
	"""Converts a domain number to three coordinates in cpu space."""
	return (domain % self.ncpu0, (domain/self.ncpu0) % self.ncpu[1],
		domain / self.ncpu01)
    
    def PointInDomain(self, point, domain):
        """Decides if a point belongs to a given processor."""
        rs = Numeric.matrixmultiply(point, self.invbasis)
        cpu = Numeric.zeros((self.dim,))
        for d in range(self.dim):
            ri = rs[d]
            if ri >= 1.0:
                ri = 0.9999999
            if ri < 0.0:
                ri = 0.0
            cpu[d] = int(self.ncpu[d] * ri)
            assert cpu[d] >= 0
            assert cpu[d] < ncpu[d]
        return domain == self.cpucoord_to_domain(cpu)
    
    def PointsInDomain(self, points, domain):
        """Decides if multiple points belong to a given processor."""
        return Numeric.equal(domain, self.WhichDomain(points))

    def WhichDomain(self, points):
        """Decides in which domain multiple points belong."""
        N = Numeric
        if len(points) == 0:
            return N.zeros((0,))
        rs = N.matrixmultiply(points, self.invbasis)
        cpu = N.zeros((self.dim, len(rs)))        
        rs = N.transpose(rs)
        for d in range(self.dim):
            ri = rs[d]
            ri = N.where(N.greater_equal(ri, 1.0), 0.9999999, ri)
            ri = N.where(N.less(ri, 0.0), 0.0, ri)
            cpu[d] = (self.ncpu[d] * ri).astype(N.Int)
            assert N.minimum.reduce(cpu[d]) >= 0
            assert N.maximum.reduce(cpu[d]) < self.ncpu[d]
        return self.cpucoord_to_domain(cpu)
        
    def _BuildSendListsLeaf(self, points, thisdomain):
        """Build lists of which points should be send to other processors.

        The list of points is a Numeric array.
        """
        N = Numeric
        target = self.WhichDomain(points)
        sendsomewhere = N.not_equal(target, thisdomain)
        sendlists = []
        for i in xrange(self.ndomains):
            sendlists.append([])
        
        # LOOP OVER ATOMS!!!
        for i in xrange(len(points)):
            if sendsomewhere[i]:
                sendlists[target[i]].append(i)
        # Turn sendlists into Numeric arrays
        #for i in xrange(self.ndomains):
        #    sendlists[i] = N.array(sendlists[i])
        return sendlists

    def BuildSendLists(self, points, thisdomain):
        """Build lists of which points should be send to other processors.

        The list of points is a DMArray.
        """
        leaflist = []
        nleaves = points.NumberOfLeaves()
        for leaf in xrange(nleaves):
            l = points.GetLeaf(leaf)[:]
            leaflist.append(self._BuildSendListsLeaf(l, thisdomain))
        #   Now reshape the list
        sendlists = []
        for i in xrange(self.ndomains):
            sl = []
            sendlists.append(sl)
            for leaf in xrange(nleaves):
                sl.append(leaflist[leaf][i])
        return sendlists

    def PointsInterestingForDomain(self, points, domain):
        """Decides if multiple points are interesting for a given processor."""
        N = Numeric
	interest = N.ones((len(points),))
	rs = N.transpose(N.matrixmultiply(points, self.invbasis))
	cpu = self.domain_to_cpucoord(domain)
	for (axis, other1, other2) in ((0,1,2), (1,2,0), (2,0,1)):
	    cross = _crossproduct(self.basis[other1], self.basis[other2])
	    cosphi = N.dot(self.basis[axis], cross) / (_norm(self.basis[axis])
						       * _norm(cross))
	    bufferzone = math.fabs(self.bufferlayer /
                                   (_norm(self.basis[axis])*cosphi))
	    limitlow = float(cpu[axis]) / self.ncpu[axis] - bufferzone
	    limithigh = float(cpu[axis]+1) / self.ncpu[axis] + bufferzone
	    # Are the points within the limits?
	    axisinterest = N.greater(rs[axis], limitlow) * N.less(rs[axis],
                                                                  limithigh)
	    interest = N.where(axisinterest, interest, 0)
	return interest	  
    
    def _BuildInterestListsLeaf(self, points, thisdomain):
	"""Build lists of which points are interesting for other processors.

        The list of points is a Numeric array.
	"""
	N = Numeric
	ilist = []
	nproc = self.ndomains
	procID = thisdomain
	for domain in xrange(nproc):
	    if domain == procID:
		ilist.append(None)
	    else:
		interest = self.PointsInterestingForDomain(points, domain)
		# LOOP OVER ATOMS!
		nint = N.add.reduce(interest)
		l = N.zeros((nint,))
		nn = 0;
		for i in xrange(len(points)):
		    if interest[i]:
			l[nn] = i
			nn = nn + 1
		ilist.append(l)
	return ilist

    def BuildInterestLists(self, points, thisdomain):
        """Build lists of which points are interesting for other processors.

        The list of points is a DMArray.
        """
        leaflist = []
        nleaves = points.NumberOfLeaves()
        for leaf in xrange(nleaves):
            l = points.GetLeaf(leaf)[:]
            leaflist.append(self._BuildInterestListsLeaf(l, thisdomain))

        # Now reshape the list
        ilists = []
        for i in xrange(self.ndomains):
            il = []
            ilists.append(il)
            for leaf in xrange(nleaves):
                il.append(leaflist[leaf][i])
        return ilists

    
	

###########################################################
#
# TEST SUITE
#
###########################################################

if __name__ == "__main__":
    import sys
#    sys.path.append("/home/data/schiotz/DigitalMaterial/src/DMSystem/alpha/obj3D")
    sys.path.append("T:/tcretegn/CVS/DigitalMaterial/src/DMSystem/")
    from DMSystem import *
    
    N = Numeric
    from RandomArray import random
    
    def testNumPy(b, n, cpu):
        ds = DomainSubdivision(b, cpu)
        ncpu = cpu[0]*cpu[1]*cpu[2]
        pos = N.matrixmultiply(random((n,3)), b)
        sl = ds._BuildSendListsLeaf(pos, 0)
        print "Positions:"
        print pos
        for i in range(ncpu):
            print "Send to node",i
            for j in sl[i]:
                print pos[j]

    def makeDM(b, n1, n2):
        a = ArrayDoubleVector()
        a1 = ArrayDoubleVector(n1)
        pos = N.matrixmultiply(random((n1,3)), b)
        a1[:] = pos
        a2 = ArrayDoubleVector(n2)
        pos = N.matrixmultiply(random((n2,3)), b)
        a2[:] = pos
        a.Attach(a1)
        a1.thisown = 0
        a.Attach(a2)
        a2.thisown = 0
        return a

    def NotOnList(lst, points):
        onlist = N.zeros((len(points),))
        for i in lst:
            onlist[i] = 1
        res = []
        for i in xrange(len(points)):
            if onlist[i] == 0:
                res.append(points[i])
        return N.array(res)

    basis = N.identity(3) * N.array((10.0,5.0,20.0))
    cpulayout = (2,1,2)
    ds = DomainSubdivision(basis, cpulayout, bufferlayer=1.0)
    positions = makeDM(basis, 100, 200)

    sendlist = ds.BuildSendLists(positions, 0)
    print 'len of send lists = ', len(sendlist)
    # Test that all atoms belong to one cpu
    # Count this cpu
    belonghere = ds.PointsInDomain(positions[:], 0)
    n = N.add.reduce(belonghere)
    print 'n = ', n
    for sl in sendlist:
        if sl is not None:
            for leaf in sl:
                n = n + len(leaf)
    if n == len(positions):
        print "Testing total length of sendlists: OK"
    else:
        print "Testing total length of sendlists: FAILED"

    # Test sendlist for cpu (0,0,1).
    # First test that all elements on sendlist belong there.
    theseatoms = []
    sl = sendlist[ds.cpucoord_to_domain((0,0,1))]
    for i in range(len(sl)):
        slleaf = sl[i]
        leaf = positions.GetLeaf(i)[:]
        for j in slleaf:
            theseatoms.append(leaf[j])
    nlist = len(theseatoms)
    print nlist, "on sendlist to (0,0,1)..."
    theseatoms = N.array(theseatoms)
    x = N.transpose(theseatoms)[0]
    y = N.transpose(theseatoms)[1]
    z = N.transpose(theseatoms)[2]
    ok = N.less_equal(x, 5.0)
    if N.add.reduce(ok) == nlist:
        print "... x coordinates: OK"
    else:
        print "... x coordinates: FAILED"
    ok = N.greater_equal(z, 10.0)
    if N.add.reduce(ok) == nlist:
        print "... z coordinates: OK"
    else:
        print "... z coordinates: FAILED"
    # Then test other sendlists
    theseatoms = []
    for cpui in (0,1):
        for cpuj in (0,):
            for cpuk in (0,1):
                cpu = (cpui, cpuj, cpuk)
                if cpu == (0,0,0) or cpu == (0,0,1):
                    continue
                sl = sendlist[ds.cpucoord_to_domain(cpu)]
                for i in range(len(sl)):
                    slleaf = sl[i]
                    leaf = positions.GetLeaf(i)[:]
                    for j in slleaf:
                        theseatoms.append(leaf[j])
    nlist = len(theseatoms)
    print nlist, "on other sendlists..."
    theseatoms = N.array(theseatoms)
    x = N.transpose(theseatoms)[0]
    y = N.transpose(theseatoms)[1]
    z = N.transpose(theseatoms)[2]
    notok = N.less(x, 5.0) * N.greater(z, 10.0)
    if N.add.reduce(notok) == 0:
        print "... x and z coordinates: OK"
    else:
        print "... x and z coordinates: FAILED"
        print "Found",  N.add.reduce(notok)
                

        
    # Test interestlist for cpu (0,0,1).
    # First test that all elements on interestlist belong there.
    interestlist = ds.BuildInterestLists(positions, 0)
    theseatoms = []
    sl = interestlist[ds.cpucoord_to_domain((0,0,1))]
    for i in range(len(sl)):
        slleaf = sl[i]
        leaf = positions.GetLeaf(i)[:]
        for j in slleaf:
            theseatoms.append(leaf[j])
    nlist = len(theseatoms)
    print nlist, "on interestlist to (0,0,1)..."
    theseatoms = N.array(theseatoms)
    #print theseatoms
    x = N.transpose(theseatoms)[0]
    y = N.transpose(theseatoms)[1]
    z = N.transpose(theseatoms)[2]
    ok = N.less_equal(x, 6.0)
    if N.add.reduce(ok) == nlist:
        print "... x coordinates: OK"
    else:
        print "... x coordinates: FAILED"
        print N.add.reduce(ok)
    ok = N.greater_equal(z, 9.0)
    if N.add.reduce(ok) == nlist:
        print "... z coordinates: OK"
    else:
        print "... z coordinates: FAILED"
        print N.add.reduce(ok)

    # Then test atoms not on list
    leaf0 = positions.GetBranch(0)[:]
    lst0 = sl[0]
    notonlist0 = NotOnList(lst0, leaf0)
    leaf1 = positions.GetBranch(1)[:]
    lst1 = sl[1]
    notonlist1 = NotOnList(lst1, leaf1)
    notonlist = N.concatenate((notonlist0, notonlist1))
    assert len(notonlist) + nlist == len(positions)

    nlist = len(notonlist)
    print nlist, "not on the interestlist..."
    theseatoms = N.array(notonlist)
    x = N.transpose(theseatoms)[0]
    y = N.transpose(theseatoms)[1]
    z = N.transpose(theseatoms)[2]
    notok = N.less(x, 6.0) * N.greater(z, 9.0)
    if N.add.reduce(notok) == 0:
        print "... x and z coordinates: OK"
    else:
        print "... x and z coordinates: FAILED"
        print "Found",  N.add.reduce(notok)
    

    print "Test of non-orthogonal basis NOT YET IMPLEMENTED."
