# NetProtocol.py

from NetBuffers import *
import Scientific.MPI
MPI = Scientific.MPI
del Scientific.MPI
import Numeric

class NetProtocol:
    """Implements network communications protocols.  Uses MPI."""

    def __init__(self, communicator=None):
	if communicator:
	    self._comm = communicator
	else:
	    world = MPI.world
	    if world is None:
		raise RuntimeError, "If you use MPI you should use an MPI aware Python executable."
	    # Create a new communicator to prevent interference from other
	    # packages using MPI
	    self._comm = world.duplicate()
	self._butterflytag = 42

    def GetProcessorNumber(self):
	"""Return the number of this processor."""
	return self._comm.rank

    def GetNumberOfProcessors(self):
	"""Get the total number of processors."""
	return self._comm.size

    def BroadcastValue(self, value, master):
        """Broadcast a scalar to all processors.  Returns the result."""
        buf = Numeric.array((value,))
        self._comm.broadcast(buf, master)
        return buf[0]

    def Broadcast(self, array, master):
        """Broadcast an array to all processors.  The array is modified in place."""
        self._comm.broadcast(array, master)

    def Barrier(self):
        self._comm.barrier()
        
    def Butterfly(self, sendbuffers):
	"""Implements the butterfly communications protocol."""
	recvbuffers = [None] * len(sendbuffers)
	tag = self._butterflytag
	
	ntasks = self.GetNumberOfProcessors()
	taskid = self.GetProcessorNumber()
	
	# We will send to the following nodes (relative positions):
	# +1, -1, +2, -2, .... ntasks/2, -ntasks/2
	# and simultaneously receive from
	# -1, +1, -2, +2, .... -ntasks/2, ntasks/2
	# omitting the last if ntasks is odd.

	for offset in xrange(1,1+ntasks/2):
	    for direction in [1,-1]:
		# Find id of tasks we want to send to and receive from
		sendcpu = taskid + direction*offset
		recvcpu = taskid - direction*offset
		if sendcpu >= ntasks:
		    sendcpu = sendcpu - ntasks
		elif sendcpu < 0:
		    sendcpu = sendcpu + ntasks
		if recvcpu >= ntasks:
		    recvcpu = recvcpu - ntasks
		elif recvcpu < 0:
		    recvcpu = recvcpu + ntasks

		# Skip the last communication if it is the same as the second
		# last (happens if ntasks is odd)
		if direction == 1:
		    lastcpu = sendcpu
		if direction == -1 and lastcpu == sendcpu:
		    continue

		if sendcpu != taskid:
		    sdata = sendbuffers[sendcpu].getvalue()
		    sendrq = self._comm.nonblocking_send(sdata, sendcpu, tag)
		if recvcpu != taskid:
		    rdata = self._comm.receiveString(recvcpu, tag)
		    rdata = rdata[0]  # Discard source and tag
		    recvbuffers[recvcpu] = NetReadBuffer(rdata)
		    assert recvbuffers[recvcpu] is not None
		if sendcpu != taskid:
		    sendrq.wait()

	# Sanity check
	assert len(recvbuffers) == ntasks
	for b in xrange(ntasks):
	    if b == taskid:
		assert recvbuffers[b] is None
	    else:
		assert recvbuffers[b] is not None
		
	# Done
	return recvbuffers


######################################################################

def test():
    """Test suite for NetProtocol."""

    prot = NetProtocol()

    taskid = prot.GetProcessorNumber()
    ntasks = prot.GetNumberOfProcessors()
    
    # Make som identifiable data to send
    sendbuffers = []
    for i in xrange(ntasks):
	if i == taskid:
	    sendbuffers.append(None)
	else:
	    str = "from %d to %d" % (taskid, i)
	    nwb = NetWriteBuffer()
	    nwb.write(str)
	    sendbuffers.append(nwb)
    recvbuffers = prot.Butterfly(sendbuffers)

    #Check the received data
    ok = 1
    for i in xrange(ntasks):
	if i != taskid:
	    str = "from %d to %d" % (i, taskid)
	    nrb = recvbuffers[i]
	    str2 = nrb.readline()
	    if str == str2:
		if taskid == 0:
		    print "Processor %d: correct result from %d" % (taskid, i)
	    else:
		ok = 0
		print "Processor %d: Wrong result from %d.  Got '%s' expected '%s'." % (taskid, i, str2, str)
    if ok:
	print "Processor %d: SUCCESS" % (taskid,)
    else:
	print "Processor %d: FAILED ***********" % (taskid,)

# Run test suite if running as a program
if __name__ == "__main__":
    test()

    
