# This example constructs a 2D small-angle grain boundary,
# using a Python AtomsInitializer class, BicrystalInitializer.py, also found in this
# directory. It then puts the grain boundary under external strain, until
# the system fractures along it.

# Path for DMSystem
DMPATH = '../..'

import sys

# Windows
sys.path.append(DMPATH + '/src/MD/SWIGBuild')
# Unix
sys.path.append(DMPATH + '/src/MD/SWIGBuild/i686/obj2D')
# Rasmol
sys.path.append(DMPATH + '/Python/MD/Observers')

from MD2D import *
from RasMolObserver import RasMolObserver2D

from Numeric import *

from BicrystalInitializer import *

# LinearPositionTransformer will take the atoms and perform a linear shear on
# them
def stretch(strain, atoms):
	""" expands all the atoms in the x direction by a factor of strain"""
	LinearPositionTransformer(array([[strain, 0.],[0.,0.]])).Transform(atoms)

potential = CutLennardJonesPotential()
lattice = TriangularLattice(potential.GetLengthScale())

# Speed at which the crack opens
velocity = 0.02

# Length and width of the original rectangle of atoms
width = 30.
length = 100.

# BicrystalInitializer will set the box size for the boundary conditions
# Length and width are the minimum size it will pick
# (it's non-trivial finding out how large a box you need to get the boundary periodic)
bc = SmartPeriodicBoundaryConditions([0, 1])


### the grain boundary will be symmetric if the Miller indices of the surface (k1,l1) and (k2,l2)
### are such that l1/k1 + l2/k2 = 1
cry = BicrystalInitializer(lattice, [width, length], [1, 10.], [1, -9.],boundaryConditions=bc, \
		       		constrLayerWidth = potential.GetCutoffDistance())
atoms = DynamicListOfAtoms(0)
atoms.SetBoundaryConditions(bc)
bulk = DynamicListOfAtoms(0)
# BiCrystalInitializer needs to create a list with three branches:
# the bulk including the grain boundary,
# and two enveloping constrained zones that will be stretched apart.
# Since it's a Python script, it has a harder time making the tree structure,
# so we build it here.
cry.Create(bulk)
constrZone1 = cry.constrLOA1 
constrZone2 = cry.constrLOA2 
atoms.Attach(bulk)
atoms.Attach(constrZone1)
atoms.Attach(constrZone2)

#ras = RasMolObserver2D(radius = 100)
#atoms.AddObserver(ras)
atoms.Notify()

# initial stretch, to speed up things
stretch(0.03, atoms)
atoms.Notify()

# We start by holding the constrained atoms fixed in position and relaxing,
# before we start opening the crack
fixedConstraint1 = FixedBody(constrZone1.GetCartesianPositions())
fixedConstraint2 = FixedBody(constrZone2.GetCartesianPositions())
constrZone1.SetConstraint(fixedConstraint1)
constrZone2.SetConstraint(fixedConstraint2)

# XXX???
fixedConstraint1.thisown = fixedConstraint2.thisown = 0

# NEIGHBOR LOCATORS
cutoff = potential.GetCutoffDistance()
neighborLocator = CellNeighborList(cutoff, 0.05*cutoff)
# You can use bondbreaker to decide where the crack will start
bondbreaker = BondBreakerNeighborLocator(neighborLocator, [0, 100.], [0, -100.], [0, 0], 1)
# we open a crack, without removing atoms
print 'use the bondbreaker neighbor locator'
atoms.SetNeighborLocator(bondbreaker)
#atoms.SetNeighborLocator(neighborLocator)

# We could minimize the energy now using one of the minimizers, but we won't
cg = ConjugateGradientMinimizer(1.0e-6, 1000)
sd = SteepestDescentMinimizer(1.0e-10, 1000, 0.01)
cg.Verbose()
sd.Verbose()
minimizer = MinimizerAtomsMover(cg, potential)



# Here the idea was to relax the crack with the broken bonds, then put the
# bonds back and relax again. We're not doing that now...
minimizer.Move(atoms)
minimizer.Move(atoms)
atoms.Notify()
# XXX ???
neighborLocator.thisown = 0
print 'restore the original neighbor locator (get rid of bondbreaker)'
atoms.SetNeighborLocator(neighborLocator)

print 'minimize energy'
minimizer.Move(atoms)

# We relax using Verlet, with a bit of randomization using AndersonDynamics
# (which is implemented as a constraint)
andersen = AndersenDynamics(0.01, 0.01)
bulk.SetConstraint(andersen)
verlet = VerletAtomsMover(potential, 0.01, 100)
verlet.Move(atoms) # we thermalize a bit
atoms.Notify()
verlet.Move(atoms) # we thermalize a bit
atoms.Notify()
verlet.Move(atoms) # we thermalize a bit
atoms.Notify()

# We now start opening up the crack
uniform1 = UniformlyMovingBody([-velocity, 0])
uniform2 = UniformlyMovingBody([velocity, 0])
constrZone1.SetConstraint(uniform1)
constrZone2.SetConstraint(uniform2)

#we remove the Andersen constraint
constraint = Constraint()
bulk.SetConstraint(constraint)

mover = verlet

# and open the crack as long as you're willing to watch
while 1:
	mover.Move(atoms)
	E = potential.CalculateEnergy(atoms)
	print "energy = ", E + atoms.GetKineticEnergy()
	atoms.Notify()

# XXX ???
atoms.SetNeighborLocator()
neighborLocator.thisown = 0
