// ASCIIReader.cpp: implementation of the ASCIIReader class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"  // because of the STL warnings under Windows

#include "ASCIIReader.h"
#include "SerialFactory.h"
#include "SprintString.h"
#include "DMException.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ASCIIReader::ASCIIReader(const std::string &filename) : DMReader(filename), SIZE(250)
{
	buffer = new char[SIZE];
}

ASCIIReader::~ASCIIReader()
{
	delete [] buffer;
}

void ASCIIReader::Open()
{
	in.open(filename.c_str(), std::ios::in);
	if(!in)
	{
		std::string error = "Cannot open" + filename;
		throw DMException("ASCIIReader","Open",error);
	}
}

void ASCIIReader::Close()
{
	in.close();
}

std::string ASCIIReader::GetType(const std::string &name)
{
	if(offset == "") Open();
	std::string type, tmp;
	while(!in.eof())
	{	
		if(offset.length() != 0) in.ignore(offset.length());
		in.getline(buffer, SIZE, '\t');
		type = buffer;
		in.getline(buffer, SIZE, ' ');
		tmp = buffer;
		in.getline(buffer, SIZE); // finish the line
		if(tmp == name) 
		{
			offset += "    ";
			return type;
		}
	}
   	throw DMException("ASCIIReader","GetType",name + string("Type not found in the file"));
}

void ASCIIReader::Done()
{
	in.getline(buffer, SIZE, '}');
	if(in.get() != '\n') throw DMException("ASCIIReader","Done","???");
	offset.resize(offset.length()-4);
}

bool ASCIIReader::Fill(Serializable *toFill, const std::string &name)
{
	if(offset == "")
	{
		string type = GetType(name), typein = toFill->GetType();
		if(type != typein) throw DMException("ASCIIReader", "Fill", "expected a "+type + " and recieved a "+typein);
	}
	else
	{
		string type = GetType( name );
	}

	try
	{
		toFill->Load(this);
		Done();
		return true;
	}
	catch (DMException e){
		std::cout<< e;}
	return false;
}

Serializable *ASCIIReader::Get(const std::string &name)
{
	Serializable *serializable = SerialFactory::Instance()->New(GetType(name));
	serializable->Load(this);
	Done();
	return serializable;
}

bool ASCIIReader::GetBool( const std::string& name )
{
	int iRead = GetInt( name );

	bool isBoolean = ( ( 0 == iRead) || ( 1 == iRead ) );
	ASSERT( isBoolean );
	if ( !isBoolean ) throw DMException( "ASCIIReader", "GetBool",
		SprintString("The boolean \"%s\" did not save correctly.",
		name.c_str() ) );

	return ( 1 == iRead );
}


int ASCIIReader::GetInt(const std::string &name)
{
	if(GetType(name) != "int") throw DMException("ASCIIReader", "GetInt",
		SprintString("\"%s\" is not an int!", name.c_str()) );
	int n;
	in >> n ;
	Done();
	return n;
}

double ASCIIReader::GetDouble(const std::string &name)
{
	if(GetType(name) != "double") throw DMException("ASCIIReader","GetInt",
												 "This is not a double!");
	double x;
	in >> x ;
	Done();
	return x;
}

std::string ASCIIReader::GetString(const std::string &name)
{
	if(GetType(name) != "string") throw DMException("ASCIIReader","GetInt",
												 "This is not a string!");
	if(offset.length() != 0) in.ignore(offset.length());
	in.getline(buffer, SIZE);
	std::string str(buffer);
	Done();
	return str;
}


int *ASCIIReader::GetIntArray(std::vector<int> &dims, const std::string &name)
{
	if(GetType(name) != "intArray") throw DMException("ASCIIReader","GetIntArray",
													  "this is not an intArray");
	dims.clear();
	int dim;
	while(in.get() != '\n') {in >> dim; dims.push_back(dim);}
	int tot = 1;
	for(unsigned int i=0; i<dims.size(); ++i) tot *= dims[i]; 
	int *data = new int[tot], *dataPtr = data;
	for(int j=0; j<tot ; ++j) in >> *dataPtr++;
	Done();
	return data;
}

double *ASCIIReader::GetDoubleArray(std::vector<int> &dims, const string &name)
{
	if(GetType(name) != "doubleArray")
		throw DMException("ASCIIReader", "GetDoubleArray",
						  "This is not a doubleArray!");
	dims.clear();
	int dim;
	while(in.get() != '\n') {in >> dim; dims.push_back(dim);}
	int tot = 1; 
	for(unsigned int i=0; i<dims.size(); ++i) tot *= dims[i]; 
	double *data = new double[tot], *dataPtr = data;
	for(int j=0; j<tot ; ++j) in >> *dataPtr++;
	Done();
	return data;
}
