// ConjugateGradientMinimizer.cpp: implementation of the ConjugateGradientMinimizer class.
//
//////////////////////////////////////////////////////////////////////
#include <assert.h>

#include "DigitalMaterial.h"
#include "ConjugateGradientMinimizer.h"
#include "blas.h"

const int ConjugateGradientMinimizer::DISPLAYPRECISION = 12;


#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


ConjugateGradientMinimizer::ConjugateGradientMinimizer(double tolerance, int maximumIterations): GOLD(1.618033988749895), CGOLD(0.381966011250105)
{

	this->tolerance = tolerance;
	this->maximumIterations = maximumIterations; 
	this->gradSq = 0.;
	this->stepsBetweenUpdate = 100;

	LINE_MIN = NEWTON; // default lineMinimizer is NewtonLineMin

	nIterations = 0;
	verbose = false;
	useFunction = true;
	outputStream = &std::cout;
	
	// parameter used by Newton line minimizer
	finiteDifferenceParameter = 1.0e-5;

	// parameters associated with Brent and DBrent line minimizers
	ITMAX_DerivBrentLineMin = 100;
	ZEPS_DerivBrentLineMin = 1.0e-10;
	ZEPS_BrentLineMin = 1.0e-10;
	TOL_LineMinimizer = 3.0e-8; //2.0e-4;// should be about sqrt(machine precision) see Numerical Recipes p398
	EPS_Minimizer = 1.0e-10;
	GLIMIT_Bracket = 100;  // limits range of checking for a bracket during one iteration
	TINY_Bracket = 1.0e-20;  // used to prevent division by zero

}


ConjugateGradientMinimizer::~ConjugateGradientMinimizer()
{

}


/*!

 This is a modified version of Numerical Recipes's "frprmn" (conjugate gradient routine):


 currentPoint is an array provided by the calling function

*/

double ConjugateGradientMinimizer::Minimize(double *currentPoint, Minimizable *minimizable)
{
	double newValue, gamma, currentValue, newGradSq, polakRibiereTerm;
	const double *externalGradientPtr;


	this->minimizable = minimizable;
	nParameters = minimizable->GetNParameters();
	this->currentPoint = currentPoint;
	int oldPrecision = outputStream->precision();
	outputStream->precision(DISPLAYPRECISION);

	nIterations = 0;
	bool done = false;

	// required for Function1D and Derivative1D:
	this->tempPoint = new double[nParameters]; 
	// store current/previous/conjugate search directions etc:
	double *negGradient   = new double[nParameters];
	double *lastDirection = new double[nParameters];
	double *newDirection  = new double[nParameters];

	// get initial function value and gradient
	currentValue = minimizable->ObjectiveFunction(currentPoint);
	externalGradientPtr = minimizable->Gradient(currentPoint); // Initial gradient in newDirection
	
	// initialize direction arrays to be negative of initial gradient
	memcpy(negGradient, externalGradientPtr, nParameters * sizeof(double));
	cblas_dscal(nParameters, -1., negGradient, 1);	
	// copy to newDirection and lastDirection
	memcpy(newDirection, negGradient, nParameters * sizeof(double));
	memcpy(lastDirection, negGradient, nParameters * sizeof(double));


	// norm squared of gradient (dot with itself)
	gradSq = cblas_ddot(nParameters, negGradient, 1, negGradient, 1);

	if(verbose && nIterations%stepsBetweenUpdate == 0)
	{
		(*outputStream) << "Initial function value: " << currentValue << std::endl;
		(*outputStream) << "Initial gradient value: " << gradSq << std::endl;
	}
	

	if( gradSq == 0.0 )
	{
		done = true;
		newValue = currentValue;
	}
	
	// whether to forget about conjugate directions and start again 
	// from the current gradient 
	bool reset;
	

	while(!done && nIterations < maximumIterations)
	{
		// do line minimization along newDirection. This changes the value of currentPoint
		newValue = LineMinimize(newDirection,currentValue,externalGradientPtr);
		if(newValue < currentValue && useFunction)
			reset = false;
		else
			reset = true;
		// this shouldn't happen--it'll be stopped in LineMin
		if( newValue > currentValue)
			cerr << "Line min increased function from " << currentValue << " to " << newValue << endl;
		if(verbose && nIterations%stepsBetweenUpdate == 0)
			(*outputStream) << "Iteration: " << nIterations << " Function value: " << newValue << std::endl;

		// if function value changes little then we are done
		if (useFunction && ( 2.0*fabs(newValue - currentValue) <= tolerance*(fabs(newValue) + fabs(currentValue) + EPS_Minimizer) ))
		{
			done = true;
			break;
		}

		// otherwise compute gradient at new position
		currentValue = newValue;
		externalGradientPtr = minimizable->Gradient(currentPoint);
		// set negGradient to minus current gradient
		memcpy(negGradient, externalGradientPtr, nParameters * sizeof(double));
		cblas_dscal(nParameters, -1., negGradient, 1);
	
		

		newGradSq = cblas_ddot(nParameters, negGradient, 1, negGradient, 1);
		if (newGradSq == 0.0)	
		{
			done = true;
			break;
		}	
		if(verbose && nIterations%stepsBetweenUpdate == 0)
			(*outputStream) << "newGradSq "  << newGradSq << endl;
		// the following is optional. It is the Polak_ribiere modification
		polakRibiereTerm = cblas_ddot(nParameters,externalGradientPtr, 1, negGradient, 1);
	

		// compute gamma (needs old gradSq)
		gamma = (newGradSq + polakRibiereTerm)/gradSq;
		
		// don't need the old gradSq any more, so can overwrite it.
		gradSq = newGradSq;

	
		// newDirection = negGradient + gamma * lastDirection
		memcpy(newDirection, negGradient, nParameters * sizeof(double));
		cblas_daxpy(nParameters, gamma, lastDirection, 1, newDirection, 1);
		// lastDirection is a copy of newDirection, that we'll need for the next iteration
		memcpy(lastDirection, newDirection, nParameters * sizeof(double));
		
		++nIterations;
	}


	delete [] tempPoint;	
	delete [] newDirection;
	delete [] lastDirection;
	delete [] negGradient;

	outputStream->precision(oldPrecision);

	return newValue;
}


double ConjugateGradientMinimizer::LineMinimize(double *direction, double currentValue, const double *currentGradient)
{
	double brackParamA, brackParamX, brackParamB, minLineParam;
	double brackValueA, brackValueX, brackValueB, minLineValue;
	
	
	this->currentDirection = direction;
	
	// Initial guess for the brackets, or, starting value of the line parameter
	// in the case of Newton (which doesn't need a bracketed minimum)
	brackParamA = 0.0;			
	brackParamX = 1.0e-4;

	if(LINE_MIN == BRENT || LINE_MIN == DBRENT)
		Bracket(&brackParamA,&brackParamX,&brackParamB,&brackValueA,&brackValueX,&brackValueB);	
	
	if(LINE_MIN == BRENT)
		minLineValue = BrentLineMin(brackParamA,brackParamX,brackParamB,TOL_LineMinimizer,&minLineParam);
	else if(LINE_MIN == DBRENT)
		minLineValue = DerivBrentLineMin(brackParamA,brackParamX,brackParamB,TOL_LineMinimizer,&minLineParam);
	else if(LINE_MIN == NEWTON)
		minLineValue = NewtonLineMin(brackParamA,brackParamX,brackParamB,TOL_LineMinimizer,&minLineParam, currentGradient);
	
	if(minLineValue > currentValue)
		// bigger-> reject and return the previous value
		// this will be recognized and the main loop will start again
		// with the gradient at the current point as the search direction
		return currentValue;
	else
	{
		// it got smaller
		for ( int pIdx = 0; pIdx < nParameters; ++pIdx )
		{
			direction[pIdx] *= minLineParam;
			currentPoint[pIdx] += direction[pIdx];
		}
		return minLineValue;
	}
}

/*!

  The following two functions are designed to look like simple scalar functions
  of one variable for the purpose of passing them to 1D minimizer routines like
  brent and dbrent

  They obviously need to access the current N-point and direction, as well as 
  have access to a temporary array for storing the N-point corresponding to the
  given value of line_x

 */

double ConjugateGradientMinimizer::Function1D(double linePos)
{
	//for ( int pIdx = 0; pIdx < nParameters; ++pIdx )
	//	this->tempPoint[pIdx] = currentPoint[pIdx] + linePos * this->currentDirection[pIdx];
	memcpy(this->tempPoint,currentPoint,nParameters*sizeof(double));
	cblas_daxpy(nParameters, linePos, currentDirection, 1, tempPoint,1);
	
	double f = minimizable->ObjectiveFunction(tempPoint);
	return f;
}


double ConjugateGradientMinimizer::Derivative1D(double linePos, const double *currentGradient)
{
	int pIdx;

	// go to the point along the line in the current direction 
	for ( pIdx = 0; pIdx < nParameters; ++pIdx )
		this->tempPoint[pIdx] = 
			this->currentPoint[pIdx] + linePos * this->currentDirection[pIdx];

	// calculate the gradient there
	const double *gradLinePos;
	if (!currentGradient)
		gradLinePos = minimizable->Gradient(tempPoint);
	else 
		gradLinePos = currentGradient;
	// convert into a directional derivative along the current direction
	double directionalDeriv = 0.0;
	for ( pIdx = 0; pIdx < nParameters; ++pIdx )
		directionalDeriv += gradLinePos[pIdx] * currentDirection[pIdx];
	
	return directionalDeriv;
}


/*!
  
  Simplest line minimizer: takes a Newton step.
  Equivalent to fitting to a parabola

*/

double ConjugateGradientMinimizer::NewtonLineMin(double ax, double bx, double cx, double tol, double *minLineParam, const double *currentGradient)
{
	// we know the current gradient so no point in computing in a second time
	// this depends on ax being zero here
	double deriv = Derivative1D(ax, currentGradient);
	double derivPlus = Derivative1D(ax + finiteDifferenceParameter);
	
	// estimate position of minimum by assuming parabola
	*minLineParam = -finiteDifferenceParameter * deriv/(derivPlus - deriv);  

	return Function1D(*minLineParam);
}




/*! This is a modified version of numerical recipes "mnbrak"

 */


void ConjugateGradientMinimizer::Bracket(double *ax, double *bx, double *cx, double *fa, double *fb,
						double *fc)

{
	double ulim, u, r, q, fu, dum=0; // Setting dum because compiler gives warning that it's not initialized
	int iter = 0;							// and SWIG doesn't like that

	*fa = Function1D(*ax);
	*fb = Function1D(*bx);

	if (*fb > *fa)
	{
	  dum = *ax; *ax = *bx; *bx = dum;
	  dum = *fb; *fb = *fa; *fa = dum;
	}
	*cx = (*bx) + GOLD * ( *bx - *ax );
	*fc = Function1D( *cx );

	while ( *fb > *fc  && iter < 1000)
	{
		iter++;
		r = ( *bx - *ax ) * ( *fb - *fc );
		q = ( *bx - *cx ) * ( *fb - *fa );
		u = ( *bx ) - ( ( *bx - *cx ) * q - ( *bx - *ax ) * r ) /
			( 2.0 * SIGN(FMAX(fabs(q - r), TINY_Bracket),q - r));
		ulim = (*bx) + GLIMIT_Bracket * ( *cx - *bx ); // Don't go farther than this. Test possibilities
		if (( *bx - u ) * ( u - *cx ) > 0.0)  // parabolic u is between b and c: try it
		{
			fu = Function1D(u);

			if (fu < *fc)		// got a minimum between b and c
			{
				*ax = (*bx);
				*bx = u;
				*fa = (*fb);
				*fb = fu;
				return;
			}
			else if (fu > *fb)
			{
				*cx = u;
				*fc = fu;
				return;
			}
			u = (*cx) + GOLD * ( *cx - *bx );	// Parabolic fit was no use. Use default magnification
			fu=Function1D(u);            // (now go back to top of while loop, after two swaps)

		}
		else if (( *cx - u ) * ( u - ulim ) > 0.0)  // parabolic fit is between c and its allowed limit
		{
			fu = Function1D(u);
			if (fu < *fc)
			{
			  *bx = *cx; *cx = u; u = *cx + GOLD*( *cx - *bx );
			  *fb = *fc; *fc = fu; fu = Function1D(u); 
			}
		} 
		else if (( u - ulim )*( ulim - *cx ) >= 0.0) 
		{
			u = ulim;		// limit parabolic u to maximum allowed value
			fu = Function1D(u);

		}
		else			   // reject parabolic u,use default magnification 
		{
			u = ( *cx ) + GOLD * ( *cx - *bx );
			fu = Function1D(u);

		}
		*ax = *bx; *bx = *cx; *cx = u; // eliminate oldest point and continue
		*fa = *fb; *fb = *fc; *fc = fu;
	}
	if(iter >=1000) throw DMException("ConjugateGradientMinimizer","Bracket","Bracket exceeded maximum steps");
}


double ConjugateGradientMinimizer::DerivBrentLineMin(double ax, double bx, double cx,double tol, double *minLineParam)
{
	int iter,ok1,ok2;
	double a,b,d,d1,d2,du,dv,dw,dx,e=0.0;
	double fu,fv,fw,fx,olde,tol1,tol2,u,u1,u2,v,w,x,xm;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=Function1D(x);
	dw=dv=dx=Derivative1D(x);
	for (iter=0;iter<ITMAX_DerivBrentLineMin;++iter)
	{
		xm=0.5*(a+b);
		tol1=tol*fabs(x)+ZEPS_DerivBrentLineMin;
		tol2=2.0*tol1;
		if (fabs(x-xm) <= (tol2-0.5*(b-a)))
		{
			*minLineParam=x;
			return fx;
		}
		if (fabs(e) > tol1) 
		{
			d1=2.0*(b-a);
			d2=d1;
			if (dw != dx) d1=(w-x)*dx/(dx-dw);
			if (dv != dx) d2=(v-x)*dx/(dx-dv);
			u1=x+d1;
			u2=x+d2;
			ok1 = (a-u1)*(u1-b) > 0.0 && dx*d1 <= 0.0;
			ok2 = (a-u2)*(u2-b) > 0.0 && dx*d2 <= 0.0;
			olde=e;
			e=d;
			if (ok1 || ok2)
			{
				if (ok1 && ok2)
					d=(fabs(d1) < fabs(d2) ? d1 : d2);
				else if (ok1)
					d=d1;
				else
					d=d2;
				if (fabs(d) <= fabs(0.5*olde))
				{
					u=x+d;
					if (u-a < tol2 || b-u < tol2)
						d=SIGN(tol1,xm-x);
				}
				else 
				{
					d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
				}
			} 
			else
			{
				d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
			}
		} 
		else 
		{
			d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
		}
		if (fabs(d) >= tol1)
		{
			u=x+d;
			fu=Function1D(u);
		}
		else
		{
			u=x+SIGN(tol1,d);
			fu=Function1D(u);
			if (fu > fx)
			{
				*minLineParam=x;
				return fx;
			}
		}
		du=Derivative1D(u);
		if (fu <= fx) 
		{
			if (u >= x) a=x; else b=x;
			MOV3(v,fv,dv, w,fw,dw);
			MOV3(w,fw,dw, x,fx,dx);
			MOV3(x,fx,dx, u,fu,du);
		}
		else 
		{
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x)
			{
				MOV3(v,fv,dv, w,fw,dw);
				MOV3(w,fw,dw, u,fu,du);
			}
			else if (fu < fv || v == x || v == w) 
			{
				MOV3(v,fv,dv, u,fu,du);
			}
		}
	}
	throw DMException("ConjugateGradientMinimizer","BrentLineMin","DerivBrentLineMin exceeded max iterations");
	return 0.0;  // Never get here
}



double ConjugateGradientMinimizer::BrentLineMin(double ax, double bx, double cx,double tol,
										 double *minLineParam)

{
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=Function1D(x);
	for (iter=0;iter<ITMAX_DerivBrentLineMin;++iter)
	{
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS_DerivBrentLineMin);
		if (fabs(x-xm) <= (tol2-0.5*(b-a)))
		{
			*minLineParam=x;
			return fx;
		}
		if (fabs(e) > tol1)
		{
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else 
			{
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} 
		else 
		{
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=Function1D(u);
		if (fu <= fx)
		{
			if (u >= x) a=x; else b=x;
			v=w; w=x; x=u;
			fv=fw; fw=fx; fx=fu;
		} else 
		{
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) 
			{
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} 
			else if (fu <= fv || v == x || v == w) 
			{
				v=u;
				fv=fu;
			}
		}
	}
	throw DMException("ConjugateGradientMinimizer","BrentLineMin","BrentLineMin exceeded max iterations");
}


