#include "DigitalMaterial.h"
#include "netcdf.h"
#include "NetCDFStructure.h"
#include "NetCDFError.h"
#include "SprintString.h"
#include "DMLastHeader.h"

/*! Clobbers any pre-existing file.

	\param filename Name of file to create.
	\param fileID Returns the NetCDF ID of the file created.
	*/

void NetCDFStructure::CreateFile( const std::string& filename, int& fileID )
{
	CheckFilename( filename, "CreateFile" );

	int status = nc_create(filename.c_str(), NC_CLOBBER, &fileID);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError( "nc_create", filename, nc_strerror(status));
	}
}

/*!
	\param filename Name of file to open.
	\param access Access mode to use. (from NetCDFStructure interface)
	\param fileID Returned NetCDF ID of file.
	*/
void NetCDFStructure::Open( const std::string& filename, Access access, int& fileID )
{
	CheckFilename( filename, "CreateFile" );

	int ncAccess;
	if ( READONLY == access ) ncAccess = NC_NOWRITE;
	else if ( WRITE == access ) ncAccess = NC_WRITE;
	else {
		bool bInvalidAccessMode = false;
		ASSERT( bInvalidAccessMode );
		throw NetCDFError( "NetCDFStructure", "Open",
			SprintString("Access mode %i is invalid.", access ) );
	}

	int status = nc_open(filename.c_str(), ncAccess, &fileID);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_open",filename,nc_strerror(status));}
}

void NetCDFStructure::Close( int fileID )
{
	int status = nc_close( fileID );
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_close","Close",nc_strerror(status));}
}

////////////////// NAME MANGLING /////////////////////////////

void NetCDFStructure::CheckItemName( const std::string& sItem )
{
	ASSERT( sItem.size() > 0 );
	if ( sItem.size() <= 0 ) throw NetCDFError("NetCDFStructure","CheckItemName",
		"Tried to pass a null name for an item.");
}


std::string NetCDFStructure::LengthMangle( const std::string& sItem )
{
	return sItem + "_nameLength";
}


std::string NetCDFStructure::DimMangle( const std::string& sItem, int dimIdx )
{
	ASSERT( dimIdx >= 0 );
	if ( dimIdx < 0 ) throw NetCDFError( "NetCDFStructure","DimMangle",
		SprintString("Asked to give name to %ith dimension.") );

	std::string sNewName( sItem );
	sNewName.append( "_DimsI" ); // zero gets one "I"
	int i;
	for ( i=0; i < dimIdx; i++ ) sNewName.append("I");

	return sNewName;
}

std::string NetCDFStructure::DataMangle( const std::string& sItem )
{
	return sItem + "_DataMangle";
}

std::string NetCDFStructure::FakeDimMangle( const std::string& sItem )
{
	return sItem + "_FakeDims";
}

////////////////// DIRECTORY FUNCTIONS /////////////////////////

std::string NetCDFStructure::CurDir( const PathType& path )
{
	if ( path.empty() ) {
		return std::string("");
	} // else
	return path.top();
}

std::string NetCDFStructure::AbsolutePath( const PathType& path,
										  const std::string& fileName )
{
	if ( path.empty() ) return fileName;

	return path.top() + "_" + fileName;
}

void NetCDFStructure::UpDir( PathType& path )
{
	ASSERT( !path.empty() );
	if ( path.empty() ) throw NetCDFError( "NetCDFStructure","UpDir",
		"Tried to move up a directory when you are at the root." );

	path.pop();
}

void NetCDFStructure::DownDir( PathType& path, const std::string& dirName )
{
	if ( path.empty() ) {
		path.push( dirName );
	} else {
		std::string sNextPath( path.top() );
		sNextPath.append( "_" );
		sNextPath.append( dirName );
		// Catch exception when there is a memory error.
		path.push( sNextPath );
	}
}


bool NetCDFStructure::IsAtRoot( const PathType& path )
{
	return path.empty();
}

////////// PRIVATE FUNCTIONS ///////////////////////

void NetCDFStructure::CheckFilename( const std::string& filename,
									const std::string& function )
{
	ASSERT( filename.size() != 0 );
	if ( filename.size() == 0 ) throw NetCDFError( "NetCDFStructure",function,
		"Filename has zero size." );
}
