#ifndef _RWLOCK_H_
#define _RWLOCK_H_

#include "DigitalMaterial.h"
#include "RWLockImp.h"

//! Read-write lock
/**
 * This is the class you work with when you make a read-write lock.
 * A read-write lock models a resource which can have multiple
 * readers but only one writer at a time.
 * You create it with a call to ProcessFactory::GetRWLock().
 * \code
 * RWLock *lock = ProcessFactory::Instance()->GetRWLock();
 * \endcode
 *
 * This class is just a wrapper on a system-dependent implementation
 * of a read-write lock.  On construction, it chooses the correct kind
 * for your system.  Then it forwards all calls to that correct
 * implementation.
 * \author Drew Dolgert
 * \date March 2000
 * \sa RWLockImp, WinRWLock, UnixRWLock
 * \ingroup dmsystem
 */
class RWLock : public RWLockImp
{
public:
	//! Initialize the lock.
	/*! You must call InitializeLock before calling any function
		except CleanupLock(). */
	virtual bool InitializeLock();
	/*! If you have called InitializeLock(), you must call CleanupLock()
		before the lock is deleted. */
	virtual void CleanupLock() { _lock->CleanupLock(); delete _lock; }
	//! Obtain a lock to read from the resource.
	/*! There can be multiple clients with read locks at the same time. */
	virtual void ReadLock() { _lock->ReadLock(); }
	//! Obtain a lock to write to the resource.
	/*! Only one client at a time may have write permission. */
	virtual void WriteLock() { _lock->WriteLock(); }
	//! Release the read lock.
	virtual void ReadRelease() { _lock->ReadRelease(); }
	//! Release the write lock.
	virtual void WriteRelease() { _lock->WriteRelease(); }
private:
	RWLockImp* _lock;
};

#endif // _RWLOCK_H_
