/*
 * SimQA: Copyright (C) Jakob Schiotz 1996, 1997
 *
 * Reimplementation of some functions in the IBM mass library, for use on
 * other architechtures.
 *
 * vsqrt - Vectorized square root.
 * vexp  - Vectorized exponential function.
 * vrec  - Vectorized reciprocal.
 * vlog  - Vectorized natural logarithm.
 */

#include <stdlib.h>
#include <math.h>

#include "mass.h"

void vsqrt(double *res, const double *a, const int *n)
{
  int i, limit;
  double a0, a1, a2, a3, a4, a5, a6, a7;

  limit = (*n)-7;
    for (i = 0; i < limit; i+=8)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	a4 = a[i+4];
	a5 = a[i+5];
	a6 = a[i+6];
	a7 = a[i+7];
	res[i] = sqrt(a0);
	res[i+1] = sqrt(a1);
	res[i+2] = sqrt(a2);
	res[i+3] = sqrt(a3);
	res[i+4] = sqrt(a4);
	res[i+5] = sqrt(a5);
	res[i+6] = sqrt(a6);
	res[i+7] = sqrt(a7);
    }
    for (i = (*n) - (*n) % 8; i < *n; i++)
    {
	a0 = a[i];
	res[i] = sqrt(a0);
    }
}

  
void vexp(double *res, const double *a, const int *n)
{
  int i, limit;
  double a0, a1, a2, a3, a4, a5, a6, a7;

  limit = (*n)-7;
    for (i = 0; i < limit; i+=8)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	a4 = a[i+4];
	a5 = a[i+5];
	a6 = a[i+6];
	a7 = a[i+7];
	res[i] = exp(a0);
	res[i+1] = exp(a1);
	res[i+2] = exp(a2);
	res[i+3] = exp(a3);
	res[i+4] = exp(a4);
	res[i+5] = exp(a5);
	res[i+6] = exp(a6);
	res[i+7] = exp(a7);
    }
    for (i = (*n) - (*n) % 8; i < (*n); i++)
    {
	a0 = a[i];
	res[i] = exp(a0);
    }
}



void vrec(double *res, const double *a, const int *n)
{
  int i, limit;
  double a0, a1, a2, a3, a4, a5, a6, a7;

  limit = (*n)-7;
    for (i = 0; i < limit; i+=8)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	a4 = a[i+4];
	a5 = a[i+5];
	a6 = a[i+6];
	a7 = a[i+7];
	res[i] = 1.0 / a0;
	res[i+1] = 1.0 / a1;
	res[i+2] = 1.0 / a2;
	res[i+3] = 1.0 / a3;
	res[i+4] = 1.0 / a4;
	res[i+5] = 1.0 / a5;
	res[i+6] = 1.0 / a6;
	res[i+7] = 1.0 / a7;
    }
    for (i = (*n) - (*n) % 8; i < (*n); i++)
    {
	a0 = a[i];
	res[i] = 1.0 / a0;
    }
}

void vlog(double *res, const double *a, const int *n)
{
  int i, limit;
  double a0, a1, a2, a3, a4, a5, a6, a7;

  limit = (*n)-7;
    for (i = 0; i < limit; i+=8)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	a4 = a[i+4];
	a5 = a[i+5];
	a6 = a[i+6];
	a7 = a[i+7];
	res[i] = log(a0);
	res[i+1] = log(a1);
	res[i+2] = log(a2);
	res[i+3] = log(a3);
	res[i+4] = log(a4);
	res[i+5] = log(a5);
	res[i+6] = log(a6);
	res[i+7] = log(a7);
    }
    for (i = (*n) - (*n) % 8; i < (*n); i++)
    {
	a0 = a[i];
	res[i] = log(a0);
    }
}


