#include "DMFirstHeader.h"
#include <stdexcept>
#include <new>
#include <stdarg.h>
#include <stdio.h>
#include "DigitalMaterial.h"
#include "SprintString.h"
#include "DMException.h"
#include "DMLastHeader.h"

const int SprintString::BufferCnt = 400;
char SprintString::threadUnsafeBuffer[BufferCnt];

/*!	This creates an STL string from a printf-like constructor.
	It uses temporary storage on the stack for small strings but
	dynamically allocates memory for larger strings.  If you use
	this for exceptions, avoid huge strings, especially if, for
	instance, the exception declares that you are out of memory.
	*/
SprintString::SprintString( const char* sMessage, ... )
{
	ASSERT( 0 != sMessage );
	if ( 0 == sMessage ) throw DMException("SprintString","SprintString",
		"Passed null pointer in constructor." );
	
	char buffer[ BufferCnt ];
	va_list marker;
	
#ifdef HAVE_VSNPRINTF
	// Try to write the string to a buffer on the stack.
	va_start( marker, sMessage );
	int writtenNotIncludingNull = vsnprintf( buffer, BufferCnt-1, sMessage, marker );
	va_end( marker );

	// If the string fit in the buffer,
	if ( writtenNotIncludingNull >= 0 ) {
		this->append( buffer, writtenNotIncludingNull );

	} else {
		// Write the string to a dynamically allocated buffer.
		char *sBuff;
		int bufferMultiple = 1;
		int messageCnt = strlen( sMessage ) + 1;
		do {
			int buffCnt = messageCnt + bufferMultiple * BufferCnt;
			bufferMultiple++;
                        
			try {
				sBuff = new char[ buffCnt ];
				} catch ( std::bad_alloc& ) {
				sBuff = 0;
				}
			ASSERT( 0 != sBuff );
			if ( 0 == sBuff ) {
				// If we can't allocate memory,
				//   then use our inviolable stash of memory to get the message out.
				va_start( marker, sMessage );
				vsnprintf( threadUnsafeBuffer, BufferCnt-1, sMessage, marker );
				va_end( marker );
				this->append( threadUnsafeBuffer );
				return;
			}

			va_start( marker, sMessage );
			writtenNotIncludingNull = vsnprintf( sBuff, buffCnt-1, sMessage, marker );
			va_end( marker );

		} while ( writtenNotIncludingNull < 0 );
		
		this->append( sBuff, writtenNotIncludingNull );
		delete[] sBuff;
	}
#else

	// Try to write the string to a buffer on the stack.
	va_start( marker, sMessage );
	int writtenNotIncludingNull = vsprintf( buffer, sMessage, marker );
	va_end( marker );

	ASSERT( writtenNotIncludingNull < BufferCnt );
	if ( writtenNotIncludingNull >= BufferCnt ) {
	  throw DMException( "SprintString", "SprintString", "Overwrote internal buffer.  Memory error.");
	}
	this->append( buffer, writtenNotIncludingNull );

#endif
}
