//! Abstract interface to simulation threads.
/**
 * This is the base class for a system-dependent implementation of a thread class.
 * \ingroup dmsystem
 * \author Drew Dolgert
 * \date 21 March 2000
 * \bug There shouldn't be a toggle function.  It is functionally redundant.
 * \sa Thread, WinThread, UnixThread, Runnable, Threadable
 */
#ifndef _THREADIMP_H_
#define _THREADIMP_H_

#ifndef SWIG
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif

#include "Threadable.h"

class ThreadImp
{
public:
	//! Possible return values.
	enum { SUCCESS, NO_THREAD, ALREADY_PAUSED, THREAD_EXISTS, THREAD_TERMINATED };
	//! Give the thread a task whose Runnable::Run method it can call.
	/*! Pass by reference because ThreadImp is not
		responsible for threadable deletion. */
	virtual void SetTask(Runnable*) = 0;
	//! Give the thread a task whose Runnable::Run method it can call.
	/*! If the task is Threadable, then it agrees to call IsRunning() every
		once in a while.  All our methods to control the task rely on the
		IsRunning() call in order to function gracefully. */
	virtual void SetTask(Threadable*) = 0;
	//! Start the thread.
	/*! \return enum value */
	virtual int Start() = 0;
	//! Stop the thread.
	/*! \return enum value */
	virtual int Stop() = 0;
	//! Pause the thread.  Fails for a running thread.
	/*! \return enum value */
	virtual int Pause() = 0;
	//! Resume the thread from a paused state.
	/*! \return enum value */
	virtual int Resume() = 0;
	//! Resume if paused.  Pause if running.
	/*! \return enum value */
	virtual int Toggle() = 0;
	//! Interface to thread.  Do not call this directly.
	/*! This method is called by the Threadable class which, in turn, was
		called by a subclass of Threadable which runs the task at hand.
		Don't call it yourself.
		\return false if we request that the task should return
		from its Runnable::Run method.
	*/
	virtual bool IsRunning() = 0;
};

#endif //_THREADIMP_H_
