// TinyVector.h: interface for the TinyVector class.
//
//////////////////////////////////////////////////////////////////////
#ifndef TINYVECTOR_H
#define TINYVECTOR_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
//! A class that provide a nice interface to built in Type[N]
/*! We cannot use arrays in STL, because they lack a copy constructor.
This class interfaces provides an interface to arrays by providing
various constructors, assignment operators and usual operators.
\author Thierry Cretegny
\ingroup DMSystem
*/

template<class Type, int N=DIMENSION>
class TinyVector
{
public:
    TinyVector(Type type) { for (int i=0; i<N; ++i) data[i] = type;}
    TinyVector() {}
    TinyVector(const Type type[N]) {memcpy(data, type, N*sizeof(Type));}
	// ~TinyVector() {} // no destructor necessary. The compiler will renerage one by default
	// we also do not need a copy constructor, since the compiler generates a default one, that will
	// be more efficient.
    const Type &operator[](int i) const {return data[i];}
    Type &operator[](int i) {return data[i];}
    TinyVector<Type, N> operator+(const TinyVector<Type, N> &toAdd) const;
    TinyVector<Type, N> operator-(const TinyVector<Type, N> &toSub) const;
    TinyVector<Type, N> &operator+=(const TinyVector<Type, N> &added) { for (int i=0; i<N; ++i) data[i] += added[i];return *this;}
    TinyVector<Type, N> &operator-=(const TinyVector<Type, N> &subst) { for (int i=0; i<N; ++i) data[i] -= subst[i];return *this;}
    TinyVector<Type, N> &operator*=(const Type &mult) {for (int i=0; i<N; ++i) data[i] *= mult;return *this;}

    TinyVector<Type, N> Cross(const TinyVector<Type, N> &vect) const;
    Type norm2() const {Type norm2=0; for (int i=0; i<N; ++i) norm2 += data[i]*data[i];return norm2;}

    inline Type operator*(const TinyVector<Type, N>& v) const ;
    
    operator Type*() {return data;}
    operator const Type*() const {return data;}
private:
    Type data[N];
};

template<class Type, int N>
TinyVector<Type, N> TinyVector<Type, N>::operator+(const TinyVector<Type, N> &toAdd) const
{
    Type res[N];
    for(int i=0; i<N; ++i) res[i] = data[i] + toAdd[i];
    return TinyVector<Type, N>(res);
}

template<class Type, int N>
TinyVector<Type, N> TinyVector<Type, N>::operator-(const TinyVector<Type, N> &toSub) const
{
    Type res[N];
    for(int i=0; i<N; ++i) res[i] = data[i] - toSub[i];
    return TinyVector<Type, N>(res);
}

template<class Type, int N>
Type TinyVector<Type, N>::operator*(const TinyVector<Type, N>& v) const
{
    Type dot = Type();
    for(int i=0; i<N; ++i) dot += data[i] * v.data[i];
    return dot;
}

template<>
inline TinyVector<double, 3> TinyVector<double, 3>::Cross(const TinyVector<double, 3> &vect) const
{
    double tmp[3] = {data[1]*vect.data[2] - data[2]*vect.data[1],
                     data[2]*vect.data[0] - data[0]*vect.data[2],
                     data[0]*vect.data[1] - data[1]*vect.data[0]};
    return TinyVector<double, 3>(tmp);
}

template<>
inline TinyVector<int, 3> TinyVector<int, 3>::Cross(const TinyVector<int, 3> &vect) const
{
    int tmp[3] = {data[1]*vect.data[2] - data[2]*vect.data[1],
                  data[2]*vect.data[0] - data[0]*vect.data[2],
                  data[0]*vect.data[1] - data[1]*vect.data[0]};
    return TinyVector<int, 3>(tmp);
}

#endif // TINYVECTOR_H
