#include "EnergyConservation.h"

/* This demonstration main function runs a 3D simulation of Silicon using the
   EDIP potential and periodic boundary conditions. Parameters such as time
   step are listed at the top of the function and are mostly self explanatory.

   More fundamental changes are to be made as follows: the value of DIMENSION
   is set in the makefile. Be sure to change the if statement at the top of 
   main to be consistent with the DIMENSION, and be sure that the potential
   allows the chosen dimension. To change the potential simply change the name
   of the class used to create the Potential object. If the lattice type and
   spacing of the new potential are available, change these as well. This is
   not necessary really for just checking energy conservation, which should 
   hold regardless of configuration, but if the lattice spacing is much smaller
   than the actual ground state spacing, then atoms may be too close for the 
   time step and go unstable
*/

void main()
{
	try
	{
		ASSERT(DIMENSION == 3);
		if(3 != DIMENSION)
			throw DMException("global","main","wrong DIMENSION");
				
		// The following must occur first thing in a C++ main function
		// or the program will seg-fault (taken care of behind the 
		// scenes in Python applications)
#ifdef _WIN32
		WinProcessFactory::Create();
#else
		UnixProcessFactory::Create();
#endif
		cout.precision(16); // show up to machine precision (otherwise fluctuations in energy may not be apparent)
		
		int coord;
		// plotting
		int plotting = 1, plotRadius = 15;
		// time steps
		double dt = 0.08;
		int nMajorSteps = 100;
		int nMinorSteps = 10;
		// thermalization
		double temperature = 0.1;
		double friction = 0.1;
		// geometry
		double latticeConst = 5.430; // Silicon SW or EDIP potential
		double superCellLength = 3 * latticeConst;
		double lengths[DIMENSION];
		for( coord = 0; coord < DIMENSION; ++coord )
			lengths[coord] = superCellLength;
		
		
		// create ListOfAtoms, Potential, BoundaryConditions and NeighborLocator
		ListOfAtoms *atoms = new DynamicListOfAtoms();
		PeriodicBoundaryConditions *bc = new SimplePeriodicBoundaryConditions();
		bc->SetLength(lengths);
		Potential *potential = new SiliconEDIPPotential();        
		NeighborLocator *neighborLocator = new SimpleNeighborLocator(potential->GetCutoffDistance());  
		
		atoms->SetBoundaryConditions(bc);
		atoms->SetNeighborLocator(neighborLocator);
		
		// create movers
		VerletAtomsMover *verlet = new VerletAtomsMover();
		verlet->SetDt(dt);
		verlet->SetNSteps(nMinorSteps);
		verlet->SetPotential(potential);
		
		LangevinAtomsMover *langevin = new LangevinAtomsMover();
		langevin->SetDt(dt);
		langevin->SetTemperature(temperature);
		langevin->SetPotential(potential);
		langevin->SetFriction(friction);
		langevin->SetNSteps(nMinorSteps);
		
		// create lattice and initialize the ListOfAtoms
		// here it's a diamond lattice for Silicon
		BravaisLatticeWithBasis *lattice = new DiamondLattice(latticeConst);
		RectangularClusterInitializer *cluster = new RectangularClusterInitializer(lengths,lattice);
		double clusterCenter[DIMENSION];
		for(coord=0;coord<DIMENSION;++coord)
			clusterCenter[coord] = lengths[coord]/2.;
		cluster->SetCenter(clusterCenter);
		cluster->Create(atoms);
		
		// the following is useful---to be able to subtract off energies later when calculating averages (helps round-off error)
		double groundEnergy = potential->CalculateEnergy(atoms);
		// thermalize (otherwise the atoms will just stay in their ground state positions)
		langevin->Move(atoms);
		
		// add the plotting observer
		PlotAtomsObserver *plot = 0;
		if(plotting)
		{
			plot = new PlotAtomsObserver(lengths[0], lengths[1], plotRadius);
			atoms->AddObserver(plot);
		}
		

		// main loop
		double aveEnergy = 0.;
		double aveEnergySq = 0.;
		for ( int i = 0; i < nMajorSteps; ++i)
		{
			verlet->Move(atoms);
			atoms->Notify();
			double kinetic = atoms->GetKineticEnergy();
			double PE = potential->CalculateEnergy(atoms);
			double totalEnergy = kinetic + PE;
			cout << "Total Energy is " << totalEnergy << endl;
			aveEnergy += totalEnergy - groundEnergy;
			aveEnergySq += (totalEnergy - groundEnergy) * (totalEnergy - groundEnergy);
		}
		aveEnergy /= nMajorSteps;
		aveEnergySq /= nMajorSteps;
		cout << "Average energy is " << groundEnergy + aveEnergy << "; variance is " << aveEnergySq - aveEnergy * aveEnergy << endl;


		// clean up
		delete atoms;
		delete neighborLocator;
		delete bc;
		delete verlet;
		delete langevin;
		delete lattice;
		delete plot;
	}
	catch(DMException exception)
	{
	
	}
}
