#include "TwoComponentLJ.h"

// This is a main function designed to test the MultiCutLennardJonesPotential class, which does Lennard-Jones for many species.




void main()
{

    cout.precision(16);
    try 
    {
#ifdef _WIN32
		WinProcessFactory::Create();
#else
		UnixProcessFactory::Create();
#endif
		int plotting      = 1;
		int drawRadius = 20;
		int stepsBetweenDraws = 5;
		
		double dt		= 0.005;  // same as SW paper
		double dtQM	= 0.04;
		double dtNEB      = 0.015;
		double dtContinuumMover = 0.04;
		double tolQM	= 1.0e-18;
		double tolNEB = 1.0e-9;
		double tolConjGrad = 1.0e-10;	
		int maxItConjGrad = 10;
		int maxQuickMinSteps = 60;		
		


		// simulation parameters
		double cellLength = 12.0;
		double lengths[DIMENSION] = {cellLength,cellLength,cellLength};
		double latticeConst = cellLength/4.;

		double startTemp = 0.02;
		double endTemp = 0.15;
		double tempStep = 0.1;
		double friction  = 0.1;
		int nAveragingSteps = 15000;
		int nEquilSteps = 50;

		int coord,a,b;
		

		int numberOfTypes = 2;
		double **epsilon,**sigma;

		epsilon = new double*[numberOfTypes];
		sigma = new double*[numberOfTypes];
		for(a=0;a<numberOfTypes;++a)
		{
			epsilon[a] = new double[numberOfTypes];
			sigma[a] = new double[numberOfTypes];
		}
		
		epsilon[0][0] = 1.5;
		epsilon[0][1] = epsilon[1][0] = 1.;
		epsilon[1][1] = 0.75;

		sigma[0][0] = 1.0;
		sigma[0][1] = sigma[1][0] = 1.5;
		sigma[1][1] = 1.2;

		
		// create the main objects
		PrimitiveListOfAtoms *leaf0  = new PrimitiveListOfAtoms();
		PrimitiveListOfAtoms *leaf1  = new PrimitiveListOfAtoms();
		BoundaryConditions *bc = new SimplePeriodicBoundaryConditions(lengths);
		MultiCutLennardJonesPotential *potential    = new MultiCutLennardJonesPotential();        
		//CutLennardJonesPotential *potential    = new CutLennardJonesPotential();        
		potential->SetNumberOfTypes(numberOfTypes);
		potential->SetEpsilons(epsilon);
		potential->SetSigmas(sigma);
		

		//NeighborLocator *neighborLocator = new SimpleNeighborLocator(potential->GetCutoffDistance());      
		NeighborLocator *neighborLocator = new CellNeighborLocator(potential->GetCutoffDistance(),potential->GetCutoffDistance());      
		LangevinAtomsMover *langevin= new LangevinAtomsMover(potential,dt);
		VerletAtomsMover *verlet = new VerletAtomsMover(potential,dt);
		QuickMinAtomsMover *qm = new QuickMinAtomsMover(potential, dtQM, tolQM);
		ConjugateGradientMinimizer *conjGrad = new ConjugateGradientMinimizer(tolConjGrad,maxItConjGrad);
		conjGrad->SetLineMinimizer(0);
		SteepestDescentMinimizer *steepest = new SteepestDescentMinimizer(1.e-4,10000,0.005);
		steepest->Verbose();     
		
		
		double origin[DIMENSION];
		for(coord=0;coord<DIMENSION;++coord)
			origin[coord] = 0.;
		PrimitiveFCCLattice lattice(latticeConst);

		RectangularClusterInitializer cluster(lengths,&lattice);
		double clusterCenter[DIMENSION];
		for(coord=0;coord<DIMENSION;++coord)
			clusterCenter[coord] = lengths[coord]/2.;
		cluster.SetCenter(clusterCenter);
		cluster.Create(leaf0);
		origin[0] += 0.5*latticeConst;
			lattice.SetOrigin(origin);
		cluster.Create(leaf1);
		
		PrimitiveListOfAtoms *atoms = new PrimitiveListOfAtoms();
		atoms->Attach(leaf0);
		atoms->Attach(leaf1);

		cout << "Number of atoms is " << atoms->GetNumber() << endl;

		atoms->SetBoundaryConditions(bc);
		atoms->SetNeighborLocator(neighborLocator);

		atoms->SetTypeName("mainAtoms");


		// Attach a PlotAtomsObserver
#ifndef _WIN32
		PlotAtomsObserver *plot = NULL; 
		if(plotting)
		{
			plot = new PlotAtomsObserver(lengths[0],lengths[1],drawRadius);
			plot->SetColor(0,0);
			plot->SetColor(1,1);
			atoms->AddObserver(plot);
			atoms->Notify();
		}
#endif // _WIN32		

		double averagePotentialEnergy;


		// loop through temperatures 
		for(double temperature=startTemp;temperature<=endTemp;temperature+=tempStep)
		{
		   

			langevin->SetFriction(friction);
			/*	double multiFriction[DIMENSION];
			for(int coord=0;coord<DIMENSION;++coord)
				multiFriction[coord] = friction;
				langevin->AddFriction("mainAtoms",multiFriction);*/
			langevin->SetTemperature(temperature);
			
			// equilibration
			langevin->SetNSteps(nEquilSteps);			
			langevin->Move(atoms);


			// Move the atoms drawing every 50, also sampling for averages (later)
			langevin->SetNSteps(stepsBetweenDraws);
			verlet->SetNSteps(stepsBetweenDraws);
			int nMajorSteps = nAveragingSteps/stepsBetweenDraws;

			averagePotentialEnergy = 0.;

			for(int movidx=0;movidx<nMajorSteps;++movidx)
			{
				langevin->Move(atoms); 
				//verlet->Move(atoms); // for checking energy conservation
				atoms->Notify();
				// calculate some averages
				double potentialEnergy = potential->CalculateEnergy(atoms);
				double kineticEnergy = atoms->GetKineticEnergy();
				averagePotentialEnergy += potentialEnergy;
				//cout << "Potential Energy per atom = " << potentialEnergy/atoms->GetNumber() << endl;
				cout << "Total energy " << potentialEnergy + kineticEnergy << endl;
				
			}
			
			averagePotentialEnergy /= nMajorSteps*atoms->GetNumber();
			cout << "Temperature: " << temperature << " PE per atom: " << averagePotentialEnergy << endl;
		}
		
	
		delete atoms; 
		delete bc;
		delete neighborLocator;
		delete potential;
		delete langevin;
		delete qm;
		delete conjGrad;
		delete steepest;

	
    }
	catch(DMException exception)
    {
		std::cerr << exception;
    }
	catch(std::runtime_error error)
    {
		std::cerr << "std::runtime_error: " << error.what() << endl;
    }
	catch(...)
    {
		std::cerr << "Unknown error" << endl;
    }
	
	return 0;


}
