// AtomicSurfaceInitializer.h: interface for the AtomicSurfaceInitializer class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ATOMICSURFACEINITIALIZER_H
#define ATOMICSURFACEINITIALIZER_H

#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#include "AtomsInitializer.h"

class ListOfAtoms;
class BravaisLatticeWithBasis;
class SkewPeriodicBoundaryConditions;

//! Initializes a surface---a layer of dynamic atoms over a layer of fixed atoms; alternatively (mode 1), it creates a single rectangular cluster with one side corresponding to the given surface orientation. 

/*!
  There are two different modes of operation. One might have different classes
  but a lot of code is the same. Maybe we will have a second class which 
  derives from the first...for now, one class, two modes.

  In mode 0:

  The spbc is set according to the surface orientation, two layers of
  atoms, one fixed, one free, are created and the top atoms have their z
  coordinate equal to 0.

  In mode 1: spbc is not changed but an error is thrown if the SuperCellVectors
  are not consistent with the orientation. A single rectangular cluster is
  created of given height and position in the z-direction

  */

#if DIMENSION == 3

class AtomicSurfaceInitializer : public AtomsInitializer
{
public:
	AtomicSurfaceInitializer( BravaisLatticeWithBasis *lattice, int millerIndices[DIMENSION], int mode = 0, SkewPeriodicBoundaryConditions *spbc = 0);
	virtual ~AtomicSurfaceInitializer();
	void SetMode(int mode);
	int GetMode() const { return mode; }
	void SetDepthOfFixedLayer(double depthOfFixedLayer);
	double GetDepthOfFixedLayer() const {return depthOfFixedLayer;}
	void SetDepthOfDynamicLayer(double depthOfDynamicLayer);
	double GetDepthOfDynamicLayer() const {return depthOfDynamicLayer;}
	void SetMillerIndices(const int millerIndices[DIMENSION]);
	const int *GetMillerIndices() const {return (const int *)millerIndices;}
	void SetMinimalLength(double minimalLength);
	double GetMinimalLength() const {return minimalLength;}

	void SetZCenter(double zCenter) { this->zCenter = zCenter; }
	double GetZCenter() const { return zCenter; }

	void ConfigureBoundaryConditions();

	const double *GetCenter() const {return center;}	//! Rough center (for graphics)
	double GetMaxSize() const;	//! Rough size (for graphics)

	void Create(ListOfAtoms *atoms);

	void SetBravaisLattice(BravaisLatticeWithBasis *lattice);
	BravaisLatticeWithBasis *GetBravaisLattice() const;

	bool SimplifiedVector(const int k,const int l, const int m) ;
private:
	void CreateModeZero(ListOfAtoms *atoms);
	void CreateModeOne(ListOfAtoms *atoms);
	BravaisLatticeWithBasis *lattice;
	int mode;
	double center[DIMENSION];
	double zCenter;
	double surfaceVectors[DIMENSION][DIMENSION];
	double superCellLength;
	double superCellWidth;
	double depthOfFixedLayer;
	double depthOfDynamicLayer;
	double minimalLength;
	int millerIndices[DIMENSION];
	void ResetPrivateData();
	SkewPeriodicBoundaryConditions *spbc;
	static const double TINY;
};

#endif // DIMENSION == 3

#endif // !defined(ATOMICSURFACEINITIALIZER_H)
