// CylindricalClusterInitializer.cpp: implementation of the CylindricalClusterInitializer class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"
#include "CylindricalClusterInitializer.h"

#include "BravaisLatticeWithBasis.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CylindricalClusterInitializer::CylindricalClusterInitializer(double radius, double length, double direction[DIMENSION], BravaisLatticeWithBasis *bravais) : ClusterInitializer(bravais)
{	
	this->radius = fabs(radius);
	this->length = fabs(length);
	for(int coord=0;coord<DIMENSION;++coord)
		this->direction[coord] = direction[coord];

	// avoid calling SetMaxSize three times---don't use the Set functions
	SetMaxSize();
}

CylindricalClusterInitializer::~CylindricalClusterInitializer()
{

}

void CylindricalClusterInitializer::SetRadius(double radius)
{
	this->radius = fabs(radius);
	SetMaxSize();
}

int CylindricalClusterInitializer::SetMaxSize()
{
	int coord;
	// first ensure the direction is of unit length
	double dSq = 0.;
	for(coord=0;coord<DIMENSION;++coord)
		dSq += direction[coord] * direction[coord];
	
	for(coord=0;coord<DIMENSION;++coord)
		direction[coord] /= sqrt(dSq);

	// Calculate maxsize[DIMENSION]...mostly the same as th espherical case,
	// except here I define extRadius as sqrt(radius*radius + length/2 * length/2) + DIMENSION * (etc)
	// to get the sphere which contains the cylinder. This is too big if radius
	// and length/2 are very different but I don't want to have to think about
	// something more efficient righ now



	double volumeOfSphere = 1.;
    // extRadius is an extended radius, used to estimate how many cells are enclosed in the sphere
	double extRadius = sqrt(radius * radius + 0.25 * length * length) +  DIMENSION*pow(bravais->GetVolumeOfCell(),1/(double)DIMENSION);

	double latticevect[DIMENSION], cartvect[DIMENSION], temp[DIMENSION];
	for(coord=0;coord<DIMENSION;coord++) temp[coord] = 0.;
	for(int k=0; k<DIMENSION; k++) 
	{
		for(coord=0;coord<DIMENSION;coord++) cartvect[coord] = (coord==k ? 1. : 0.);
		bravais->LatticeCoordinatesFromCartesianCoordinates(latticevect,cartvect);
		for(coord=0;coord<DIMENSION;coord++) temp[coord] += latticevect[coord]*latticevect[coord];
	}
	for(coord=DIMENSION; coord>=2;coord-=2) volumeOfSphere *= extRadius*extRadius*2.*PI/(double)coord ;
	if(coord==1) volumeOfSphere *= 2.*extRadius;
	for(coord=0;coord<DIMENSION;coord++)
	{
		maxSize[coord] = int(sqrt(temp[coord])*radius) + 1;
		minSize[coord] = -maxSize[coord];
	}
	double shift[DIMENSION];
	bravais->LatticeCoordinatesFromCartesianCoordinates(shift, clusterCenter);
	for(coord=0; coord<DIMENSION; coord++) 
	{
		maxSize[coord] += (int) shift[coord];
		minSize[coord] += (int) shift[coord];
	}
	return int(volumeOfSphere*bravais->GetNAtomsPerCell()/bravais->GetVolumeOfCell());
}

bool CylindricalClusterInitializer::Inside(double X[DIMENSION])
{
	int coord;
	double rPerp2 = 0, rPar2 = 0., dotProduct = 0., vect[DIMENSION], parallelComp[DIMENSION], perpComp[DIMENSION];

	// some of the following is borrowed from Quasi2DElasticField
	for(coord=0;coord<DIMENSION;coord++) 
    {
		vect[coord] = X[coord] - clusterCenter[coord];
		dotProduct += vect[coord]*direction[coord];
    }

	for(coord=0;coord<DIMENSION;coord++)
    {
		parallelComp[coord] = dotProduct*direction[coord];
		perpComp[coord] = vect[coord] - parallelComp[coord];
		rPerp2 += perpComp[coord]*perpComp[coord];
		rPar2 += parallelComp[coord] * parallelComp[coord];
    }

	// the following way of dealing with boundaries is borrowed from
	// RectangularClusterInitializer
	bool inLengthwise = true;
	double smallFraction = 1e-14;
	
	if ((length!=0) && 
		(fabs((fabs(dotProduct)- 0.5 * length) / length) < smallFraction))
	{
		// Near boundary
		if (dotProduct < 0) inLengthwise = false;
	}
	else
	{
		inLengthwise = (fabs(dotProduct) <= 0.5 * length);
	}

	return (rPerp2 < radius * radius && inLengthwise);
}


void CylindricalClusterInitializer::SetLength(double length)
{
	this->length = fabs(length);
	SetMaxSize();
}

void CylindricalClusterInitializer::SetDirection(double direction[DIMENSION])
{
	for(int coord=0;coord<DIMENSION;++coord)
		this->direction[coord] = direction[coord];
	SetMaxSize();
}
