// PolyClusterInitializer.cpp: implementation of the PolyClusterInitializer class.
//
//////////////////////////////////////////////////////////////////////
#include "DMFirstHeader.h"
#include <vector>
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include "DigitalMaterial.h"
#include "PolyClusterInitializer.h"
#include "BravaisLatticeWithBasis.h"
#include "DMLastHeader.h"

#ifndef MAX_INT
#define MAX_INT   2147483647 
#endif

using namespace std ;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

PolyClusterInitializer::PolyClusterInitializer(DMArray<double, DIMENSION>* pcorners, BravaisLatticeWithBasis *bravais, 
											   DMArray<int, DIMENSION>* pconnectivities) : ClusterInitializer(bravais)
{
	SetCorners(pcorners);
	if (pconnectivities) {
		SetConnectivities(pconnectivities) ;
		givenConnectivities = true ;
	}
	else {
		SetConvexConnectivities() ;
		givenConnectivities = false ;
	}
}

PolyClusterInitializer::~PolyClusterInitializer()
{
	for (unsigned int i =0; i<faces.size(); ++i)
		delete faces[i] ;
}

void PolyClusterInitializer::SetCorners(DMArray<double, DIMENSION>* pcorners)
{
    corners.SetShape(pcorners->Size()) ;
    corners.Copy(pcorners) ;
}

void PolyClusterInitializer::SetConnectivities(DMArray<int, DIMENSION>* pconnectivities)
{
    connectivities.SetShape(pconnectivities->Size()) ;
    connectivities.Copy(pconnectivities) ;
}

void PolyClusterInitializer::SetConvexConnectivities()
{
	int i[DIMENSION], setOfCorners;
	unsigned int j, k, l, temp;
	bool different, newFace;
	double matrix[DIMENSION][DIMENSION], tempmatrix[DIMENSION][DIMENSION];
	double normal[DIMENSION] ;
	double normOfNormal, dotProduct, norm;
	unsigned int numberOfCorners = corners.Size() ;
	vector<face* > tempFaces;
	vector<vector<int> > tempConns ;
	
	int cornerLim = numberOfCorners;
	for (j = 1; j < DIMENSION; ++j) cornerLim *= numberOfCorners;
	
	for(setOfCorners=0;setOfCorners<cornerLim;setOfCorners++)	 
		// a potential face is a set of DIMENSION corners, which we code with a number in base numberOfCorners.
	{
		temp = setOfCorners;
		different = true;
		for(k=0;k<DIMENSION;k++)  // calculate which corners correspond to the code setOfCorners.
		{
			i[k] = temp%numberOfCorners;
			temp /= numberOfCorners;
			for(l=0;l<k;l++) different = different && (i[l] < i[k]); 
			// we consider indices of corners in increasing order. 
			for(l=0;l<DIMENSION;l++) matrix[k][l] = corners[i[k]][l];
			//matrix is the matrix containing the vectors of the corners forming the potential face.
		}
		
		if (!different) continue; // these corners do not define a plane of interest.
		normOfNormal = 0.;
		for(k=0;k<DIMENSION;k++)
		{								 // calculate a vector normal to the potential face.
			for(l=1;l<DIMENSION;l++)
			{
				norm = 0.;
				for(j=0;j<k;j++) 
				{  
					tempmatrix[l-1][j] = matrix[l][j]-matrix[0][j];
					norm += tempmatrix[l-1][j]*tempmatrix[l-1][j];
				}
				norm += (matrix[l][k]-matrix[0][k])*(matrix[l][k]-matrix[0][k]);
				for(j=k+1;j<DIMENSION;j++)
				{
					tempmatrix[l-1][j-1] = matrix[l][j]-matrix[0][j];
					norm += tempmatrix[l-1][j-1]*tempmatrix[l-1][j-1];
				}
				norm = sqrt(norm);
				for(j=0;j<DIMENSION-1;j++) tempmatrix[l-1][j] /= norm; 
			}
			normal[k] = Determinant(tempmatrix,DIMENSION-1) * (k%2 == 0? 1.0 : -1.0);
			normOfNormal += normal[k]*normal[k];
		}
		normOfNormal = sqrt(normOfNormal);
		if (normOfNormal < 1.0e-6) continue ; // means that the corners are on the same line.
		for(k=0;k<DIMENSION;k++) normal[k] /= normOfNormal;  // normalize the normal vector.
		temp = -1; // look for a corner out of the potential face, in order to orient it.
		do
		{
			temp++;
			if(temp == i[0]) temp++;
			dotProduct = XdotAminusB(normal,corners[temp],matrix[0]);
			
		} while ((fabs(dotProduct)<1.0e-6) && (temp<numberOfCorners)  );
		if (temp == numberOfCorners)
		{
//			std::cerr << "no cluster can be defined here" << std::endl;
//			exit(-1); // if this happens, then no cluster can be defined!
			return;
		}
		if (dotProduct > 0.) 
			   for(k=0;k<DIMENSION;k++) normal[k] = -normal[k];  // the vector points outside.
			   newFace = true;
			   for(l=0;l<numberOfCorners;l++)
			   {
				   dotProduct = XdotAminusB(normal,corners[l],matrix[0]);
				   if(dotProduct>1.0e-6)	  // check that all points are on the same side (dot prod is <0)
				   {
					   newFace = false;
					   continue;
				   }
			   }
			   if (newFace)
			   {		//we have a face, but must check it's a new face.
				   for(l=0;(l<tempFaces.size()) && newFace ;l++)
				   {
					   different = false;
					   for(k=0;k<DIMENSION;k++)
					   {
						   dotProduct = XdotAminusB(tempFaces[l]->normal,matrix[k],tempFaces[l]->point);
						   different = different || (fabs(dotProduct)>1.0e-6);
					   }
					   newFace = newFace && different;
				   }
				   if (newFace)
				   {
                       vector<int> conn ;
					   face* tempface = new face;
					   for(k=0;k<DIMENSION;k++) {
						   tempface->normal[k] = normal[k] ;
						   tempface->point[k] = corners[i[0]][k] ;
						   conn.push_back(i[k]) ;
					   }
                       tempConns.push_back(conn) ;
					   tempFaces.push_back(tempface) ;
				   }
			   }
	}
	connectivities.SetShape(tempFaces.size()) ;
	faces.resize(tempFaces.size()) ;
	for(j=0;j<tempFaces.size();j++)
	{
		faces[j] = new face ;
		for(k=0;k<DIMENSION;k++) {
			connectivities[j][k] = tempConns[j][k];
			faces[j]->normal[k] = tempFaces[j]->normal[k];
			faces[j]->point[k] = tempFaces[j]->point[k];
		}
		delete tempFaces[j];
	}
}

double PolyClusterInitializer::XdotAminusB(double x[DIMENSION], double a[DIMENSION], double b[DIMENSION])
{ 
	double norm=0., dot = 0., tmp;
	for(int coord=0;coord<DIMENSION;coord++) 
	{
		tmp = a[coord]-b[coord];
		dot += x[coord]*tmp;
		norm += tmp*tmp;
	}
	if (norm != 0) dot = dot/sqrt(norm);
	return dot;
}

int PolyClusterInitializer::SetMaxSize()
{
	double tmpvect[DIMENSION], maxi[DIMENSION], mini[DIMENSION];
	int maxVolume = 1;
	for(int j=0;j<corners.Size();j++)
	{
		bravais->LatticeCoordinatesFromCartesianCoordinates(tmpvect,corners[j]);
		for(int coord=0;coord<DIMENSION;coord++)
		{
			if((j==0) || (tmpvect[coord] > maxi[coord])) maxi[coord] = tmpvect[coord];
			if((j==0) || (tmpvect[coord] < mini[coord])) mini[coord] = tmpvect[coord];
		}
	}
	for(int coord=0;coord<DIMENSION;coord++)
	{
		maxVolume *= int(maxi[coord]-mini[coord]) + 2;	// add one site each side
		maxSize[coord] = (int)maxi[coord] + 1;
		minSize[coord] = (int)mini[coord] - 1;
	}
	return int(maxVolume*bravais->GetNAtomsPerCell()*bravais->GetNAtomsPerCell());
}

bool PolyClusterInitializer::StrictlyInside(double X[DIMENSION])
{
	//! Has to use InsideRayCrossings to check the strictly inside points
	for (int i=0; i<DIMENSION;++i) queryPoint[i] = X[i] ;
	int status = 0 ;
	status = InsideRayCrossings() ;
	if (status == 1) return (true) ;
	return (false) ;
}

bool PolyClusterInitializer::Inside(double X[DIMENSION])
{
	for (int i=0; i<DIMENSION;++i) queryPoint[i] = X[i] ;
	if(connectivities.Size() == 0) {
		throw DMException("PolyClusterInitializer", "Inside", "No external faces for the region") ;
		return false;
	}
    if (givenConnectivities) {
		int status = InsideRayCrossings() ;
	    if (status >= 1) return true ;
		else return false ;
	}
	else
		return InsideConvex() ;
}

//! Use ray crossings to determine whether the given point is inside the box
int PolyClusterInitializer::InsideRayCrossings()
{
	int status ;
	
	if (DIMENSION == 2) 
        status = InPolygon() ;
	else
		status = InPolyhedron() ;
    return status ;
}

bool PolyClusterInitializer::InsideConvex()
{
	bool in=true;
	double dotproduct;	// the dot product between the normal to a face and the vector binding a point
	// of the face to the potential point. Its sign tells on which side of the face the potential point lies.
	for(unsigned int nface=0;nface<faces.size();nface++)
	{
		dotproduct = XdotAminusB(faces[nface]->normal,queryPoint,faces[nface]->point);
		in = in && (dotproduct <= 1.0e-5);
	}	
	return in;
}

double PolyClusterInitializer::Determinant(double M[DIMENSION][DIMENSION], int dim)
{
	if(dim==1)
	{
		return M[0][0];
	}
	else if(dim==2)
	{
		return M[0][0]*M[1][1] - M[1][0]*M[0][1];
	}
	else 
	{
		double temp = 0., matrixtemp[DIMENSION][DIMENSION]; 
		int i, j, k;
		{
			for(i=0; i<dim; i++)
			{
				for(j=1;j<dim;j++)
				{
					for(k=0;k<i;k++) matrixtemp[j-1][k] = M[j][k];
					for(k=i+1;k<dim;k++) matrixtemp[j-1][k-1] = M[j][k];
				}
				temp += M[0][i]*Determinant(matrixtemp,dim-1)* (i%2 == 0 ? 1. : -1.); 
			}
			return temp;
		}
	}
}

/* the rest of routines for RayCrossing are given by O'Rourke 
   with slight modification 
   XXX This is a mess. Someone someday get to clean this

This code is described in "Computational Geometry in C" (Second Edition),
Chapter 7.  It is not written to be comprehensible without the 
explanation in that book.

Written by Joseph O'Rourke, contributions by Min Xu, June 1997.
Questions to orourke@cs.smith.edu.
--------------------------------------------------------------------
This code is Copyright 1998 by Joseph O'Rourke.  It may be freely 
redistributed in its entirety provided that this copyright notice is 
not removed.
--------------------------------------------------------------------
*/

int PolyClusterInitializer::InPolygon() 
{
  int	 d;          /* dimension index */
  double x;          /* x intersection of e with ray */
  int	 Rcross = 0; /* number of right edge/ray crossings */
  int    Lcross = 0; /* number of left edge/ray crossings */
  double TOL = 1.0E-6 ;

  
  /* Shift so that q is the origin. Note this destroys the polygon.
     This is done for pedogical clarity. */
  DMArray<double, DIMENSION> copyCorners(corners.Size()) ;

  int i, i1 ;
  int n = copyCorners.Size() ;
  for( i = 0; i < n; i++ ) {
    for( d = 0; d < DIMENSION; d++ )
      copyCorners[i][d] = corners[i][d] - queryPoint[d];
  }
  
  /* For each edge e=(i-1,i), see if crosses ray. */
  n = connectivities.Size();
  for( int iedge = 0; iedge < n; iedge++ ) {
    /* First see if queryPoint=(0.0,0.0) is a vertex. */
      i  = connectivities[iedge][0] ;
	  i1 = connectivities[iedge][1] ;
	  if ( fabs(copyCorners[i][0]-0.0)<TOL && fabs(copyCorners[i][1]-0.0)<TOL ) {
		  return 2;
	  }
 
      if ((copyCorners[i][1] > 0) && (copyCorners[i1][1] <= 0) 
		  || (copyCorners[i][1] <= 0) && (copyCorners[i1][1] > 0)) {
      
		/* e straddles ray, so compute intersection with ray. */
		x = (copyCorners[i][0] * copyCorners[i1][1] - 
			copyCorners[i1][0] * copyCorners[i][1])
			/ (copyCorners[i1][1] - copyCorners[i][1]);
      
      /* crosses ray if strictly positive intersection. */
		if (x > 0) Rcross++;
	  }
    
      if ((copyCorners[i][1] > 0) && (copyCorners[i1][1] <= 0) 
		  || (copyCorners[i][1] <= 0) && (copyCorners[i1][1] > 0)) {
      
      /* e straddles ray, so compute intersection with ray. */
		x = (copyCorners[i][0] * copyCorners[i1][1] - 
			copyCorners[i1][0] * copyCorners[i][1])
			/ (copyCorners[i1][1] - copyCorners[i][1]);

      /* crosses ray if strictly positive intersection. */
		if (x < 0) Lcross++;
	  }
  }	
  
  /* q on the edge if left and right cross are not the same parity. */
  if( ( Rcross % 2 ) != (Lcross % 2 ) )
    return 3;
  
  /* q inside iff an odd number of crossings. */
  if( (Rcross % 2) == 1 )
    return 1;
  else	return 0;
}

bool PolyClusterInitializer::InBox() 
{
    int i, j ;
    for (i=0; i<DIMENSION; ++i) bmin[i]=bmax[i]=corners[0][i] ;

	for( i = 1; i < corners.Size(); i++ ) {
		for( j = 0; j < DIMENSION; j++ ) {
			if( corners[i][j] < bmin[j] )
				bmin[j] = corners[i][j];
			if( corners[i][j] > bmax[j] ) 
				bmax[j] = corners[i][j];
		}
	}


	if (DIMENSION == 2) {
		if( ( bmin[0] <= queryPoint[0] ) && ( queryPoint[0] <= bmax[0] ) &&
		( bmin[1] <= queryPoint[1] ) && ( queryPoint[1] <= bmax[1] )  )
		return true;		
	}
	else {
		if( ( bmin[0] <= queryPoint[0] ) && ( queryPoint[0] <= bmax[0] ) &&
		( bmin[1] <= queryPoint[1] ) && ( queryPoint[1] <= bmax[1] )  &&
		( bmin[2] <= queryPoint[2] ) && ( queryPoint[2] <= bmax[2] ))
		return true;		
	}
	
	return false;
}

int PolyClusterInitializer::InPolyhedron()
{
	double r[DIMENSION];  /* Ray endpoint. */
    int k = 0, crossings = 0;
    char code = '?';
   /* If query point is outside bounding box, finished. */
   if ( !InBox() )
      return 0; // Exterior

  
   LOOP:
   while( k++ < corners.Size() ) {
      crossings = 0;
  
      RandomRay( r, this->ComputeBoxRadius() ); 
      int i ;
	  for (i=0; i<DIMENSION; ++i) r[i] += queryPoint[i] ;
  
      for ( i = 0; i < connectivities.Size(); i++ ) {  /* Begin check each face */
         if ( BoxTest( i, r ) == '0' ) {
              code = '0';
         }
         else code = SegTri( i, r );

         /* If ray is degenerate, then goto outer while to generate another. */
         if ( code == 'p' || code == 'v' || code == 'e' ) {
            goto LOOP;
         }
   
         /* If ray hits face at interior point, increment crossings. */
         else if ( code == 'f' ) {
            crossings++;
         }

         /* If query endpoint q sits on a V/E/F, return that code. */
         else if ( code == 'V' )
            return( 2 );
         else if ( code == 'E' )
            return( 3 );
         else if ( code == 'F' )
            return( 4 );

         /* If ray misses triangle, do nothing. */
         else if ( code == '0' )
            ;

         else 
            cerr << "Error in InPolyhedron" << endl;

      } /* End check each face */

      /* No degeneracies encountered: ray is generic, so finished. */
      break;

   } /* End while loop */
 
   /* q strictly interior to polyhedron iff an odd number of crossings. */
   if( ( crossings % 2 ) == 1 )
      return   1;
   else return 0;
}

double PolyClusterInitializer::ComputeBoxRadius()
{
	  return (sqrt( pow( (bmax[0] - bmin[0]), 2.0 ) +
                    pow( (bmax[1] - bmin[1]), 2.0 ) +
                    pow( (bmax[2] - bmin[2]), 2.0 ) ) );
}

void PolyClusterInitializer::RandomRay( double ray[], double radius )
{
  double x, y, z, w, t;

  /* Generate a random point on a sphere of radius 1. */
  /* the sphere is sliced at z, and a random point at angle t
     generated on the circle of intersection. */
  z = 2.0 * (double) rand() / MAX_INT - 1.0;
  t = 2.0 * 3.14159 * (double) rand() / MAX_INT;
  w = sqrt( 1 - z*z );
  x = w * cos( t );
  y = w * sin( t );
  ray[0] = ( radius * x );
  ray[1] = ( radius * y );
  ray[2] = ( radius * z );
  
}

void PolyClusterInitializer::ComputeTriangleBoxes() 
{
	Boxes.resize(connectivities.Size()) ;
	int i, j, k;
	double w ;
	for ( i = 0; i < connectivities.Size(); i++ ) {
		for ( j=0; j < DIMENSION; j++ ) {
		   Boxes[i]->minMax[0][j] = corners[ connectivities[i][0] ][j];
		   Boxes[i]->minMax[1][j] = corners[ connectivities[i][0] ][j];
		}
		/* Check k=1,2 vertices of face. */
		for ( k=1; k < 3; k++ )
			for ( j=0; j < DIMENSION; j++ ) {
			   w = corners[ connectivities[i][k] ][j];
			   if ( w < Boxes[i]->minMax[0][j] ) 
				   Boxes[i]->minMax[0][j] = w;
			   if ( w > Boxes[i]->minMax[1][j] ) 
				   Boxes[i]->minMax[1][j] = w;
			}
	}
} 
  
/*
  This function returns a char:
    '0': the segment [ab] does not intersect (completely misses) the 
         bounding box surrounding the n-th triangle T.  It lies
         strictly to one side of one of the six supporting planes.
    '?': status unknown: the segment may or may not intersect T.
*/

char PolyClusterInitializer::BoxTest( int n, double b[] )
{
   int i; /* Coordinate index */
   double w;

   for ( i=0; i < DIMENSION; i++ ) {
       w = Boxes[n]->minMax[0][i]; /* min: lower left */
       if ( (queryPoint[i] < w) && (b[i] < w) ) return '0';
       w = Boxes[n]->minMax[1][i]; /* max: upper right */
       if ( (queryPoint[i] > w) && (b[i] > w) ) return '0';
   }
   return '?';
}

char PolyClusterInitializer::SegTri( int faceNum, double ray[])
{
    int code = '?';
    int m = -1;
    double interSect[DIMENSION] ;
    int Tri[3] ;
	int i ;
	for (i=0; i<3; ++i)
		Tri[i] = connectivities[faceNum][i] ;
	
	code = SegPlane( Tri, ray, interSect, &m );

    if      ( code == '0')
       return '0';
    else if ( code == 'q')
       return InTri3D( Tri, m, queryPoint );
    else if ( code == 'r')
       return InTri3D( Tri, m, ray );
    else if ( code == 'p' )
       return 'p';
    else if ( code == '1' )
       return SegTriCross( Tri, ray );
    else /* Error */ {
       cerr << "Error" << endl ;
	   return code;
	}
}

/* Assumption: p lies in the plane containing T.
    Returns a char:
     'V': the query point p coincides with a Vertex of triangle T.
     'E': the query point p is in the relative interior of an Edge of triangle T.
     'F': the query point p is in the relative interior of a Face of triangle T.
     '0': the query point p does not intersect (misses) triangle T.
*/

char PolyClusterInitializer::InTri3D( int Tri[], int m, double interSect[] )
{
   int i;           /* Index for X,Y,Z           */
   int j;           /* Index for X,Y             */
   int k;           /* Index for triangle vertex */
   double pp[DIMENSION];      /* projected p */
   double Tp[3][DIMENSION];   /* projected T: three new vertices */

   /* Project out coordinate m in both p and the triangular face */
   j = 0;
   for ( i = 0; i < DIMENSION; i++ ) {
     if ( i != m ) {    /* skip largest coordinate */
       pp[j] = interSect[i];
       for ( k = 0; k < 3; k++ )
			Tp[k][j] = corners[Tri[k]][i];
			j++;
     }
   }
   return( InTri2D( Tp, pp ) );
}

char PolyClusterInitializer::InTri2D( double Tp[][DIMENSION], double pp[] )
{
   int area0, area1, area2;

   /* compute three AreaSign() values for pp w.r.t. each edge of the face in 2D */
   area0 = AreaSign( pp, Tp[0], Tp[1] );
   area1 = AreaSign( pp, Tp[1], Tp[2] );
   area2 = AreaSign( pp, Tp[2], Tp[0] );

   if ( ( area0 == 0 ) && ( area1 > 0 ) && ( area2 > 0 ) ||
        ( area1 == 0 ) && ( area0 > 0 ) && ( area2 > 0 ) ||
        ( area2 == 0 ) && ( area0 > 0 ) && ( area1 > 0 ) ) 
     return 'E';

   if ( ( area0 == 0 ) && ( area1 < 0 ) && ( area2 < 0 ) ||
        ( area1 == 0 ) && ( area0 < 0 ) && ( area2 < 0 ) ||
        ( area2 == 0 ) && ( area0 < 0 ) && ( area1 < 0 ) )
     return 'E';                 
   
   if ( ( area0 >  0 ) && ( area1 > 0 ) && ( area2 > 0 ) ||
        ( area0 <  0 ) && ( area1 < 0 ) && ( area2 < 0 ) )
     return 'F';

   if ( ( area0 == 0 ) && ( area1 == 0 ) && ( area2 == 0 ) )
     cerr << "Error in InTriD" << endl;

   if ( ( area0 == 0 ) && ( area1 == 0 ) ||
        ( area0 == 0 ) && ( area2 == 0 ) ||
        ( area1 == 0 ) && ( area2 == 0 ) )
     return 'V';

   else  
     return '0';  
}

int PolyClusterInitializer::AreaSign( double a[], double b[], double c[] )  
{
    double area2;

    area2 = ( b[0] - a[0] ) * (double)( c[1] - a[1] ) -
            ( c[0] - a[0] ) * (double)( b[1] - a[1] );

    /* The area should be an integer. */
    if      ( area2 >  0.5 ) return  1;
    else if ( area2 < -0.5 ) return -1;
    else                     return  0;
} 

/*---------------------------------------------------------------------
    'p': The segment lies wholly within the plane.
    'q': The q endpoint is on the plane (but not 'p').
    'r': The r endpoint is on the plane (but not 'p').
    '0': The segment lies strictly to one side or the other of the plane.
    '1': The segement intersects the plane, and 'p' does not hold.
---------------------------------------------------------------------*/
char	PolyClusterInitializer::SegPlane( int Tri[], double ray[], double interSect[], int *m)
{
    double N[DIMENSION]; 
	double D;
    double rq[DIMENSION];
    double num, denom, t;
	int i;

    *m = PlaneCoeff( Tri, N, &D );
    num = D - Dot( queryPoint, N );
    SubVec( ray, queryPoint, rq );
    denom = Dot( rq, N );

    if ( denom == 0.0 ) {  /* Segment is parallel to plane. */
       if ( num == 0.0 )   /* q is on plane. */
           return 'p';
       else
           return '0';
    }
    else
       t = num / denom;
    /*printf("SegPlaneInt: t=%lf \n", t );*/

    for( i = 0; i < DIMENSION; i++ )
       interSect[i] = queryPoint[i] + t * ( ray[i] - queryPoint[i] );

    if ( (0.0 < t) && (t < 1.0) )
         return '1';
    else if ( num == 0.0 )   /* t == 0 */
         return 'q';
    else if ( num == denom ) /* t == 1 */
         return 'r';
    else return '0';
}
                           
/*---------------------------------------------------------------------
Computes N & D and returns index m of largest component.
---------------------------------------------------------------------*/
int	PolyClusterInitializer::PlaneCoeff( int Tri[], double N[], double *D )
{
    int i;
    double t;              /* Temp storage */
    double biggest = 0.0;  /* Largest component of normal vector. */
    int m = 0;             /* Index of largest component. */

    NormalVec( corners[Tri[0]], corners[Tri[1]], corners[Tri[2]], N );
    *D = Dot( corners[Tri[0]], N );

    /* Find the largest component of N. */
    for ( i = 0; i < DIMENSION; i++ ) {
      t = fabs( N[i] );
      if ( t > biggest ) {
        biggest = t;
        m = i;
      }
    }
    return m;
}

/*---------------------------------------------------------------------
Compute the cross product of (b-a)x(c-a) and place into N.
---------------------------------------------------------------------*/
void	PolyClusterInitializer::NormalVec( double a[], double b[], 
										  double c[], double N[] )
{
    N[0] = ( c[2] - a[2] ) * ( b[1] - a[1] ) -
           ( b[2] - a[2] ) * ( c[1] - a[1] );
    N[1] = ( b[2] - a[2] ) * ( c[0] - a[0] ) -
           ( b[0] - a[0] ) * ( c[2] - a[2] );
    N[2] = ( b[0] - a[0] ) * ( c[1] - a[1] ) -
           ( b[1] - a[1] ) * ( c[0] - a[0] );
}

/*---------------------------------------------------------------------
Returns the dot product of the two input vectors.
---------------------------------------------------------------------*/
double	PolyClusterInitializer::Dot( double a[], double b[] )
{
    int i;
    double sum = 0.0;

    for( i = 0; i < DIMENSION; i++ )
       sum += a[i] * b[i];

    return  sum;
}
/*---------------------------------------------------------------------
a - b ==> c.
---------------------------------------------------------------------*/
void    PolyClusterInitializer::SubVec( double a[], double b[], double c[] )
{
   int i;

   for( i = 0; i < DIMENSION; i++ )
      c[i] = a[i] - b[i];
}

/*---------------------------------------------------------------------
The signed volumes of three tetrahedra are computed, determined
by the segment qr, and each edge of the triangle.  
Returns a char:
   'v': the open segment includes a vertex of T.
   'e': the open segment includes a point in the relative interior of an edge
   of T.
   'f': the open segment includes a point in the relative interior of a face
   of T.
   '0': the open segment does not intersect triangle T.
---------------------------------------------------------------------*/

char PolyClusterInitializer::SegTriCross( int Tri[], double ray[] )
{
   int vol0, vol1, vol2;
   
   vol0 = VolumeSign( queryPoint, corners[ Tri[0] ], corners[ Tri[1] ], ray ); 
   vol1 = VolumeSign( queryPoint, corners[ Tri[1] ], corners[ Tri[2] ], ray ); 
   vol2 = VolumeSign( queryPoint, corners[ Tri[2] ], corners[ Tri[0] ], ray );
 
     
   /* Same sign: segment intersects interior of triangle. */
   if ( ( ( vol0 > 0 ) && ( vol1 > 0 ) && ( vol2 > 0 ) ) || 
        ( ( vol0 < 0 ) && ( vol1 < 0 ) && ( vol2 < 0 ) ) )
      return 'f';
   
   /* Opposite sign: no intersection between segment and triangle */
   if ( ( ( vol0 > 0 ) || ( vol1 > 0 ) || ( vol2 > 0 ) ) &&
        ( ( vol0 < 0 ) || ( vol1 < 0 ) || ( vol2 < 0 ) ) )
      return '0';

   else if ( ( vol0 == 0 ) && ( vol1 == 0 ) && ( vol2 == 0 ) )
     cerr << "Error 1 in SegTriCross" << endl;
   
   /* Two zeros: segment intersects vertex. */
   else if ( ( ( vol0 == 0 ) && ( vol1 == 0 ) ) || 
             ( ( vol0 == 0 ) && ( vol2 == 0 ) ) || 
             ( ( vol1 == 0 ) && ( vol2 == 0 ) ) )
      return 'v';

   /* One zero: segment intersects edge. */
   else if ( ( vol0 == 0 ) || ( vol1 == 0 ) || ( vol2 == 0 ) )
      return 'e';
   
   else {
     cerr << "Error 1 in SegTriCross" << endl;
   }
return '0' ;
}

int 	PolyClusterInitializer::VolumeSign( double a[], double b[], double c[], double d[] )
{ 
   double vol;
   double ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz;
   double bxdx, bydy, bzdz, cxdx, cydy, czdz;

   ax = a[0];
   ay = a[1];
   az = a[2];
   bx = b[0];
   by = b[1];
   bz = b[2];
   cx = c[0]; 
   cy = c[1];
   cz = c[2];
   dx = d[0];
   dy = d[1];
   dz = d[2];

   bxdx=bx-dx;
   bydy=by-dy;
   bzdz=bz-dz;
   cxdx=cx-dx;
   cydy=cy-dy;
   czdz=cz-dz;
   vol =   (az-dz) * (bxdx*cydy - bydy*cxdx)
         + (ay-dy) * (bzdz*cxdx - bxdx*czdz)
         + (ax-dx) * (bydy*czdz - bzdz*cydy);


   /* The volume should be an integer. */
   if      ( vol > 0.5 )   return  1;
   else if ( vol < -0.5 )  return -1;
   else                    return  0;
}


