// CompositeLangevinAtomsMover.cpp: implementation of the CompositeLangevinAtomsMover class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"

#include "blas.h"

#include "DigitalMaterial.h"
#include "CompositeLangevinAtomsMover.h"
#include "DMException.h"
#include "ListOfAtoms.h"
#include "LeafToLeafIterator.h"
#include "DMException.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"

using namespace std;


static SerialBuilder<CompositeLangevinAtomsMover> *_CompLangevincreator = 
    new SerialBuilder<CompositeLangevinAtomsMover>("CompositeLangevinAtomsMover");

const std:: string CompositeLangevinAtomsMover::className = string("CompositeLangevinAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CompositeLangevinAtomsMover::CompositeLangevinAtomsMover(Potential *potential,double dt,int n):VerletAtomsMover(potential,dt,n)
{  
  randomPosInc = NULL;
  randomVelInc = NULL;

  parametersOutOfDate = true;
  numberOfLeaves = 0;

  for(int coord=0;coord<DIMENSION;++coord)
	  defaultFriction[coord] = 1.0;
}

CompositeLangevinAtomsMover::~CompositeLangevinAtomsMover()
{
	if(randomPosInc) delete randomPosInc;
	if(randomVelInc) delete randomVelInc;
}

/*!
  As well as allocating the main arrays---forces, position and velocity, Allocate also allocates the arrays of parameters.

 */

void CompositeLangevinAtomsMover::Allocate(ListOfAtoms *atoms)
{
	VerletAtomsMover::Allocate(atoms->GetCartesianPositions()); // allocates forces

	if(randomPosInc)
		delete randomPosInc;
	// ajd27 shapeChange
	randomPosInc = new DMArray<double, DIMENSION>(atoms->GetCartesianPositions());
	
	if(randomVelInc)
		delete randomVelInc;
	// ajd27 shapeChange
	randomVelInc = new DMArray<double, DIMENSION>(atoms->GetCartesianPositions());
   
	numberOfLeaves = atoms->NumberOfLeaves();

	// clear the lists
	 friction.clear();
	 mass.clear();
	 c0.clear();
	 c1.clear();
	 c2.clear();
	 c1_m_c2.clear();
	 stdDevPos.clear();
	 stdDevVel.clear();
	 pvCorrelation.clear();
	 corrConstant.clear();
	 
	// iterate over the leaves of the LOA, setting the frictions and masses, making space for the other parameters
	 
	 for(LeafToLeafIterator<ListOfAtoms> iter(atoms);!iter.Done(); iter.Next())
	 {
		 std::string typeName = iter.Current()->GetTypeName();
		 std::map<std::string,TinyVector<double> >::const_iterator leafFriction = frictionMap.find(typeName);
		 const TinyVector<double> *nextFriction;

		 if(leafFriction == frictionMap.end())
			 nextFriction = &defaultFriction;
		 else nextFriction = &(leafFriction->second);
		 
		 friction.push_back(*nextFriction);
		 mass.push_back(iter.Current()->GetMass());
		 
		 c0.resize(numberOfLeaves);
		 c1.resize(numberOfLeaves);
		 c2.resize(numberOfLeaves);
		 c1_m_c2.resize(numberOfLeaves);
		 stdDevPos.resize(numberOfLeaves);
		 stdDevVel.resize(numberOfLeaves);
		 pvCorrelation.resize(numberOfLeaves);
		 corrConstant.resize(numberOfLeaves);
		 
	 }
	 
	 CalculateParameters();
}


void CompositeLangevinAtomsMover::AddFriction(std::string typeName, double friction[DIMENSION])
{
	// pass in a type name and friction value which are added to an STL map
	frictionMap.insert(make_pair(typeName,TinyVector<double>(friction)));

}


const double *CompositeLangevinAtomsMover::GetFriction(std::string typeName) const
{
	std::map<std::string,TinyVector<double> >::const_iterator friction = frictionMap.find(typeName);
	if(friction == frictionMap.end())
		throw DMException("CompositeLangevinAtomsMover","GetFriction","Couldn't find typeName");
	return friction->second;
}


void CompositeLangevinAtomsMover::Move(ListOfAtoms *atoms)
{
	if (!(forces && (forces->GetShape() == atoms->GetCartesianPositions()->GetShape())))
		Allocate(atoms); // includes call to CalculateParameters
	else if(parametersOutOfDate) // eg if only the temperature or time step has changed since the last time move was called
		CalculateParameters();
	
	
	// Calculate forces
	potential->CalculateForces(atoms,forces);
	atoms->AdjustForces(forces);
	double temp[DIMENSION];
	int coord;
	
	for(int s=0;s<nSteps;s++)
	{
		// for second and subsequent times through the loop, am relying on forces as calculated for previous time step
		
		GenerateRandomNumbers();
		atoms->AdjustForceIncrements(randomPosInc); // because these are like forces, except we can't add the external force to these!
		atoms->AdjustForceIncrements(randomVelInc);
		
		
		LeafToLeafIterator<ListOfAtoms> iter(atoms);
		LeafToLeafIterator<DMArray<double,DIMENSION> > forceIter(forces);
		LeafToLeafIterator<DMArray<double,DIMENSION> > randPosIncIter(randomPosInc);
		
		int leaf;
		for(leaf=0 ; !iter.Done(); iter.Next(),forceIter.Next(),randPosIncIter.Next(),++leaf)
		{
			const DMArray<double, DIMENSION> *leafVelocities = iter.Current()->GetCartesianVelocities();
			const DMArray<double, DIMENSION> *leafPositions  = iter.Current()->GetCartesianPositions();
			const DMArray<double, DIMENSION> *leafForces = forceIter.Current();
			const DMArray<double, DIMENSION> *leafRandPosInc = randPosIncIter.Current();
			
			double mass = iter.Current()->GetMass();
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = c1[leaf][coord]*dt;
			iter.Current()->IncrementCartesianPositions(leafVelocities,temp);
			
			for(coord=0;coord<DIMENSION;++coord)
			  temp[coord] = c2[leaf][coord]*dt*dt/mass;
			iter.Current()->IncrementCartesianPositions(leafForces,temp);
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = stdDevPos[leaf][coord];
			iter.Current()->IncrementCartesianPositions(leafRandPosInc, temp);
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = c0[leaf][coord];
			iter.Current()->ScaleCartesianVelocities(temp);
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = c1_m_c2[leaf][coord]*dt/mass;
			iter.Current()->IncrementCartesianVelocities(leafForces,temp);
		}
		
		// Calculate forces
		
		potential->CalculateForces(atoms,forces);
		atoms->AdjustForces(forces);
		
		LeafToLeafIterator<ListOfAtoms> iter2(atoms);
		LeafToLeafIterator<DMArray<double,DIMENSION> > forceIter2(forces);
		LeafToLeafIterator<DMArray<double,DIMENSION> > randVelIncIter2(randomVelInc);
		LeafToLeafIterator<DMArray<double,DIMENSION> > randPosIncIter2(randomPosInc);
		
		for (leaf=0; !iter2.Done(); iter2.Next(),forceIter2.Next(),randPosIncIter2.Next(),randVelIncIter2.Next(),++leaf)	 
		{
			const DMArray<double, DIMENSION> *leafVelocities = iter2.Current()->GetCartesianVelocities();
			const DMArray<double, DIMENSION> *leafPositions  = iter2.Current()->GetCartesianPositions();
			const DMArray<double, DIMENSION> *leafForces = forceIter2.Current();
			const DMArray<double, DIMENSION> *leafRandVelInc = randVelIncIter2.Current();
			const DMArray<double, DIMENSION> *leafRandPosInc = randPosIncIter2.Current();
			double mass = iter2.Current()->GetMass();
			
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = c2[leaf][coord]*dt/mass;
			iter2.Current()->IncrementCartesianVelocities(leafForces,temp);
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = stdDevVel[leaf][coord]*pvCorrelation[leaf][coord];
			iter2.Current()->IncrementCartesianVelocities(leafRandPosInc,temp);
			
			for(coord=0;coord<DIMENSION;++coord)
				temp[coord] = stdDevVel[leaf][coord]*corrConstant[leaf][coord];
			iter2.Current()->IncrementCartesianVelocities(leafRandVelInc,temp);
		}
		
		atoms->AdjustVelocities(); 
	}
  
  
}


void CompositeLangevinAtomsMover::GenerateRandomNumbers()
{
	DMArrayIterator<double, DIMENSION> randPosIncIter(randomPosInc), randVelIncIter(randomVelInc);
	double *currentPos, *currentVel;
	for(randPosIncIter.Init(), randVelIncIter.Init(); !randPosIncIter.IsDone();
		randPosIncIter.Next(), randVelIncIter.Next())
	{
		currentPos = randPosIncIter.Current();
		currentVel = randVelIncIter.Current();
		for(int coord=0;coord<DIMENSION;++coord)
		{
			double r1 = 1.0 - rand.uniform();
			// to avoid r1 being zero (don't want to take the log of zero!)
			double r2 = 2*PI*rand.uniform();
			double t = sqrt(-2.0 * log(r1));
			currentPos[coord] = t*cos(r2);
			currentVel[coord] = t*sin(r2);
		}
	}
}



void CompositeLangevinAtomsMover::CalculateParameters()
{
/*

	std::vector<double> zero;
	int coord;
	for(coord=0;coord<DIMENSION;++coord) zero.push_back(0.);
	
	if(i==c1.size())
    {
		c0.push_back(zero);
		c1.push_back(zero);
		c2.push_back(zero);
		c1_m_c2.push_back(zero);
		stdDevPos.push_back(zero);
		stdDevVel.push_back(zero);
		pvCorrelation.push_back(zero);
		corrConstant.push_back(zero);
    }
*/	
	for(int i=0;i<numberOfLeaves;++i)
		for(int coord=0;coord<DIMENSION;++coord)
		{
			double fricDt = friction[i][coord] * dt;
			c0[i][coord] = exp(-fricDt);
			if(friction[i][coord] == 0)
			{
				c1[i][coord] = 1.;
				c2[i][coord] = 0.5;
				stdDevPos[i][coord] = 0.0;
				stdDevVel[i][coord] = 0.0;
			}
			else
			{
				c1[i][coord] = (1-c0[i][coord])/fricDt;
				c2[i][coord] = (1-c1[i][coord])/fricDt;
				
				// need to add the mass in the following two quantities
				stdDevPos[i][coord] = dt*sqrt(temperature/mass[i] *((2.0 - (3 - 4*exp(-fricDt) + exp(-2*fricDt))/fricDt)/fricDt));
				stdDevVel[i][coord] = sqrt(temperature/mass[i] * (1 - exp(-2*fricDt)));
			}
			c1_m_c2[i][coord] = c1[i][coord] - c2[i][coord];
			pvCorrelation[i][coord] = sqrt(3.0)/2. * (1. - fricDt / 8.0);
			corrConstant[i][coord] = sqrt((1.0 - pvCorrelation[i][coord]) * (1.0 + pvCorrelation[i][coord]));
		
		}// end loop over i and coord
	
	parametersOutOfDate = false;
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void CompositeLangevinAtomsMover::Save(DMWriter *writer) const
{
	VerletAtomsMover::Save(writer);	

	writer->PutDouble(temperature, "temperature");
	writer->Put(&rand, "rand");

	// Need to save the frction and mass arrays!!!!!!
}

void CompositeLangevinAtomsMover::Load(DMReader *reader)
{
	VerletAtomsMover::Load(reader);

	SetTemperature(reader->GetDouble("temperature"));
	reader->Fill(&rand, "rand");
}
