// PredictorCorrectorAtomsMover.cpp: implementation of the PredictorCorrectorAtomsMover class.
//
//////////////////////////////////////////////////////////////////////

#include "blas.h"

#include "DigitalMaterial.h"

#include "PredictorCorrectorAtomsMover.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

static SerialBuilder<PredictorCorrectorAtomsMover> *_PCmovercreator = 
    new SerialBuilder<PredictorCorrectorAtomsMover>("PredictorCorrectorAtomsMover");

const std::string PredictorCorrectorAtomsMover::className = std::string("PredictorCorrectorAtomsMover");

//////////////////////////////////////////////////////////////////////

// Construction/Destruction

//////////////////////////////////////////////////////////////////////


PredictorCorrectorAtomsMover::PredictorCorrectorAtomsMover(Potential *potential, double dt, int nSteps):VerletAtomsMover(potential,dt,nSteps)
{
  // I don't think this is true anymore:
        if(N_VALUES != 4)
	  throw DMException("PredictorCorrectorAtomsMover",
						"PredictorCorrectorAtomsMover",
						"Only works for N_VALUES = 4 yet!");
	forces = 0; // Won't be used in this AtomsMover
	corrections = 0;
	secondDerivs = 0;
	thirdDerivs = 0;
	fourthDerivs = 0;
	fifthDerivs = 0;
	CalculateCoefficients();
}

PredictorCorrectorAtomsMover::~PredictorCorrectorAtomsMover()
{
	if (corrections) delete [] corrections;
	if(secondDerivs) delete [] secondDerivs;
	if(thirdDerivs) delete [] thirdDerivs;
	if(fourthDerivs) delete [] fourthDerivs;
	if(fifthDerivs) delete [] fifthDerivs;
}

void PredictorCorrectorAtomsMover::Move(ListOfAtoms *atoms)	// PredictorCorrector; returns time
{
  const DMArray<double, DIMENSION> *velocities = atoms->GetCartesianVelocities();
  const DMArray<double, DIMENSION> *positions  = atoms->GetCartesianPositions();
  if (!(corrections && (corrections->GetShape() == positions->GetShape())))
	  Allocate(positions);

  atoms->AdjustVelocities(); 
  
  for(int step=0;step<nSteps;step++)
    {
      // Predictor step
      atoms->IncrementCartesianPositions(velocities,s[1]);
      atoms->IncrementCartesianPositions(secondDerivs,s[2]);
      atoms->IncrementCartesianPositions(thirdDerivs,s[3]);

      atoms->IncrementCartesianVelocities(secondDerivs,s[1]);
      atoms->IncrementCartesianVelocities(thirdDerivs,s[2]);


      secondDerivs->Axpy(thirdDerivs,s[1]);


      // Usually we pass a "forces" array to the potential, but we're going to do a subtraction such that the result will be the corrections. There's no need for two arrays.

      potential->CalculateForces(atoms,corrections);
      atoms->AdjustForces(corrections);
      corrections->Axpy(secondDerivs,-1.0);

	  // Corrector step
      atoms->IncrementCartesianPositions(corrections,coeff[0]);
      atoms->IncrementCartesianVelocities(corrections,coeff[1]);
      secondDerivs->Axpy(corrections,1.0);
      thirdDerivs->Axpy(corrections,coeff[3]);
    }

   
    // For testing energy conservation
     /*cout.precision(12);
     double kineticEnergy = atoms->GetKineticEnergy();
     double potentialEnergy = potential->CalculateEnergy(atoms);
     cout << "Total Energy is " << kineticEnergy + potentialEnergy << endl;
	  */
	
  time += nSteps*dt;

}

void PredictorCorrectorAtomsMover::Allocate(const DMArray<double, DIMENSION> *array)
{
	if (corrections) delete [] corrections;
	if(secondDerivs) delete [] secondDerivs;
	if(thirdDerivs) delete [] thirdDerivs;
	if(fourthDerivs) delete [] fourthDerivs;
	if(fifthDerivs) delete [] fifthDerivs;	


	// ajd27 shapeChange many TIMES
	corrections = new DMArray<double, DIMENSION> (array);
	secondDerivs = new DMArray<double, DIMENSION> (array);
	thirdDerivs = new DMArray<double, DIMENSION> (array);
	if(N_VALUES > 4)
	  fourthDerivs = new DMArray<double, DIMENSION> (array);
	if(N_VALUES > 5)
	  fifthDerivs = new DMArray<double, DIMENSION> (array);


	// use an iterator and memset?
	double zero[DIMENSION];for(int c=0;c<DIMENSION;++c) zero[c] = 0.;
	corrections->Set(zero);
	secondDerivs->Set(zero);
	thirdDerivs->Set(zero);
	if(N_VALUES > 4)
	  secondDerivs->Set(zero);
	if(N_VALUES > 5)
	  secondDerivs->Set(zero);
	

}


void PredictorCorrectorAtomsMover::CalculateCoefficients()
	  {
	    int a;

	    s[0] = 1;
	    for(a=1;a<N_VALUES;++a) s[a] = s[a-1] * dt/a;
#if (N_VALUES == 4)
	    coeff[0] = 1.0/6.0;
	    coeff[1] = 5.0/6.0;
	    coeff[2] = 1.0;
	    coeff[3] = 1.0/3.0;
#elif (N_VALUES == 5)
	    coeff[0] = 19.0/120.0;
	    coeff[1] = 3.0/4.0;
	    coeff[2] = 1.0;
	    coeff[3] = 1.0/32.0;
	    coeff[4] = 1.0/12.0;
#elif (N_VALUES == 6)
	    coeff[0] = 3.0/20.0;
	    coeff[1] = 251.0/360.0;
	    coeff[2] = 1.0;
	    coeff[3] = 11.0/18.0;
	    coeff[4] = 1.0/6.0;
	    coeff[5] = 1.0/60.0;
#else
#error "Invalid value for N_VALUES in Predictor-Corrector"
#endif
	    for(a=0;a<N_VALUES;++a)
	      coeff[a] *= s[2]/s[a];

	  }

 


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void PredictorCorrectorAtomsMover::Save(DMWriter *writer) const
{
	VerletAtomsMover::Save(writer);	

	// Probably should save the arrays...I'll come back to this sometime (NB, 8/6/2001)
}

void PredictorCorrectorAtomsMover::Load(DMReader *reader)
{
	VerletAtomsMover::Load(reader);
}
