// QuickMinAtomsMover.h: interface for the QuickMinAtomsMover class.
//
//////////////////////////////////////////////////////////////////////

#ifndef QUICKMINATOMSMOVER_H
#define QUICKMINATOMSMOVER_H

#include "DigitalMaterial.h"
#include "AtomsMover.h"
#include "Potential.h"
#include <iostream>

//! Minimization by Molecular Dynamics method (sometimes called MDmin)

/*! 
  This version goes atom-by-atom, zeroing the velocity of an atom
  when the dot-product with its force is negative. Other variants are in
  subclasses.


*/

class QuickMinAtomsMover : public AtomsMover  
{
public:
	QuickMinAtomsMover(Potential *potential = 0,double dt = 1.,double tolerance = 1.e-6);
	~QuickMinAtomsMover() ;
	void Move(ListOfAtoms *atoms);
	double GetTolerance() const {return tolerance;}
	void SetTolerance(double t) {tolerance = t;}
	int GetMaxIterations() const {return maxIterations;}
	void SetMaxIterations(int n) {maxIterations = n;}
	double GetDt() const {return dt;}
	void SetDt(double t) {dt = t;}

	void SetOutputStream(std::ostream *stream){this->outputStream = stream;}
	void Verbose() {verbose = true;}
	void Quiet() {verbose = false;}
	void SetStepsBetweenUpdate(int steps){ stepsBetweenUpdate = steps; }
	int GetStepsBetweenUpdate(){ return stepsBetweenUpdate; }
	void SetInstabilityThreshold(double instTh){ instabilityThreshold = instTh; }
	double GetInstabilityThreshold(){ return instabilityThreshold; }
	
	// Accessing "return" information
	double GetConvergence() const {return totalForceSq;}
	int GetNIterations() const {return nIterations;}
	
	// not sure about whether this should be SWIGged or not...NB
	class UnstableError
    {
    public:
		UnstableError(double p,int n): pe_gf(p), N(n) {}
		double pe_gf;
		int N;
    };

#ifndef SWIG
    void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
    const std::string &GetType() const
    {
        return className;
    }
#endif // SWIG

protected:
	virtual void AdjustVelocities(const DMArray<double, DIMENSION> *forces, const DMArray<double, DIMENSION> *velocities);
	DMArray<double, DIMENSION> *adjustedVelocities;

private:
	void CheckQMParameter(DMReader *reader, double parameter, const std::string parameterName);

	//! if totalForceSq increases by a factor greater than this then throw an exception
    double instabilityThreshold;
	//! if verbose is set, print out status every \a stepsBetweenUpdate iterations
	int stepsBetweenUpdate;
    static const int DISPLAYPRECISION;

	double dt;
	double tolerance;
	int maxIterations;
	Potential *potential;
	int nAtoms;
	DMArray<double, DIMENSION> *forces;

	void Allocate(int nAtoms, const DMArray<double, DIMENSION> *array);
	double totalForceSq;
	int nIterations;
	bool verbose;
	std::ostream *outputStream;
	static const std::string className;
};

#endif //  QUICKMINATOMSMOVER_H
