// Constraint.h: interface for the Constraints class.
//
//////////////////////////////////////////////////////////////////////

#ifndef CONSTRAINT_H
#define CONSTRAINT_H



#include "DigitalMaterial.h"
#include <string>

#ifndef SWIG
class ListOfAtoms;
template<class Type, int N> class DMArray;

#endif // SWIG

//! Defines interface for implenting constraints on a ListOfAtoms


/*!
  The Constraint class and its derivatives will allow manipulation of 
  constrained ListOfAtomses. Examples of derived classes are
  NonRotatingRigidBody, FixedBody, UniformlyMovingBody.

  There are two ways to use a Constraint class, corresponding to two sets of
  functions defined in the interface.

  The first is by the "adjust" mechanism. ListOfAtoms functions such as
  Set/IncrementCartesianPositions, Set/IncrementCartesianVelocities, etc
  include calls to Constraint::AdjustPositions(),
  Constraint::AdjustVelocities(), etc. A Mover would call
  Constraint::AdjustForces(). All this assumes that the constraint is such that
  it can be implemented by such postprocessing. In some cases only one or two
  adjust functions are implemented, depending on the nature of the Constraint

  Ideally the Adjust functions would be "projection operators" in the sense
  that calling them a second time would have no effect. However this is not
  for ExternalForceConstraint::AdjustForces() (the external forces would get
  added twice) or TimeDependentConstraint::AdjustPositions() (the internal 
  time would be incremented twice).

  The second method is for the Constraint class to be an interface between the
  dynamically independent degrees of freedom of the system, and the actual
  atomic positions. Even in a unconstrained ListOfAtoms it is useful minimizers
  to be able to see the atomic positions as an array of doubles.

  For this purpose there exist the "GeneralizedCoordinate" functions. You can
  ask for the number of generalized coordinates, pass an array in order to set
  them, pass an array to be filled in with the values of the generalized 
  coordinates, increment a single generalized coordinate, get the derivative of
  the position of an atom with respect to a particular generalized coordinate
  and ask for an array of forces previously calculated by a potential to be 
  used to calculated the corresponding generailzed forces.

  It's not clear that all of these functions will remain. For example the
  derivative function is used in one application to compute generalized forces,
  but this should not be necessary if the constraint already has a function to
  do that.

  Caution: Most of this interface and the various Constraint classes assume 
  that Constraints are attached to leaves only. The Constraint interface
  appears in ListOfAtoms as well. For the "Adjust" functions, the corresponding
  function in the ListOfAtoms's Constraint is called and then the same function
  in each branch of the LOA (passing appropriate branches of forces or 
  whatever the array is). For the GeneralizedCoordinateFunctions, if the LOA
  is a leaf it will pass the calls on to its own Constraint, otherwise it will
  ignore its own constraint and pass the calls onto the branches, modifying
  pointer in aa ppropriate way.

  
  The base class is a non-constraint: the Adjust functions do nothing and the
  GeneralizedCoordinate functions are simply an interface to the full set of 
  atomic positions (useful for minimizers).

  A Constraint is attached to a particular ListOfAtoms and is destroyed when a
  new one is attached or when the LOA is itself destroyed. The constraint has a
  pointer to the LOA which is set when it is attached. It is private, so cannot
  be set by any other class than ListOfAtoms, which is a friend.

  \ingroup MolecularDynamics
  \author Nicholas Bailey
  \date March 1, 2001

 */

class Constraint
{
public:
	Constraint() ;
	virtual ~Constraint();

	virtual Constraint *Copy() const;

	
	// Adjust functions
	virtual void AdjustPositions(DMArray<double, DIMENSION> *positions) {}
	virtual void AdjustVelocities(DMArray<double, DIMENSION> *velocities) {}
	virtual void AdjustForces(DMArray<double, DIMENSION> *forces) {}
	virtual void AdjustForceIncrements(DMArray<double, DIMENSION> *forces) {}

	// Generalized Coordinate functions
	virtual int NumberGeneralizedCoordinates();
	virtual void SetGeneralizedCoordinates(double *generalizedCoordinates) ;
	virtual void GetGeneralizedCoordinates(double *generalizedCoordinates); 
	virtual void IncrementGeneralizedCoordinate(int n,double inc);

	// I think we need to Serialize Constraints. But I want this function now

	const std::string &GetType() const
		{
			return className; 
		}
#ifndef SWIG
	virtual void CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces);
	//! This function returns the DIMENSION components of the derivative of atom number "atom" with respect to each of the generalized coordinates
  	virtual void DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION]);
	friend class ListOfAtoms;
	
protected:
	ListOfAtoms *atoms; // only set when the constraint is added to a particular LOA
	static const std::string className;
#endif // SWIG
};


#endif // !defined(CONSTRAINT_H)
