// NonRotatingRigidBody.cpp: implementation of the NonRotatingRigidBody class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"

#include "NonRotatingRigidBody.h"
#include "DMArray.h"
#include "ListOfAtoms.h"

#include "DMLastHeader.h"

const std::string NonRotatingRigidBody::className = std::string("NonRotatingRigidBody");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

NonRotatingRigidBody::NonRotatingRigidBody(double weight): weightFactor(weight)
{
    for(int coord=0;coord<DIMENSION;++coord)
        centerOfMass[coord] = 0.;
    displacements = 0;
    centerOfMassIsDefined = false;
}

NonRotatingRigidBody::~NonRotatingRigidBody()
{
    if (displacements) delete displacements;
}

Constraint *NonRotatingRigidBody::Copy() const
{
    NonRotatingRigidBody *newConstraint = new NonRotatingRigidBody(weightFactor);
    newConstraint->SetExternalForce(externalForce);
    return newConstraint;
}

void NonRotatingRigidBody::AdjustForces(DMArray<double, DIMENSION> *forces)
{
    int coord;
    // add the external force to any atom---doesn't matter which because of the subsequent averaging.
    if(forces->Size() > 0)
    {
        for(coord=0;coord<DIMENSION;++coord)
            (*forces)[0][coord] += externalForce[coord];
    }
    AdjustForceIncrements(forces); // does averaging only
}


void NonRotatingRigidBody::AdjustForceIncrements(DMArray<double, DIMENSION> *forces)
{
    double averaged[DIMENSION];
    forces->Average(averaged);
    forces->Set(averaged);
}

void NonRotatingRigidBody::AdjustVelocities(DMArray<double, DIMENSION> *velocities)
{
    AdjustForceIncrements(velocities);
}



int NonRotatingRigidBody::NumberGeneralizedCoordinates()
{
    if(!centerOfMassIsDefined)
        InitializeCOMAndDisplacements();
    return DIMENSION;
}

void NonRotatingRigidBody::InitializeCOMAndDisplacements()
{
    // assumes all atoms have same mass!
    atoms->GetCartesianPositions()->Average(centerOfMass);
    // make COM the negative in order to subtract below; correct afterwards
    int coord;
    for(coord=0;coord<DIMENSION;++coord)
        centerOfMass[coord] *= -1.;
    
    // displacements is siblings of positions
    displacements = new DMArray<double, DIMENSION>(atoms->GetCartesianPositions());
    displacements->Copy(atoms->GetCartesianPositions());
    displacements->Increment(centerOfMass);
    for(coord=0;coord<DIMENSION;++coord)
        centerOfMass[coord] *= -1.;
	
	centerOfMassIsDefined = true;
}


void NonRotatingRigidBody::SetCartesianPositionsFromCOM()
{
    if(!centerOfMassIsDefined)
    {
        InitializeCOMAndDisplacements();
        return;
    }
    int coord;
    // easiest way to set the positions: add the COM to the displacements
    // and set the positions using the displacements
    displacements->Increment(centerOfMass);
    atoms->SetCartesianPositions(displacements);
    
    // then need to subtract the COM from the displacements again.
    for(coord=0;coord<DIMENSION;++coord)
        centerOfMass[coord] *= -1.;
    displacements->Increment(centerOfMass);
    for(coord=0;coord<DIMENSION;++coord)
        centerOfMass[coord] *= -1.;	
}

void NonRotatingRigidBody::SetGeneralizedCoordinates(double *generalizedCoordinates)
{
    if(!centerOfMassIsDefined)
        InitializeCOMAndDisplacements();	// to have displacements be defined
    
    for(int coord=0;coord<DIMENSION;++coord)
        centerOfMass[coord] = generalizedCoordinates[coord]*weightFactor;
    
    SetCartesianPositionsFromCOM();
    
}

void NonRotatingRigidBody::GetGeneralizedCoordinates(double *generalizedCoordinates)
{
    if(!centerOfMassIsDefined)
        InitializeCOMAndDisplacements();	
    
    for(int coord=0;coord<DIMENSION;++coord)
        generalizedCoordinates[coord] = centerOfMass[coord]/weightFactor;
    
}


void NonRotatingRigidBody::IncrementGeneralizedCoordinate(int n,double inc)
{
    if(!centerOfMassIsDefined)
        InitializeCOMAndDisplacements();	
    
    ASSERT(n < DIMENSION);
    
    centerOfMass[n] += inc*weightFactor;
    
    SetCartesianPositionsFromCOM();
    
}


void NonRotatingRigidBody::CalculateGeneralizedForces(
                            DMArray<double, DIMENSION> *forces,
                            double *generalizedForces)
{
    AdjustForces(forces);
    
    for(int coord=0;coord<DIMENSION;++coord)
        generalizedForces[coord] = (*forces)[0][coord] * atoms->GetNumber()*weightFactor;
    
}

void NonRotatingRigidBody::DerivativesWRTGeneralizedCoordinates(int atom,
                                                                double (*gradient)[DIMENSION])
{
    for(int genCoord=0;genCoord<DIMENSION;++genCoord)
        for(int coord=0;coord<DIMENSION;++coord)
            gradient[genCoord][coord] = weightFactor;
}

