// SingleDirectionRigidBody.h: interface for the SingleDirectionRigidBody class.
//
//////////////////////////////////////////////////////////////////////

#ifndef SINGLEDIRECTIONRIGIDBODY_H
#define SINGLEDIRECTIONRIGIDBODY_H

#include "DigitalMaterial.h"
#include "Constraint.h"

//! Like NonRotatingRigidBody, but only coord==0 positions are constrained

/*!

  The single direction can be the coord == 0 direction, coord == 1 direction 
  etc. We store just the constrainedCoord of the center of mass, and average 
  only the corresponding components of the forces etc. So the number
  of degrees of freedom is 1 (the directional COM) plus (DIMENSION-1) 
  times the number of atoms. Having the to deal with all the other positions
  makes the functions dealing with GeneralizedCoordinates a bit messy.

  Perhaps this could be made more general: Different directions other than the
  coordinate axes would be allowed (eg constrain motion along an arbitrary
  straight line to be rigid). Also, it could all be included in the same class
  as NonRotatingRigidBody---just turn on and off the components you want
  contrained,  or pass some kind of matrix to specify them.

  The date below refers to creation of this file. The class previously existed
  in the single Constraint.h file.

  \ingroup MolecularDynamics
  \author Nicholas Bailey
  \date March 1, 2001
  
 */ 


class SingleDirectionRigidBody : public Constraint
{
 public:
	SingleDirectionRigidBody(int constrainedCoord =0,double weight = 1.0); 
	~SingleDirectionRigidBody() ;
	Constraint *Copy() const;

	void AdjustForces(DMArray<double, DIMENSION> *forces);
	void AdjustForceIncrements(DMArray<double, DIMENSION> *forces);
	void AdjustVelocities(DMArray<double, DIMENSION> *velocities);

	int NumberGeneralizedCoordinates();
	void SetGeneralizedCoordinates(double *generalizedCoordinates) ;
    void GetGeneralizedCoordinates(double *generalizedCoordinates); 
	void IncrementGeneralizedCoordinate(int n,double inc);
#ifndef SWIG
    void CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces);
    void DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION]);
#endif // SWIG
	const std::string &GetType() const
		{
			return className;
		}


private:
	int constrainedCoord;
	double weightFactor;
	void InitializeCOMAndDisplacements();
	void SetCartesianPositionsFromCOM(double *restOfPositions = 0);
	double constrainedCenterOfMass;
	bool centerOfMassIsDefined;
	DMArray<double, DIMENSION> *displacements;
	static const std::string className;
};


#endif // !defined(SINGLEDIRECTIONRIGIDBODY_H)
