// DMArray.h: interface for the DMArray classes.
//
//////////////////////////////////////////////////////////////////////

#ifndef DMARRAY_H
#define DMARRAY_H

#include "DMFirstHeader.h"
#include <algorithm>

#include "DigitalMaterial.h"
#include <math.h>
#include "SprintString.h"

#include "DMException.h"
#include "IndexError.h"
#include "DMArray_AnyType.h"
#include "SynchronizationKit.h"
#include "DMWriter.h"
#include "DMReader.h"

template<class Type, int N> class DMArrayIterator;

template<class Type, int N=1>
struct ReturnType{
    typedef Type &IndexOp;
    static IndexOp convert(Type (&a)[1]) {return a[0];}
};

struct ReturnType<double, 2>{
    typedef double (&IndexOp)[2];
    static IndexOp convert(double  (&a)[2]) {return a;}
};
struct ReturnType<double, 3>{
    typedef double (&IndexOp)[3];
    static IndexOp convert(double  (&a)[3]) {return a;}
};
struct ReturnType<double, 4>{
    typedef double (&IndexOp)[4];
    static IndexOp convert(double  (&a)[4]) {return a;}
};
struct ReturnType<double, 6>{
    typedef double (&IndexOp)[6];
    static IndexOp convert(double  (&a)[6]) {return a;}
};
struct ReturnType<int, 2>{
    typedef int (&IndexOp)[2];
    static IndexOp convert(int  (&a)[2]) {return a;}
};
struct ReturnType<int, 3>{
    typedef int (&IndexOp)[3];
    static IndexOp convert(int  (&a)[3]) {return a;}
};
struct ReturnType<int, 4>{
    typedef int (&IndexOp)[4];
    static IndexOp convert(int  (&a)[4]) {return a;}
};


//! This tree structure holds objects which are pointers to storage of fixed length.
/*! Each leaf of the tree is a dmarray class.

  \author Thierry Cretegny
*/

template<class Type, int N=1>
class DMArray : public DMArray_AnyType, public ReturnType<Type, N>
{
    friend class DMArrayIterator<Type, N>;
    friend class SynchronizationKit<Type, N>;
public:
    //! Create a new DMArray as an empty leaf
    DMArray( int n = 0);
    //! Constructor where you pre-allocate space. Pass an in to get a leaf
    DMArray(const Shape &shape);
    //! Constructor where you copy the shape of a DMArray.
    /*! This constructor specifies that the new array should
    be related to the old array.  A call to GetSiblings
    will return both arrays.
    */
    template<class OtherType, int OtherN>
        DMArray(const DMArray<OtherType, OtherN> *array) : DMArray_AnyType(array)
    {
        //TRACE("DMArray::DMArray(const DMARRAY&) executing.\n");
//        upperNode = 0;
        branches = 0;
        data = 0;
        dataCapacity = 1;
        contiguous = false;
        synchroKit = new SynchronizationKit<Type, N>(this);
        synchroKit->SetSynchronizer(array->GetSynchronizationKit()->GetSynchronizer());
        AllocateFromArray(array);
    }
    //! Destructor.
    virtual ~DMArray();
    
    //! Functions which recurse through the tree.
    //@{
    //! Delete all the data and the branches.
    /*! Everything is LOST! Beware : the shape is set to 0
    (meaning the DMArray is a leaf with 0 length)
    
      This deletes leaf data if it is called on a leaf.
      If it is called on a composite, it deletes the branches.
    */
    
    void Clear();
    //! Change the shape the array.
    /*! If the new and the old shapes are similar (ie have the same tree structure), then the data that fits in the 
    new shape is preserved
    */
    void SetShape(const Shape &shape);
    //! Copy the content of array (must have the same shape).
    /*! We use memcpy, so make sure the type is such that we
    can really use memcpy!
    */
    void Copy(const DMArray<Type, N> *array);
    //! Set all entries to \a newValue
    void Set(const Type newValue);
    void Set(const Type newValue[N]) ;
    //@}
    
    //! Accessor functions for values of leaves.
    //@{
    //! Remove the elements in the set
    void RemoveElements(const std::set<Index> &indices);
    //! Add space at the end.
    void AddElementsAtEnd( int n );
    //@}
    
    //! Mathematical access to array.
    //@{
    //!	initializes the array with zeroes
    void Zero();
    
    //! increment all the elements with shift.
    void Increment(const Type shift);
    void Increment(const Type shift[N]);
    
    //! y -> a*x + y. (x and y must have the same shape). Defined only for double[DIMENSION]
    void Axpy(const DMArray<Type, N> *shift, Type scaleFactor);
    //! Like the previous function, but the scale factor can be different for each column of the DMArray
    void Axpy(const DMArray<Type, N> *shift, Type scaleFactor[N]); 
    
	//! Applies a N x N matrix to each element, returning the result in a new DMArray. Defined only for double[DIMENSION]
	DMArray<Type, N> *ApplyMatrix(double rotation[DIMENSION][DIMENSION]) const;

	//! returns Euclidean norm of the DMarray
	double Nrm2();

    //! x *= a. Defined only for double[DIMENSION]
    DMArray<Type, N> &operator*=(Type rescale);
    DMArray<Type, N> &operator*=(Type rescale[N]);
    
    void Sum(Type sum[N]) const;
    void Average(Type avg[N]) const;
    const Shape &ExtendedShape() const {return GetSynchronizationKit()->ExtendedShape();}

    typename ReturnType<Type,N>::IndexOp operator[](int i) const;
    typename ReturnType<Type,N>::IndexOp operator[](const Index &index) const;

    //@}
    
    //! Accessor functions for branches of tree.
    //@{
    //! Whether this is a leaf or a node.
    /*! Even if you have no branches, the container of branches
    is created for nodes. */
    bool IsLeaf() const { return ( 0==branches ); }
    //! How much data this SubArray holds.
    DMArray<Type, N> *GetLeaf(unsigned int i);
    const DMArray<Type, N> *GetLeaf(unsigned int i) const;
    
    //! Return the number of subarrays.
    int NumberOfBranches() const;
    int NumberOfLeaves() const {return cache.flatList.size();}
    //! access the subarrays of a composite array (error thrown if DMArray isn't composite)
    DMArray<Type,N> *GetBranch(int i);
    //! access the subarrays of a composite array (error thrown if DMArray isn't composite)
    const DMArray<Type,N> *GetBranch(int i) const;
    
    //! Remove a child. The DMArray must be composite,
    void RemoveBranch(DMArray<Type, N> *toRemove);
    //! Merge two DMArray. Their shape must be similar (ie have the same tree structure)
    void Merge(const DMArray<Type, N> *array);
    //! Attach a DMArray to \a this. If \this is a leaf it becomes a composite 
    void Attach(DMArray<Type, N> *array);
    //@}
    
    //! Serializable interface. Implemented in ArrayIO.h
    //@{
    //! Save the array to a writer. (In parallel gathers data on node 0 first)
    void Save(DMWriter *writer) const;
    //! Load the array from a reader. 
    void Load(DMReader *reader);
    const std::string &GetType() const ;
    //@}
    SynchronizationKit<Type, N> *GetSynchronizationKit() const
    {return static_cast<SynchronizationKit<Type,N>*>(synchroKit);}

    void Sort(const DMArray<int> *ids);
    
protected:
    //! Set the upperNode of a DMArray. WARNING: we do NOT update the upperNode at this time
    void AddUpperNode(DMArray<Type, N> *upperNode);
    void RemoveUpperNode(DMArray<Type, N> *array);
    class Cache    // I had to inline these functions because they were unresolved by the linker
    {
    public:
        std::vector<int> sizes;
        std::vector<int> starts;
        std::vector<Type (*)[N]> ptrs;
        std::vector<DMArray<Type, N> *> flatList;
        void Clear()
        {
            sizes.clear();
            starts.clear();
            ptrs.clear();
            flatList.clear();
        }
        void push_back(int size, Type (*ptr)[N], DMArray<Type, N> *leaf)
        {
            if(starts.size() == 0) starts.push_back(0);
            else starts.push_back(starts.back()+sizes.back());
            sizes.push_back(size);
            ptrs.push_back(ptr);
            flatList.push_back(leaf);
        }
        void push_back(const Cache &cache){ for(unsigned int i=0; i<cache.sizes.size(); ++i) push_back(cache.sizes[i], cache.ptrs[i], cache.flatList[i]);}
    } cache;
    bool contiguous;
    void Allocate(const Shape &shape);
    template<class OtherType, int OtherN>
        void AllocateFromArray(const DMArray<OtherType,OtherN> *array)
    {
        // we first delete everything and reallocate  ALL THE STORED DATA IS LOST!
        Clear();
        delete [] data;
        data = 0;
        dataCapacity = 1;
        const Shape& shape = array->GetShape();
        if(shape.IsLeaf())
        {
            branches = 0;
            dataCapacity = shape.GetSize();
            if (dataCapacity == 0) dataCapacity = 1;
            data = new Type[dataCapacity][N];
            this->shape = shape;
        }
        else 
        {
            branches = new std::vector<DMArray<Type, N> *>;	
            for(int i=0; i<shape.GetNSubShapes(); ++i)
            {
                branches->push_back(new DMArray<Type, N>(array->GetBranch(i)));
                branches->back()->AddUpperNode(this);
            }
        }
        UpdateShapeAndCache();
    }
    void UpdateShapeAndCache();
    typedef std::vector<DMArray<Type, N> *> ArrayContainer;
    ArrayContainer upperNodes;
    std::vector<DMArray<Type, N> *> *branches;
    Type (*data)[N];
    int dataCapacity;
	static const std::string className;
};



// ######################################################################
// DMArray Definitions
// ######################################################################

template <class Type, int N>
DMArray<Type, N>::DMArray(int n) : DMArray_AnyType(n)
{
    Shape shape(n);
    //	TRACE("_DMArray<Type, N>::_DMArray void constructor\n");
    branches = 0;
    data = 0;
    dataCapacity = 1;
    contiguous = false;
    Allocate(shape);
    synchroKit = new SynchronizationKit<Type, N>(this);
}

template <class Type, int N>
DMArray<Type, N>::DMArray(const Shape& shape) : DMArray_AnyType(shape)
{
    //TRACE( "DMArray<Type, N>::DMArray constructed from Shape.\n");
    branches = 0;
    data = 0;
    dataCapacity = 1;
    contiguous = false;
    Allocate(shape);
    synchroKit = new SynchronizationKit<Type, N>(this);
}

template <class Type, int N>
DMArray<Type, N>::~DMArray()
{
    delete synchroKit;
    ArrayContainer upperNodesCopy = upperNodes; // see ASSERT below
    for(typename ArrayContainer::iterator it=upperNodesCopy.begin(); it!=upperNodesCopy.end();
        ++it)
            (*it)->RemoveBranch(this);
    ASSERT(upperNodes.size() == 0); // upperNodes was emptied when we removed branches
    Clear(); // will allocate 1 length data;
    delete [] data;
}

template<class Type, int N>
void DMArray<Type, N>::Clear()
{
    if(data)
    {
        delete [] data;
        data = 0;
        dataCapacity = 1;
    }
    if(branches)
    {
        for(typename std::vector<DMArray<Type, N>* >::reverse_iterator sub = branches->rbegin(); 
        sub != branches->rend(); ++sub) 
        {
            (*sub)->RemoveUpperNode(this);
            if((*sub)->upperNodes.size() == 0) delete *sub;
        }
        delete branches;
        branches = 0;
    }
    shape = Shape(0);
    ASSERT((dataCapacity == 1) && (data==0) && (branches == 0)) ;
    data = new Type[dataCapacity][N]; // we don't want to let the internal state in an undefined state (ie with both data and branches ==0)
    UpdateShapeAndCache();
}

template <class Type, int N>
void DMArray<Type, N>::Allocate(const Shape &shape)
{
    // we first delete everything and reallocate  ALL THE STORED DATA IS LOST!
    Clear();
    delete [] data;
    data = 0;
    dataCapacity = 1;
    if(shape.IsLeaf())
    {
        branches = 0;
        dataCapacity = shape.GetSize();
        if(dataCapacity == 0) dataCapacity = 1;
        data = new Type[dataCapacity][N];
        this->shape = shape;
    }
    else 
    {
        branches = new std::vector<DMArray<Type, N> *>;	
        for(int i=0; i<shape.GetNSubShapes(); ++i)
        {
            branches->push_back(new DMArray<Type, N>(shape.GetSubShape(i)));
            branches->back()->AddUpperNode(this);
        }
    }
    UpdateShapeAndCache();
}


template <class Type, int N>
void DMArray<Type, N>::SetShape(const Shape &shape)
{
    if(!this->shape.IsSimilar(shape))
    {
        Clear();
        Allocate(shape);
        return;
    }
    if(shape.IsLeaf()) 
    {
        if(shape.GetSize() != this->shape.GetSize())
        {
            int oldlength = this->shape.GetSize();
            int newlength = shape.GetSize();
            // let's say we can afford up to 100 entries unused 
            if((newlength > dataCapacity) || (dataCapacity-newlength > 100)) 
            {
                Type (*oldData)[N] = data;
                data = new Type[newlength][N];
                dataCapacity = newlength;
                if (newlength <= oldlength) memcpy(data, oldData, newlength*N*sizeof(Type));
                else 
                {
                    memcpy(data, oldData, oldlength*N*sizeof(Type));
                    memset(data+oldlength, 0, (newlength-oldlength)*N*sizeof(Type));
                }
                delete [] oldData;
            }
            else
                if(newlength > oldlength)
                    memset(data+oldlength, 0, (newlength-oldlength)*N*sizeof(Type));
                this->shape = shape;
                UpdateShapeAndCache();
        }
    }
    else
        for(unsigned int i=0; i<branches->size();++i)
            (*branches)[i]->SetShape(shape.GetSubShape(i));
    //UpdateShapeAndCache will be called up the tree and shape will be reset,
    //so we don't need to change it here. 
}


template<class Type, int N>
void DMArray<Type, N>::RemoveBranch(DMArray<Type, N> *toRemove)
{
    if(!branches) throw DMException("DMArray<Type, N>", "RemoveBranch",
        "Cannot remove a branch from a leaf");
    typename std::vector<DMArray<Type, N> *>::iterator found = 
        std::find(branches->begin(), branches->end(),toRemove);
    if(found == branches->end())
        throw DMException("DMArray<Type, N>", "RemoveBranch", "branch to remove not found");
    (*found)->RemoveUpperNode(this);  // we also want to notify the child that we cut our relations!
    branches->erase(found);
    UpdateShapeAndCache();
}


template <class Type, int N>
void DMArray<Type,N>::AddUpperNode(DMArray<Type,N> *upperNode)
{
    typename ArrayContainer::iterator found;
    found = std::find(upperNodes.begin(), upperNodes.end(), upperNode);
    if(found == upperNodes.end()) upperNodes.push_back(upperNode);
}


template<class Type, int N>
void DMArray<Type, N>::RemoveUpperNode(DMArray<Type, N> *array)
{
    typename ArrayContainer::iterator found;
    found = std::find(upperNodes.begin(), upperNodes.end(), array);
    ASSERT(found != upperNodes.end());
    if(found != upperNodes.end()) upperNodes.erase(found);
}


template<class Type, int N>
void DMArray<Type, N>::UpdateShapeAndCache()
{
    cache.Clear();
    if(branches)
    {
        shape = Shape(); // sets it as an empty composite
        for(typename std::vector<DMArray<Type, N> *>::iterator iter=branches->begin(); iter != branches->end(); ++iter)
        {
            cache.push_back((*iter)->cache);
            shape.Attach((*iter)->shape);
        }
        contiguous = true;
        for(unsigned int i=1; i<cache.sizes.size(); ++i) contiguous &= (cache.ptrs[i] == cache.ptrs[i-1] + cache.sizes[i-1]);
    }
    else
    {
        cache.push_back(shape.GetSize(), data, this);                   
        contiguous = true;
        // shape is up to date for a leaf.
    }
    for(typename ArrayContainer::iterator it=upperNodes.begin(); it!=upperNodes.end();
        ++it)
        (*it)->UpdateShapeAndCache();
}



/*! When this removes an element, it moves all data below the element
up one notch to fill in the gap.
*/
template<class Type, int N>
void DMArray<Type, N>::RemoveElements(const std::set<Index> &indices)
{
    // user should be aware that the index of the remaining entries 
    // are screwed up after calling this routine!!!
    if(indices.size() == 0)	return;
    int oldSize = Size();
    std::set<Index>::const_iterator rmIt=indices.begin();
    int currentLeaf    = rmIt->leaf;
    int previousItem   = rmIt->inLeaf;
    int removedInCurrentLeafCnt = 0;
    Type (*currentData)[N] = cache.ptrs[currentLeaf];
    for(++rmIt; rmIt != indices.end(); ++rmIt)
    {
        if(rmIt->leaf == currentLeaf)
        {
            memmove(currentData[previousItem-removedInCurrentLeafCnt],
                currentData[previousItem+1],
                (rmIt->inLeaf - previousItem - 1)*N*sizeof(Type));
            ++removedInCurrentLeafCnt;
        }
        else
        {
            int thisSize = cache.sizes[currentLeaf];
            memmove(currentData[previousItem - removedInCurrentLeafCnt],
                currentData[previousItem+1],
                (thisSize - previousItem - 1)*N*sizeof(Type));
            cache.flatList[currentLeaf]->shape = thisSize - removedInCurrentLeafCnt - 1;
            cache.flatList[currentLeaf]->UpdateShapeAndCache();
            removedInCurrentLeafCnt = 0;
            currentLeaf = rmIt->leaf;
            currentData = cache.ptrs[currentLeaf];
        }
        previousItem = rmIt->inLeaf;
    }
    int thisSize = cache.sizes[currentLeaf];
    memmove(currentData[previousItem - removedInCurrentLeafCnt],
        currentData[previousItem + 1],
        (thisSize - previousItem - 1)*N*sizeof(Type));
    cache.flatList[currentLeaf]->shape = thisSize - removedInCurrentLeafCnt - 1;
    cache.flatList[currentLeaf]->UpdateShapeAndCache();
    ASSERT(oldSize == Size() + indices.size());
}

template <class Type, int N>
void DMArray<Type,N>::AddElementsAtEnd( int n )
{
    if (n == 0) return;
    ASSERT( true==IsLeaf() );
    if (n<=0) throw DMException("DMArray<Type, N>","AddElementsAtEnd",
                                "You have to add a positive number.");
    if (!IsLeaf()) throw DMException("DMArray<Type, N>","AddElementsAtEnd",
        "Cannot add elements to a node.");

    int m = shape.GetSize();
    Shape s(m+n);
    ASSERT( s.GetSize() == shape.GetSize()+n );
    this->SetShape( s );
}


template <class Type, int N>
void DMArray<Type, N>::Copy(const DMArray<Type,N> *array)
{
    if(!(shape == array->GetShape()))
        throw DMException("DMArray<Type, N>", "Copy",
                          "tried to copy arrays of different shapes");
    int size=cache.sizes.size();
    for(int i=0; i<size; ++i)
        memcpy(cache.ptrs[i], array->cache.ptrs[i],
        cache.sizes[i]*N*sizeof(Type)); 
}

template <class Type, int N>
void DMArray<Type, N>::Set(const Type newValue)
{
    int size=cache.sizes.size();
    for(int i=0; i<size; ++i) 
    {
        int length = cache.sizes[i];
        Type (*ptr)[N] = cache.ptrs[i];
        for(int j=0; j<length; ++j)
            for(int coord=0; coord<N; coord++)
                ptr[j][coord] = newValue;
    }
}

template <class Type, int N>
void DMArray<Type, N>::Set(const Type newValue[N])
{
    int size=cache.sizes.size();
    for(int i=0; i<size; ++i) 
    {
        int length = cache.sizes[i];
        Type (*ptr)[N] = cache.ptrs[i];
        for(int j=0; j<length; ++j)
            for(int coord=0; coord<N; coord++)
                ptr[j][coord] = newValue[coord];
    }
}

template<class Type, int N>
void DMArray<Type, N>::Zero()
{
    int size=cache.sizes.size();
    for(int i=0; i<size; ++i) memset(cache.ptrs[i], 0, cache.sizes[i]*N*sizeof(Type));
}


template<class Type, int N>
void DMArray<Type, N>::Increment(const Type shift[N])
{
    int size=cache.sizes.size();
    for(int i=0; i<size; ++i) 
    {
        int length = cache.sizes[i];
        Type (*ptr)[N] = cache.ptrs[i];
        for(int j=0; j<length; ++j)
            for(int coord=0; coord<N; coord++)
                ptr[j][coord] += shift[coord];  // there might be a way to speed this up
    }
}

template<class Type, int N>
void DMArray<Type, N>::Increment(const Type shift)
{
    int size=cache.sizes.size();
    for(int i=0; i<size; ++i) 
    {
        int length = cache.sizes[i];
        Type (*ptr)[N] = cache.ptrs[i];
        for(int j=0; j<length; ++j)
            for(int coord=0; coord<N; coord++)
                ptr[j][coord] += shift;  // there might be a way to speed this up
    }
}


template<>
inline void DMArray<double, DIMENSION>::Axpy(const DMArray<double, DIMENSION> *shift, double scaleFactor)
{
    if(!(shape == shift->GetShape()))
        throw DMException("DMArray<double, DIMENSION>", "Axpy (scalar version)",
                          "tried to Axpy arrays of different shapes");
    int size = cache.sizes.size();
    for(int i=0; i<size; ++i) if(cache.sizes[i] > 0)
        cblas_daxpy(cache.sizes[i]*DIMENSION, scaleFactor, (const double *)(shift->cache.ptrs[i]),
            1, (double *)(cache.ptrs[i]), 1); 
}

template<>
inline void DMArray<double, DIMENSION>::Axpy(const DMArray<double, DIMENSION> *shift, double scaleFactor[DIMENSION])
{
    if(!(shape == shift->GetShape())) 
        throw DMException("DMArray<double, DIMENSION>", "Axpy (array version)", 
                          "tried to Axpy arrays of different shapes");
    int size = cache.sizes.size();
    for(int i=0; i<size; ++i) if(cache.sizes[i] > 0)
        for(int coord=0;coord<DIMENSION;++coord)
            cblas_daxpy(cache.sizes[i], scaleFactor[coord], ((const double *)(shift->cache.ptrs[i])) + coord, 
                DIMENSION, ((double *)(cache.ptrs[i])) + coord, DIMENSION);
        
}


template<>
inline DMArray<double, DIMENSION> *DMArray<double, DIMENSION>::ApplyMatrix(double rotation[DIMENSION][DIMENSION]) const
{
	DMArray<double, DIMENSION> *rotatedArray = new DMArray<double, DIMENSION>(this->GetShape());
	rotatedArray->Zero();

	int size = cache.sizes.size();
    for(int i=0; i<size; ++i) if(cache.sizes[i] > 0)
	{
		int leafSize = cache.sizes[i];

		// coord1 is first index of matrix, refers to destination
		for( int coord1 = 0; coord1 < DIMENSION; ++coord1 )
			// coord2 is second index, refers to source
			for( int coord2 = 0; coord2 < DIMENSION; ++coord2 )
            cblas_daxpy(leafSize, rotation[coord1][coord2], ((const double *)(cache.ptrs[i])) + coord2, 
                DIMENSION, ((double *)(rotatedArray->cache.ptrs[i])) + coord1, DIMENSION);
        

	}

	return rotatedArray;
}


template<>
inline double DMArray<double, DIMENSION>::Nrm2()
{
	double result = 0.;
	int size = cache.sizes.size();
    for(int i=0; i<size; ++i) if(cache.sizes[i] > 0)
	{
		double temp = cblas_dnrm2(cache.sizes[i] * DIMENSION, ((const double *)(cache.ptrs[i])), 1);
		result += temp * temp;
	}
	
	return sqrt(result);
}


template<class Type, int N>		
typename ReturnType<Type,N>::IndexOp DMArray<Type, N>::operator [](int i) const
{
    //! We perform range checking in DEBUG mode only.
    ASSERT(i>=0 && (i<shape.GetSize())); 
    if(contiguous)
        return convert(cache.ptrs.front()[i]);
    else
    {
        int j = cache.sizes.size()-1;
        while(cache.starts[j] > i) --j;
        return convert(*(cache.ptrs[j]+i-cache.starts[j]));
    }
}

template<class Type, int N>		
typename ReturnType<Type,N>::IndexOp DMArray<Type, N>::operator [](const Index &ind) const
{
    //! We perform range checking in DEBUG mode only.
    ASSERT(ind.leaf>=0 && (ind.leaf<cache.ptrs.size()));
    ASSERT(ind.inLeaf>=0 && (ind.inLeaf<cache.sizes[ind.leaf]));
    return convert(cache.ptrs[ind.leaf][ind.inLeaf]);
}

/*! This might be called GetNumberOfSubArrays, but more general language for
a leaf structure seems appropriate.
*/

template<class Type, int N>
int DMArray<Type, N>::NumberOfBranches( ) const
{
    if (0 != branches) return branches->size();
    return 0;
}

//! access the leaves of a tree structure
template<class Type, int N>
DMArray<Type, N> *DMArray<Type, N>::GetLeaf(unsigned int i)
{
    ASSERT(i < cache.flatList.size());
    return cache.flatList[i];
}
//! access the leaves of a tree structure
template<class Type, int N>
const DMArray<Type, N> *DMArray<Type, N>::GetLeaf(unsigned int i) const 
{
    ASSERT(i < cache.flatList.size());
    return cache.flatList[i];
}

//! access the subarrays of a composite array
template<class Type, int N>
DMArray<Type, N> *DMArray<Type, N>::GetBranch(int i)
{
    if(branches) return (*branches)[i];
    else throw DMException("DMArray<Type, N>", "GetBranch", "attempt to access a branch of a leaf"); // we may want to accept i=0 in this case. 
}

template<class Type, int N>
const DMArray<Type, N> *DMArray<Type, N>::GetBranch(int i) const
{
    if(branches) return (*branches)[i];
    else throw DMException("DMArray<Type, N>", "GetBranch","attempt to access a branch of a leaf"); // we may want to accept i=0 in this case. 
}

template<>
inline DMArray<double, DIMENSION> &DMArray<double, DIMENSION>::operator*=(double rescale)
{
    int size = cache.sizes.size();
    for(int i=0; i<size; ++i)
        if(cache.sizes[i] > 0)   // some blas versions seem to have problems with size 0.
            cblas_dscal(cache.sizes[i]*DIMENSION, rescale, (double *)cache.ptrs[i], 1);
    return *this;
}

template<>
inline DMArray<double, DIMENSION> &DMArray<double, DIMENSION>::operator*=(double rescale[DIMENSION])
{
    int size = cache.sizes.size();
    for(int i=0; i<size; ++i)
    {
        if(cache.sizes[i] > 0) // some blas versions seem to have problems with size 0.
            for(int coord=0;coord<DIMENSION;++coord)
                cblas_dscal(cache.sizes[i], rescale[coord], ((double *)cache.ptrs[i]) + coord, DIMENSION);
    }
    return *this;
}


template<class Type, int N>
void DMArray<Type, N>::Merge(const DMArray<Type, N> *array)
{
    if(branches)
        for(unsigned int i=0; i<branches->size(); ++i) (*branches)[i]->Merge(array->GetBranch(i));
    else
    {
        assert(0 != data);
        if(!array->shape.IsLeaf()) throw DMException("DMArray<Type, N>", "Merge", "cannot merge arrays of different shape");
        int lengthA = shape.GetSize();
        int lengthB = array->shape.GetSize();
        Type (*oldData)[N] = data;
        data = new Type[lengthA+lengthB][N];
        dataCapacity = lengthA+lengthB;
        memcpy(data, oldData, lengthA*N*sizeof(Type));
        memcpy(data+lengthA, array->data, lengthB*N*sizeof(Type));
        delete [] oldData;
        shape = Shape(lengthA+lengthB);
        UpdateShapeAndCache();
    }
}

template<class Type, int N>
void DMArray<Type, N>::Attach(DMArray<Type, N> *array)
{
    if(!branches)
    {
        if(shape.GetSize() > 0) throw DMException("DMArray<Type, N>", "Attach", "Cannot attach an array to a leaf array");
        if(data) delete [] data;
        data = 0;
        dataCapacity = 1;
        branches = new std::vector<DMArray<Type, N> *>;
    }
    branches->push_back(array);
    array->AddUpperNode(this);
    UpdateShapeAndCache();
}

template<class Type, int N>
std::ostream &operator<<(std::ostream &out, const DMArray<Type, N> *array)
{
    DMArrayIterator<Type, N> iter(array);
    for(iter.Init(); !iter.IsDone(); iter.Next())
    {
        for(int j=0; j<N; ++j) out << iter.Current()[j] << " ";
        out << endl;
    }
    return out;
}

#include "DMArrayIterator.h"
#include "ArrayMathOps.h"
#include "ArrayIO.h"

#endif // !defined(DMARRAY_H)
