// MDSimulation.cpp: implementation of the MDSimulation class.
//
//////////////////////////////////////////////////////////////////////

#ifdef _WINDOWS
#include "stdafx.h"
#include "MD.h"
#include "MDSimulation.h"
#endif // _WINDOWS

#include "PrimitiveListOfAtoms.h"

#include "VerletAtomsMover.h"
#include "HooverAtomsMover.h"
#include "VerletThermalizeAtomsMover.h"
#include "QuickMinAtomsMover.h"
#include "LangevinAtomsMover.h"
#include "PredictorCorrectorAtomsMover.h"

#include "CutLennardJonesPotential.h"
#include "EMTPotential.h"
#include "GravityPotential.h"

#include "FreeBoundaryConditions.h"
#include "PeriodicBoundaryConditions.h"
#include "SimplePeriodicBoundaryConditions.h"

#include "SimpleNeighborLocator.h"
#include "CellNeighborLocator.h"
#include "NeighborList.h"

#include "RectangularClusterInitializer.h"
#include "SphericalClusterInitializer.h"

#include "DislocationMaker.h"	// XXX Should be called "DislocationTransformer.h"
#include "OverlapPruner.h"

#include "EnergyObserver.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MDSimulation::MDSimulation() : listOfListsOfAtoms("ListsOfAtoms"), listOfAtomsMovers("AtomsMovers"),
		listOfPotentials("Potentials"), listOfBoundaryConditions("BoundaryConditions"),
		listOfNeighborLocators("NeighborLocators"), listOfBravaisLattices("BravaisLattices"),
		listOfAtomsInitializers("AtomsInitializers"), listOfTransformers("Transformers"),
		listOfObservers("Observers")
{

	// Set up list of lists: order here determines order on menu

	listOfLists.push_back(&listOfListsOfAtoms);
	listOfLists.push_back(&listOfPotentials);
	listOfLists.push_back(&listOfBravaisLattices);
	listOfLists.push_back(&listOfAtomsInitializers);
	listOfLists.push_back(&listOfAtomsMovers);
	listOfLists.push_back(&listOfBoundaryConditions);
	listOfLists.push_back(&listOfNeighborLocators);
	listOfLists.push_back(&listOfTransformers);
	listOfLists.push_back(&listOfObservers);

	// (1) Lists of atoms
	// Need to attach boundary conditions and neighbor list to the list of atoms
	listOfListsOfAtoms.push_back(new PrimitiveListOfAtoms);
	listOfListsOfAtoms.push_back(new PrimitiveListOfAtoms);
	listOfListsOfAtoms.push_back(new PrimitiveListOfAtoms);
	listOfListsOfAtoms.push_back(new PrimitiveListOfAtoms);
	listOfListsOfAtoms.SetCurrentNumber(0);

	// (2) Potentials
	listOfPotentials.push_back(new CutLennardJonesPotential);
	if (DIMENSION==3)	listOfPotentials.push_back(new EMTPotential);
	listOfPotentials.push_back(new GravityPotential);
	listOfPotentials.SetCurrent("CutLennardJonesPotential");
	oldPotentialNumber = listOfPotentials.GetCurrentNumber();	// XXX Crude

	// (3) Atoms Movers
	listOfAtomsMovers.push_back(new VerletAtomsMover(GetPotential(),GetTimeStep()/10.0,10));
	listOfAtomsMovers.push_back(new HooverAtomsMover(GetPotential(),0.005,GetTimeStep()/10.0,2.0,10));
	listOfAtomsMovers.push_back(new VerletThermalizeAtomsMover(GetPotential(),0.2,GetTimeStep()/10.0,10));
	listOfAtomsMovers.push_back(new QuickMinAtomsMover(GetPotential(),GetTimeStep()/10.0,0.001));
	listOfAtomsMovers.push_back(new LangevinAtomsMover(GetPotential(),0.2,GetTimeStep()/10.0,10));
	listOfAtomsMovers.push_back(new PredictorCorrectorAtomsMover(GetPotential(),GetTimeStep()/10.0,10));
	listOfAtomsMovers.SetCurrent("VerletThermalizeAtomsMover");


	// (4) Bravais lattices
	if (DIMENSION==3)
	{
		listOfBravaisLattices.push_back(new FCCLattice);
		listOfBravaisLattices.push_back(new BCCLattice);
		listOfBravaisLattices.SetCurrent("FCCLattice");
	}
	else if (DIMENSION==2)
	{
		listOfBravaisLattices.push_back(new SquareLattice);
		listOfBravaisLattices.push_back(new TriangularLattice);
		listOfBravaisLattices.SetCurrent("TriangularLattice");
	}

	//  Reset lattice constants for Bravais lattices
	for (int numBravais=0; numBravais < listOfBravaisLattices.size(); numBravais++)
	{
		BravaisLatticeWithBasis *bravais = (BravaisLatticeWithBasis *)listOfBravaisLattices[numBravais];
		bravais->RescaleLatticeVectors(GetDiameter() / bravais->GetMinInteratomicSpacing());
	}


	// (5) Boundary Conditions
	// Pick a rectangle that extends at least one cutoff distance in each direction
	// and is an integer number of lattice constants

	int cutoffDividedByLatticeConstant = int(GetPotential()->GetCutoffDistance() / GetBravaisLattice()->GetLatticeSpacing() +1);
	if (cutoffDividedByLatticeConstant == 0) cutoffDividedByLatticeConstant ++;
	for (int coord=0; coord<DIMENSION; coord++)
	{
		rectangleSize[coord] = (DIMENSION-coord+cutoffDividedByLatticeConstant)*GetBravaisLattice()->GetLatticeSpacing();
		center[coord] = 0;
	}
	listOfBoundaryConditions.push_back(new FreeBoundaryConditions);
	listOfBoundaryConditions.push_back(new SimplePeriodicBoundaryConditions(rectangleSize));
	listOfBoundaryConditions.SetCurrent("FreeBoundaryConditions");
	for (int atomList = 0; atomList < listOfListsOfAtoms.size(); atomList++)
	{
		((ListOfAtoms *)listOfListsOfAtoms[atomList])->SetBoundaryConditions(GetBoundaryConditions());
	}

	// (6) Neighbor Locators
	listOfNeighborLocators.push_back(new SimpleNeighborLocator(GetPotential()->GetCutoffDistance()));
	listOfNeighborLocators.push_back(new CellNeighborLocator(
		GetPotential()->GetCutoffDistance(), 0.2*GetPotential()->GetCutoffDistance()));
	if (DIMENSION==3)	
	{
		listOfNeighborLocators.push_back(new NeighborList(
							GetPotential()->GetCutoffDistance(), 1.2*GetPotential()->GetCutoffDistance()));
	}
	listOfNeighborLocators.SetCurrent("SimpleNeighborLocator");

	// (7) Atoms Initializers
	listOfAtomsInitializers.push_back(new RectangularClusterInitializer(rectangleSize,GetBravaisLattice()));
	listOfAtomsInitializers.push_back(new SphericalClusterInitializer(3*GetDiameter(),GetBravaisLattice()));
	listOfAtomsInitializers.SetCurrent("SphericalClusterInitializer");

	if (listOfBoundaryConditions.GetCurrentNumber() == 1)	// Periodic BC go from 0 to L...
	{
		for (int coord=0; coord<DIMENSION; coord++) center[coord] = rectangleSize[coord]/2.;
		((ClusterInitializer *)GetAtomsInitializer())->SetCenter(center);
	}

//	for (atomList = 0; atomList < listOfListsOfAtoms.size(); atomList++)
//	{
//		((ListOfAtoms *)listOfListsOfAtoms[atomList])->SetNeighborLocator(GetNeighborLocator());
//	}
	GetListOfAtoms()->SetNeighborLocator(GetNeighborLocator());

	// (8) Transformers
	double direction[3] = {1.0,0.0,0.0};	// Two dimensions direction is ignored
	
	listOfTransformers.push_back(new DislocationMaker(GetBravaisLattice()->GetLatticeVector(0), direction, center,0.3));
	listOfTransformers.SetCurrent("DislocationMaker");


	// (9) Observers
	listOfObservers.push_back(new EnergyObserver(GetPotential()));
	listOfObservers.SetCurrent("EnergyObserver");

	// Should attach all "selected" observers

	GetListOfAtoms()->Attach(GetObserver());
	
	Reset();
}

MDSimulation::~MDSimulation()
{
	int item;

	for (item=0; item<listOfAtomsInitializers.size(); item++)	
		delete listOfAtomsInitializers[item];
	for (item=0; item<listOfBravaisLattices.size(); item++)	
		delete listOfBravaisLattices[item];
	for (item=0; item<listOfAtomsMovers.size(); item++)	
		delete listOfAtomsMovers[item];
	for (item=0; item<listOfBoundaryConditions.size(); item++)	
		delete listOfBoundaryConditions[item];
	for (item=0; item<listOfListsOfAtoms.size(); item++)	
		delete listOfListsOfAtoms[item];
	for (item=0; item<listOfNeighborLocators.size(); item++)	
		delete listOfNeighborLocators[item];
	for (item=0; item<listOfPotentials.size(); item++)	
		delete listOfPotentials[item];
}

void MDSimulation::Reset()
{
	int potentialNumber = listOfPotentials.GetCurrentNumber();

	// (3)
	if (oldPotentialNumber != potentialNumber)	// Ugly! Otherwise resets lattice to original
	{
		for (int atomsMoverNumber = 0; atomsMoverNumber < listOfAtomsMovers.size(); atomsMoverNumber++)
		{
			AtomsMover *mover = (AtomsMover *)listOfAtomsMovers[atomsMoverNumber];

			mover->SetPotential(GetPotential());
			mover->SetDt(GetTimeStep()/10.0);
		}
	}
	
	// (4)
	//  Reset lattice constants for Bravais lattices
	if (oldPotentialNumber != potentialNumber)	// Ugly! Otherwise resets lattice to original
	{
		for (int numBravais=0; numBravais < listOfBravaisLattices.size(); numBravais++)
		{
			BravaisLatticeWithBasis *bravais = (BravaisLatticeWithBasis *)listOfBravaisLattices[numBravais];
			bravais->RescaleLatticeVectors(GetDiameter() / bravais->GetMinInteratomicSpacing());
		}
	}

	// (5) Boundary Conditions
	// Pick a rectangle that extends at least one cutoff distance in each direction
	// and is an integer number of lattice constants
	if (oldPotentialNumber != potentialNumber)
	{
		int cutoffDividedByLatticeConstant = int(GetPotential()->GetCutoffDistance() / GetBravaisLattice()->GetLatticeSpacing() +1);
		if (cutoffDividedByLatticeConstant == 0) cutoffDividedByLatticeConstant ++;
		for (int coord=0; coord<DIMENSION; coord++)
		{
			rectangleSize[coord] = (DIMENSION-coord+cutoffDividedByLatticeConstant)*GetBravaisLattice()->GetLatticeSpacing();
		}
	}

	for (int coord=0; coord < DIMENSION; coord++)
	{
		if (listOfBoundaryConditions.GetCurrentNumber() == 1) 
			center[coord] = rectangleSize[coord]/2.;
		else center[coord] = 0;
	}

	for (int atomList = 0; atomList < listOfListsOfAtoms.size(); atomList++)
	{
		((ListOfAtoms *)listOfListsOfAtoms[atomList])->SetBoundaryConditions(GetBoundaryConditions());
		((ListOfAtoms *)listOfListsOfAtoms[atomList])->SetNeighborLocator(GetNeighborLocator());
	}

	// (6)
	if (oldPotentialNumber != potentialNumber)
	{
		for (int numNL = 0; numNL < listOfNeighborLocators.size(); numNL++)
		{
			NeighborLocator *nL = (NeighborLocator *) listOfNeighborLocators[numNL];
			nL->SetCutoffDistance(GetPotential()->GetCutoffDistance());
			nL->SetDrift(0.2*GetPotential()->GetCutoffDistance());
		}
	}
	for (atomList = 0; atomList < listOfListsOfAtoms.size(); atomList++)
	{
		((ListOfAtoms *)listOfListsOfAtoms[atomList])->SetBoundaryConditions(GetBoundaryConditions());
	}

	// (7)
	((ClusterInitializer *)GetAtomsInitializer())->SetCenter(center);

	for (int initializerNumber = 0; initializerNumber < listOfAtomsInitializers.size(); initializerNumber++)
	{
			if (oldPotentialNumber != potentialNumber)
			{
				if (initializerNumber==0)	// Rectangular
					((RectangularClusterInitializer *) listOfAtomsInitializers[initializerNumber])->SetRectangleSize(rectangleSize);
				else if (initializerNumber==1) // Spherical
				((SphericalClusterInitializer *) listOfAtomsInitializers[initializerNumber])->SetRadius(2.0*GetDiameter());
			}
			((ClusterInitializer *)listOfAtomsInitializers[initializerNumber])->SetBravaisLattice(GetBravaisLattice());
	}

	GetListOfAtoms()->SetNeighborLocator(GetNeighborLocator()); //WARNING: The old list of atoms still points to this neighborlocator

	// (9) 
	// XXX Crude 
	((EnergyObserver *)listOfObservers[0])->SetPotential(GetPotential());

	// Note new potential
	oldPotentialNumber = listOfPotentials.GetCurrentNumber();	// XXX Crude

	// Reset velocities to zero
	GetListOfAtoms()->ScaleCartesianVelocities(0.0);	// Zero velocities

	// Create cluster
	GetAtomsInitializer()->Create(GetListOfAtoms());	// Sets up cluster

	// Check for overlapping atoms
	if (listOfBoundaryConditions.GetCurrentNumber()==1) // SimplePeriodicBoundaryConditions
	{
		PeriodicBoundaryConditions *pbc = (PeriodicBoundaryConditions *)GetBoundaryConditions();
		bool overlap = false;
		int initializerNumber = listOfAtomsInitializers.GetCurrentNumber();
		if (initializerNumber==0) // Rectangular
		{
			RectangularClusterInitializer *rect = 
				(RectangularClusterInitializer *) GetAtomsInitializer();
			for (int coord=0; coord<DIMENSION; coord++)
			{
				if (pbc->GetLength(coord)< 1.01*rect->GetRectangleSize()[coord]) overlap=true;
			}
		}
		else if (initializerNumber==1) // Spherical
		{
			SphericalClusterInitializer *sphere = 
				(SphericalClusterInitializer *) GetAtomsInitializer();
			for (int coord=0; coord<DIMENSION; coord++)
			{
				if (pbc->GetLength(coord)< 2.01*sphere->GetRadius()) overlap=true;
			}
		}
		if (overlap)
		{
			OverlapPruner overlapPruner(0.9*GetDiameter());
			overlapPruner.Transform((PrimitiveListOfAtoms *)GetListOfAtoms());
		}
	}

/*	// XXX For class, should change these
//	atomView->SetPlottedRegionSize(3*sim->GetMaxSize());
//	atomView->SetCenter(sim->GetAtomsInitializer()->GetCenter());
	double plotCenter[2] = {0,0};
	atomView->SetCenter(plotCenter);
	atomView->SetPlottedRegionSize(40);

	// Change for class: old
	// Create cluster
// sim->GetAtomsInitializer()->Create(sim->GetListOfAtoms());	// Sets up cluster

// Thought for class
	PrimitiveListOfAtoms *atoms0 = (PrimitiveListOfAtoms *)sim->listOfListsOfAtoms.components[0];
	PrimitiveListOfAtoms *atoms1 = (PrimitiveListOfAtoms *)sim->listOfListsOfAtoms.components[1];
	RectangularClusterInitializer* rect = (RectangularClusterInitializer *)sim->listOfAtomsInitializers.components[0];
	SphericalClusterInitializer* sphere = (SphericalClusterInitializer *)sim->listOfAtomsInitializers.components[1];
	

	rect->Create(atoms0);	// Sets up surface

*/

}

double MDSimulation::GetMaxSize()
{
	if (GetBoundaryConditions()->GetName() == "SimplePeriodicBoundaryConditions")
	{
		double maxSize = 0;
		for (int coord=0; coord<DIMENSION; coord++) 
		{
			double extent = 
				((SimplePeriodicBoundaryConditions *)GetBoundaryConditions())->GetLength(coord);
			if (maxSize < extent) maxSize = extent;
		}
		return maxSize;
	}
	else return GetAtomsInitializer()->GetMaxSize();
}
