// CellNeighborList.cpp: implementation of the CellNeighborList class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"

#include "CellNeighborList.h"
#include "BoundaryConditions.h"
#include "DMException.h"
#include "LeafToLeafIterator.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<CellNeighborList> *_cellNListcreator = 
	new SerialBuilder<CellNeighborList>("CellNeighborList");


const std::string DMArray<CellNeighborList::TypedNBListElement*>::className = std::string("DMArray<TypedNBListElement*>");


const string CellNeighborList::className = string("CellNeighborList");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CellNeighborList::CellNeighborList(double cutoffDistance, double drift) : maxNeighbors(100), 
	CellNeighborLocator(cutoffDistance, drift) 
{
	this->neighborList          = 0;
	this->typedNeighborList     = 0;
	this->typedNBListBuilt      = false;
	this->numberOfHalfNeighbors = 0;
	this->timesRebuilt          = 0;
}

CellNeighborList::~CellNeighborList()
{
	if(neighborList) delete [] neighborList;
	if(typedNeighborList)
	{
		DMArrayIterator<TypedNBListElement *> iter(typedNeighborList);
		for(iter.Init(); !iter.IsDone(); iter.Next()) delete [] iter.Current();
		delete typedNeighborList;
	}
	if(numberOfHalfNeighbors) delete [] numberOfHalfNeighbors;

}


NeighborLocator *CellNeighborList::Copy() const
{
    CellNeighborList *newNeighborLocator = 
        new CellNeighborList(cutoffDistance,drift);
    return newNeighborLocator;
}


void CellNeighborList::UpdateMyData()
{
	CellNeighborLocator::UpdateMyData();
	MakeNeighborList();
}


bool CellNeighborList::HalfNeighbors(int atomNumber,
									 int *neighbors, 
									 int& numNeighbors,
									 double (*displacements)[DIMENSION], 
									 double* displacementsSquared,
									 int& spacesLeftInArrays)
{
	numNeighbors = 0;
	const int *neighborCandidates = 
		(const int *)(&neighborList[atomNumber][numberOfHalfNeighbors[atomNumber]]);
	const int maxNumberOfNeighbors = 
		neighborList[atomNumber].size()-numberOfHalfNeighbors[atomNumber];

	if(maxNumberOfNeighbors > spacesLeftInArrays) return false;
	
	memcpy(neighbors, neighborCandidates, maxNumberOfNeighbors*sizeof(int));
	const double *center = (*positions)[atomNumber];
	//cout << " " << maxNumberOfNeighbors;
	PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors,positions,
		displacements,displacementsSquared, cutoffDistance2, numNeighbors);
	spacesLeftInArrays -= numNeighbors; 
	return true;	
}

/* PROJECT
bool CellNeighborList::HalfNeighbors(Index atomIndex, Index* neighbors, 
		int& numNeighbors, double (*displacements)[DIMENSION],
		double* displacementsSquared, int& spacesLeftInArrays)
{
	numNeighbors = 0;
	if (!typedNBListBuilt) MakeTypedNeighborList();

	const int globalIndex = partialSum[atomIndex.leaf] + atomIndex.inLeaf;
	const int *neighborCandidates = 
		neighborList[globalIndex].begin() + numberOfHalfNeighbors[globalIndex];
	const int maxNumberOfNeighbors = 
		neighborList[globalIndex].end()-neighborCandidates;

	if(maxNumberOfNeighbors > spacesLeftInArrays) return false;

	const TypedNBListElement *current = (*typedNeighborList)[atomIndex];
	const int myType = atomIndex.leaf;
	const int size = current[myType].size;
	int itmp = 0;
	const int *end = current[myType].begin + size;
	while(neighborCandidates != end)
	{
		neighbors[itmp] = Index(myType, *neighborCandidates - partialSum[myType]);
		++itmp;
		++neighborCandidates;
	}
	for(int type=myType+1; type<nTypes; ++type)
	{
		const int size = current[type].size;
		for(int cand=0; cand<size; ++cand)
			neighbors[itmp+cand] =
				Index(type, *neighborCandidates++ - partialSum[type]);
		itmp += size;
	}
	ASSERT(itmp == maxNumberOfNeighbors);
	ASSERT(neighborCandidates == neighborList[globalIndex].end());
	const double *center = (*positions)[atomIndex];
	PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors,positions,
		displacements,displacementsSquared, cutoffDistance2, numNeighbors);
	spacesLeftInArrays -= numNeighbors; 
	return true;	
}
*/

bool CellNeighborList::Neighbors(int atomNumber,
								 int *neighbors,
								 int& numNeighbors, 
								 double (*displacements)[DIMENSION],
								 double* displacementsSquared,
								 int& spacesLeftInArrays)
{
	numNeighbors = 0;
	int maxNumberOfNeighbors = neighborList[atomNumber].size();

	if(maxNumberOfNeighbors > spacesLeftInArrays) return false;

	int *neighborCandidates = (int *) (&neighborList[atomNumber][0]);
	memcpy(neighbors, neighborCandidates, maxNumberOfNeighbors*sizeof(int));
	const double *center = (*positions)[atomNumber];
	PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors,positions,
		displacements,displacementsSquared, cutoffDistance2, numNeighbors);
	spacesLeftInArrays -= numNeighbors; 
	return true;	
}

bool CellNeighborList::Neighbors(int atomNumber,
								 int type1,
								 int type2,
								 int *neighbors, 
								 int& numNeighbors,
								 double (*displacements)[DIMENSION],
								 double* displacementsSquared,
								 int& spacesLeftInArrays)
{
	numNeighbors = 0;
	if (!typedNBListBuilt) MakeTypedNeighborList();

	const Index ind(type1, atomNumber);
	const TypedNBListElement &thisTypedNBList = (*typedNeighborList)[ind][type2];
	int maxNumberOfNeighbors = thisTypedNBList.size;
	if(maxNumberOfNeighbors > spacesLeftInArrays) return false;

	const int *neighborCandidates = thisTypedNBList.begin;
	for(int i=0; i<maxNumberOfNeighbors; ++i)
		neighbors[i] = neighborCandidates[i] - partialSum[type2];
	const double *center = (*positions)[ind];
	PruneNeighborCandidates(center, maxNumberOfNeighbors, neighbors, 
		positions->GetLeaf(type2), displacements, displacementsSquared, 
		cutoffDistance2, numNeighbors);
	spacesLeftInArrays -= numNeighbors; 
	return true;	
}

bool CellNeighborList::HalfNeighbors(int atomNumber,
						    		 int type1,
							    	 int type2,
								     int *neighbors, 
						    		 int& numNeighbors,
							    	 double (*displacements)[DIMENSION],
								     double* displacementsSquared,
								     int& spacesLeftInArrays)
{
	numNeighbors = 0;
    if(type2 < type1) return true;
	if (!typedNBListBuilt) MakeTypedNeighborList();

	const Index ind(type1, atomNumber);
	const TypedNBListElement &thisTypedNBList = (*typedNeighborList)[ind][type2];
	int maxNumberOfNeighbors = thisTypedNBList.size; // an upper bound
	if(maxNumberOfNeighbors > spacesLeftInArrays) return false;

	int *neighborCandidates = thisTypedNBList.begin;
    if(type1 == type2)
    {
        int tmp=0;
        const int global = atomNumber + partialSum[type1];
        while(*neighborCandidates < global)
        {
            neighborCandidates++;
            tmp++;
        }
        maxNumberOfNeighbors -= tmp;
    }
	for(int i=0; i<maxNumberOfNeighbors; ++i)
		neighbors[i] = neighborCandidates[i] - partialSum[type2];
	const double *center = (*positions)[ind];
	PruneNeighborCandidates(center, maxNumberOfNeighbors, neighbors, 
		positions->GetLeaf(type2), displacements, displacementsSquared, 
		cutoffDistance2, numNeighbors);
	spacesLeftInArrays -= numNeighbors; 
	return true;	
}

void CellNeighborList::MakeNeighborList()   // build a list of pairs with cutList.
{
#ifndef NDEBUG
    int ok = 0, bad = 0;
#endif
	if(neighborList)
	{
		delete [] neighborList;
		delete [] numberOfHalfNeighbors;
	}
	neighborList = new std::vector<int> [nAtoms] ;
	for(int i=0; i<nAtoms; ++i) neighborList[i].reserve(maxNeighbors);
	numberOfHalfNeighbors = new int[nAtoms];

	// Strategy: part 1. for each cell do the following: for each of its atoms,
	// compute the pairs with atoms in the half the neighbor cells 
	if(nextLevelNeighborLocator)
	{
		int nbCell, cellsAround;
		const int maxCellsAround      = 500;
		int slotsLeft                 = maxCellsAround;
		int *neighborCells            = new int[maxCellsAround];
		double (*cellDisp)[DIMENSION] = new double[maxCellsAround][DIMENSION];
		double *cellSqDisp            = new double[maxCellsAround];
		DMArrayIterator<double, DIMENSION> cellPosIter(cellsPositions);
		int cell=0;
		for(cellPosIter.Init(); !cellPosIter.IsDone(); ++cell, cellPosIter.Next())
		{
		    int slotsLeft = maxCellsAround;
  			if(!nextLevelNeighborLocator->HalfNeighbors(cell, neighborCells, 
														cellsAround, cellDisp, cellSqDisp, slotsLeft)){
				throw DMException("CellNeighborList","MakeNeighborList",
								  "Need more cells around");
			}
			PruneNeighborCells(neighborCells, cellsAround, cellDisp, cutoffDistance, true); 
            // 2*drift is added to cutoffDistance inside the function
			//	now we want to know how many atoms are located in the neighbor cells:
			int candidates = 0;
			for(nbCell=0; nbCell < cellsAround; ++nbCell)
				candidates += numberOfAtomsInCell[neighborCells[nbCell] ];
			// then we allocate 3 arrays of indexes, positions and displacements of the candidates
			int *neighbors  =  new int[candidates];
			double (*candidatesPositions)[DIMENSION] = new double[candidates][DIMENSION];
			double (*displacements)[DIMENSION]       = new double[candidates][DIMENSION];
			double (*candPosPtr)[DIMENSION]          = candidatesPositions;
			// now we store these indexes and the positions
			int iTmp = 0;
			for(nbCell=0; nbCell<cellsAround; ++nbCell) 
			{
				int *neighCellList = cellListGlobal[ neighborCells[nbCell] ];
				int  neighCellSize = numberOfAtomsInCell[ neighborCells[nbCell] ];
				memcpy(neighbors+iTmp, neighCellList, neighCellSize*sizeof(int));
				for(int i=0; i<neighCellSize; ++i)
					memcpy(candPosPtr++, (*oldPositions)[neighCellList[i]], DIMENSION*sizeof(double) );
				iTmp += neighCellSize;
			}
			// now we sweep over the atoms in the cell number cell.
			int *tmpCellList = cellListGlobal[cell];
			int nAtomsInThisCell = numberOfAtomsInCell[cell];
			for(int atom=0; atom<nAtomsInThisCell; ++atom)
			{
				int indexOfThisAtom           = tmpCellList[atom];
				vector<int> &nbListOfThisAtom = neighborList[indexOfThisAtom];

				memcpy(displacements, candidatesPositions, candidates*DIMENSION*sizeof(double));
				
				double *positionOfThisAtom = (*oldPositions)[indexOfThisAtom];
				for(int cnt=0; cnt<candidates; ++cnt)
					for(int coord=0; coord<DIMENSION; ++coord)
						displacements[cnt][coord] -= positionOfThisAtom[coord];
				bc->DifferenceBoundaryConditions(displacements, candidates);
				double cut2 = cellSize*cellSize;
				for(int i=0; i<candidates; ++i)
				{
					double dr2 = 0.0;
					for (int coord = 0; coord < DIMENSION; coord++) 
						dr2 += displacements[i][coord] * displacements[i][coord];
					if (dr2 <= cut2)
					{
						int other = neighbors[i];
						if(indexOfThisAtom > other) nbListOfThisAtom.push_back(other);
						else neighborList[other].push_back(indexOfThisAtom);
#ifndef NDEBUG
                        ok++;
					}
					else ++bad;
#else
                    }
#endif
                }
			}
			delete [] neighbors;
			delete [] candidatesPositions;
			delete [] displacements;
		}
		delete [] neighborCells;
		delete [] cellDisp;
		delete [] cellSqDisp;
	}

	// Strategy: part 2. For each cell do the following:
	// compute half the pairs of atoms which are located in this cell
	for(int cell=0; cell<numberOfNonEmptyCells; ++cell)
	{
		int atom;
		int *thisList = cellListGlobal[cell];
		int size = numberOfAtomsInCell[cell];
		// for efficiency we'll store the positions:
		double (*atomsPos)[DIMENSION] = new double[size][DIMENSION];
		for(atom=0; atom<size; ++atom)
			memcpy(atomsPos[atom], (*oldPositions)[thisList[atom]], DIMENSION*sizeof(double));
		double (*displacements)[DIMENSION] = new double[size][DIMENSION];
		for(atom=1; atom<size; ++atom)
		{
			for(int cnt=0; cnt < atom; ++cnt)
				for(int coord=0; coord<DIMENSION; ++coord)
					displacements[cnt][coord] = atomsPos[cnt][coord] - atomsPos[atom][coord];
			bc->DifferenceBoundaryConditions(displacements, atom);
			for(int i=0; i<atom; ++i)
			{
				double dr2 = 0.0;
				for (int coord = 0; coord < DIMENSION; coord++) 
					dr2 += displacements[i][coord] * displacements[i][coord];
				if (dr2 <= cellSize*cellSize)
				{
					int pos1 = thisList[atom];
					int pos2 = thisList[i];
					if(pos1 > pos2) neighborList[pos1].push_back(pos2);
					else neighborList[pos2].push_back(pos1);
#ifndef NDEBUG
                    ok++;
                }
				else ++bad;
#else
                }
#endif
			}
		}
		delete [] atomsPos;
		delete [] displacements;
	}


#ifndef NDEBUG
//cout << "ratio ok = " << ok << "/" << (ok+bad) << " = " << ok/double(ok+bad) << endl;
#endif
    //Finally we symmetrize the neighborList.
	int atom;
	for(atom=0; atom<nAtoms; ++atom) 
		numberOfHalfNeighbors[atom] = neighborList[atom].size();
	for(atom=0; atom<nAtoms; ++atom)
		for(int i=0; i<numberOfHalfNeighbors[atom]; ++i) 
			neighborList[neighborList[atom][i]].push_back(atom);
	++timesRebuilt;
	this->typedNBListBuilt = false;

}


//*************************************************************************

void CellNeighborList::MakeTypedNeighborList()
{
	nTypes = positions->NumberOfLeaves();
	if (nTypes == 0)
		throw DMException("CellNeighborList","MakeTypedNeighborList","No types");

	const Shape &shape = positions->GetShape();
	if(!typedNeighborList) 
		typedNeighborList = new DMArray<TypedNBListElement *> (shape);
	else
	{
		DMArrayIterator<TypedNBListElement *> iter(typedNeighborList);
		for(iter.Init(); !iter.IsDone(); iter.Next()) delete [] iter.Current();
		typedNeighborList->SetShape(shape);
	}

	const vector<int> &leavesSize = positions->GetShape().GetLeavesSize();
	this->partialSum.clear();
	partialSum.push_back(0);
	for(int type=1; type<nTypes; ++type)
		partialSum.push_back(partialSum.back()+leavesSize[type-1]);

	DMArrayIterator<TypedNBListElement *> iter(typedNeighborList);
	int atomCnt = 0;
	for(iter.Init(); !iter.IsDone(); iter.Next())
	{
		vector<int> &currentNL = neighborList[atomCnt];
		sort(currentNL.begin(), currentNL.begin() + numberOfHalfNeighbors[atomCnt]);
		sort(currentNL.begin() + numberOfHalfNeighbors[atomCnt], currentNL.end());
		iter.Current() = new TypedNBListElement[nTypes];
		TypedNBListElement *current = iter.Current();
		int currentLimit = 0;
		int currentType  = 0;
		int size = 0;
		vector<int>::iterator ptr;
		for( ptr=currentNL.begin(); ptr != currentNL.end(); ptr++ )
		{
			int dPtr = *ptr;
			while(dPtr >= currentLimit)
			{
				current[currentType].begin = &dPtr;
				if(currentType != 0) current[currentType-1].size = size;
				currentLimit += leavesSize[currentType];
				++currentType;
				size = 0;
			}
			++size;
		}
		current[currentType-1].size = size;
		while(currentType < nTypes) current[currentType++].size = 0;
#ifndef NDEBUG
		int total  = 0;
		for(int type=0; type<nTypes; ++type) total += current[type].size;
		ASSERT(total == currentNL.size());
		int nbInd = 0;
		for(int ty=0; ty<nTypes; ++ty)
		{
			for(int o=0; o<current[ty].size; ++o)
			{
				ASSERT(current[ty].begin[o] == currentNL[nbInd]);
				ASSERT(current[ty].begin[o] >= partialSum[ty]);
				if(ty < 0) ASSERT(current[ty].begin[o] < partialSum[ty-1]);
				nbInd++;
			}
		}
#endif // NDEBUG
		++atomCnt;
	}
	this->typedNBListBuilt = true;
}

