// SimpleNeighborLocator.cpp: implementation of the SimpleNeighborLocator class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <math.h>
#include "DigitalMaterial.h"

#include "SimpleNeighborLocator.h"
#include "DMArray.h"
#include "DMException.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"

static SerialBuilder<SimpleNeighborLocator> *_simpleNLcreator = 
	new SerialBuilder<SimpleNeighborLocator>("SimpleNeighborLocator");

const std::string SimpleNeighborLocator::className = std::string("SimpleNeighborLocator");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SimpleNeighborLocator::SimpleNeighborLocator(double cutoffDistance): 
NeighborLocator(cutoffDistance)
{
}

SimpleNeighborLocator::~SimpleNeighborLocator()
{

}

void SimpleNeighborLocator::UpdatePosition()
{	
}

bool SimpleNeighborLocator::HalfNeighbors(int atomNumber, 
                                          int *neighbors,
                                          int &numNeighbors,
                                          double (*displacements)[DIMENSION],
                                          double *displacementsSquared,
                                          int &spacesLeftInArrays)
{
    numNeighbors = 0;
    int nAtoms = positions->Size();
    if (nAtoms - atomNumber - 1 > spacesLeftInArrays) return false;
    int count=0;
    // taking the atoms with higher index allows to include the ghost atoms
    for(int j=atomNumber+1;j<nAtoms;j++) neighbors[count++] = j;
    double *center = (*positions)[atomNumber];
    PruneNeighborCandidates(center, count, neighbors, positions,
        displacements, displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors;
    return true;
}

bool SimpleNeighborLocator::Neighbors(int atomNumber,
                                      int *neighbors,
                                      int &numNeighbors,
                                      double (*displacements)[DIMENSION],
                                      double *displacementsSquared,
                                      int &spacesLeftInArrays)
{
    numNeighbors = 0;
    int nAtomsMinus1 = positions->Size() - 1;
    if (nAtomsMinus1 > spacesLeftInArrays) return false;
    
    int j;
    for(j=0;j<atomNumber;j++) neighbors[j] = j;
    for(j=atomNumber+1; j<=nAtomsMinus1;++j) neighbors[j-1] = j;
    double *center = (*positions)[atomNumber];
    PruneNeighborCandidates(center, nAtomsMinus1, neighbors, positions,
        displacements, displacementsSquared, cutoffDistance2, numNeighbors); 
    spacesLeftInArrays -= numNeighbors;
    return true;
}


bool SimpleNeighborLocator::Neighbors(const double *center,
                                      int* neighbors,
                                      int& numNeighbors,
                                      double (*displacements)[DIMENSION],
                                      double* displacementsSquared,
                                      int& spacesLeftInArrays,
                                      double cutoff)
{
    numNeighbors = 0;
    int nAtoms = positions->Size();
    if ( nAtoms > spacesLeftInArrays) return false;

    if(cutoff<0.) cutoff = cutoffDistance;
    double cutoffSq = cutoff*cutoff;

    for(int j=0;j<nAtoms;++j) neighbors[j] = j;
    PruneNeighborCandidates(center, nAtoms, neighbors,positions,
        displacements, displacementsSquared, cutoffSq, numNeighbors); 
    spacesLeftInArrays -= numNeighbors;
    return true;
}


bool SimpleNeighborLocator::Neighbors(int atomNumber,
                                      int type1,
                                      int type2,
                                      int* neighbors,
                                      int& numNeighbors,
                                      double (*displacements)[DIMENSION],
                                      double* displacementsSquared,
                                      int& spacesLeftInArrays)
{
    const DMArray<double, DIMENSION> *posType2 = positions->GetLeaf(type2);
    const int nAtoms2 = posType2->Size();
    if(spacesLeftInArrays < nAtoms2) return false;
    int j, maxNeighbors;
    if(type1 == type2)
    {
        for (j=0 ; j<atomNumber; ++j) neighbors[j] = j ;
        for (j=atomNumber+1 ; j<nAtoms2; ++j) neighbors[j-1] = j ;
        maxNeighbors = nAtoms2 - 1;
    }
    else
    {
        for (j=0 ; j<nAtoms2; ++j) neighbors[j] = j ;
        maxNeighbors = nAtoms2;
    }
    const double *center = (*(positions->GetLeaf(type1)))[atomNumber];
    
    PruneNeighborCandidates(center, maxNeighbors, neighbors, posType2,
        displacements, displacementsSquared, cutoffDistance2, numNeighbors); 
    spacesLeftInArrays -= numNeighbors;
    return true;
}

bool SimpleNeighborLocator::HalfNeighbors(int atomNumber,
                                      int type1,
                                      int type2,
                                      int* neighbors,
                                      int& numNeighbors,
                                      double (*displacements)[DIMENSION],
                                      double* displacementsSquared,
                                      int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(type2 < type1) return true;

    const DMArray<double, DIMENSION> *posType2 = positions->GetLeaf(type2);

    int maxNeighbors;
    if(type2 == type1) maxNeighbors = atomNumber;
    else maxNeighbors = posType2->Size();

    if(spacesLeftInArrays < maxNeighbors) return false;

    for (int i=0 ; i < maxNeighbors; ++i) neighbors[i] = i ;

    const double *center = (*(positions->GetLeaf(type1)))[atomNumber];
    
    PruneNeighborCandidates(center, maxNeighbors, neighbors, posType2,
        displacements, displacementsSquared, cutoffDistance2, numNeighbors); 
    spacesLeftInArrays -= numNeighbors;
    return true;
}


NeighborLocator *SimpleNeighborLocator::Copy() const
{
    SimpleNeighborLocator *newNeighborLocator = 
                new SimpleNeighborLocator(cutoffDistance);
    return newNeighborLocator;
}
