// ColorMethod.cpp: implementation of the ColorMethod class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <string>
#include "DigitalMaterial.h"

#include "ColorMethod.h"
#include "ListOfAtoms.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "BravaisLatticeWithBasis.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ForceColorMethod::ForceColorMethod(Potential *potential)
{
	this->potential = potential;
	this->component = 0;
	this->nAtoms = 0;
	this->forces = 0;
}

ForceColorMethod::~ForceColorMethod()
{
	if(forces)
		delete forces;
}


void ForceColorMethod::Update(ListOfAtoms *atoms)
{
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber(),atoms->GetCartesianPositions());
	potential->CalculateForces(atoms,forces);	
}


void ForceColorMethod::Allocate(int nAtoms, const DMArray<double, DIMENSION> *array)
{
	this->nAtoms = nAtoms;
	if(forces) delete forces;
	forces = new DMArray<double, DIMENSION> (array);
}



CoordinationColorMethod::CoordinationColorMethod()
{
	maxNeighbors = 0;
	displacementsSq = NULL;
	displacements = NULL;
	neighbors = NULL;
	
	AllocateForNeighborLocator(150); // a rough estimate of the expected number of neighbors
}

CoordinationColorMethod::~CoordinationColorMethod()
{
	delete [] neighbors;
	delete [] displacements;
	delete [] displacementsSq;
}

void CoordinationColorMethod::AllocateForNeighborLocator(int maxNeighbors)
{
	this->maxNeighbors = maxNeighbors;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(maxNeighbors > 0)
	{
		neighbors       = new (std::nothrow) int[maxNeighbors];
		displacements   = new (std::nothrow) double[maxNeighbors][DIMENSION];
		displacementsSq = new (std::nothrow) double[maxNeighbors];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("CoordinationColorMethod",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", maxNeighbors));
		if(0 == displacements)
			throw NotEnoughMemory("CoordinationColorMethod",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", maxNeighbors*DIMENSION));
		if(0 == displacementsSq)
			throw NotEnoughMemory("CoordinationColorMethod",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d integers", maxNeighbors));
	}
	else
	{
		neighbors       = NULL;
		displacements   = NULL;
		displacementsSq = NULL;
	}


}


double CoordinationColorMethod::operator()(ListOfAtoms *atoms, int i)
{
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	double cutoff = neighborLocator->GetCutoffDistance();

	int spacesLeftInArrays = maxNeighbors, numNeighbors;
	bool arrayNotTooFull = neighborLocator->Neighbors(i, 
												 neighbors,
												 numNeighbors, 
												 displacements,
												 displacementsSq,
												 spacesLeftInArrays);

	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*maxNeighbors);
		spacesLeftInArrays = maxNeighbors; // maxNeighbors has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i, 
													 neighbors,
													 numNeighbors, 
													 displacements,
													 displacementsSq,
													 spacesLeftInArrays);
	}
		
	ASSERT(arrayNotTooFull);
	if (numNeighbors > 4)
		return -1;
	else
		return numNeighbors;
}




CentroSymmetryColorMethod::CentroSymmetryColorMethod(BravaisLatticeWithBasis *undeformedLattice, int numberNearestNeighbors, DMArray<double, DIMENSION> *nearestNeighborVectors)
{
	this ->	undeformedLattice = undeformedLattice;
	double latticeSpacing = undeformedLattice->GetLatticeSpacing();

	// will have to make this be obtainable from the lattice. For now it is passedin through the constructor
	this->numberNearestNeighbors = numberNearestNeighbors; 

	this->nearestNeighborVectors = new double*[numberNearestNeighbors];
	actualNeighborDisplacements = new double*[numberNearestNeighbors];
	for (int a=0;a<numberNearestNeighbors;++a)
	{
		this->nearestNeighborVectors[a] = new double[DIMENSION];
		actualNeighborDisplacements[a] = new double[DIMENSION];
		for(int coord=0;coord<DIMENSION;++coord)
			this->nearestNeighborVectors[a][coord] = (*nearestNeighborVectors)[a][coord];
	}


	maxNeighbors = 0;
	displacementsSq = NULL;
	displacements = NULL;
	neighbors = NULL;
	
	AllocateForNeighborLocator(150); // a rough estimate of the expected number of neighbors
}

CentroSymmetryColorMethod::~CentroSymmetryColorMethod()
{
	for (int a=0;a<numberNearestNeighbors;++a)
	{
		delete nearestNeighborVectors[a];
		delete actualNeighborDisplacements[a];
	}

	delete nearestNeighborVectors;
	delete actualNeighborDisplacements;

	delete [] neighbors;
	delete [] displacements;
	delete [] displacementsSq;
}

void CentroSymmetryColorMethod::AllocateForNeighborLocator(int maxNeighbors)
{
	this->maxNeighbors = maxNeighbors;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(maxNeighbors > 0)
	{
		neighbors       = new (std::nothrow) int[maxNeighbors];
		displacements   = new (std::nothrow) double[maxNeighbors][DIMENSION];
		displacementsSq = new (std::nothrow) double[maxNeighbors];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("CentroSymmetryColorMethod",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", maxNeighbors));
		if(0 == displacements)
			throw NotEnoughMemory("CentroSymmetryColorMethod",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", maxNeighbors*DIMENSION));
		if(0 == displacementsSq)
			throw NotEnoughMemory("CentroSymmetryColorMethod",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d integers", maxNeighbors));
	}
	else
	{
		neighbors       = NULL;
		displacements   = NULL;
		displacementsSq = NULL;
	}


}


double CentroSymmetryColorMethod::operator()(ListOfAtoms *atoms, int i)
{
	int coord, vecIdx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	double cutoff = neighborLocator->GetCutoffDistance();
	double position[DIMENSION], neighborPosition[DIMENSION];

	vector<int> neighborIndices;

	for(coord=0;coord<DIMENSION;++coord)
		position[coord] = atoms->GetCartesianPosition(i,coord);
	neighborIndices.clear();
	
	// add the nearNeighbor vectors to the current position in turn and 
	// find the atom whose current position is closest to the resulting position 
	for (vecIdx=0;vecIdx<numberNearestNeighbors;++vecIdx)
	{
		int spacesLeftInArrays = maxNeighbors;
		for(coord=0;coord<DIMENSION;++coord)
			neighborPosition[coord] = position[coord] + nearestNeighborVectors[vecIdx][coord];
		int numNeighbors;
		bool arrayNotTooFull = neighborLocator->Neighbors(neighborPosition,
														  neighbors,
														  numNeighbors,
														  displacements,
														  displacementsSq,
														  spacesLeftInArrays,
														  cutoff);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*maxNeighbors);
			spacesLeftInArrays = maxNeighbors; // maxNeighbors has been doubled
			arrayNotTooFull = neighborLocator->Neighbors(neighborPosition, 
														 neighbors,
														 numNeighbors, 
														 displacements,
														 displacementsSq,
														 spacesLeftInArrays,
														 cutoff);
		}


		if(numNeighbors == 0)
		{
			cout << "zero neighbors within " << 2*cutoff << " found for neighbor " << vecIdx << " of atom " << i << endl;
			cout << "position is " << atoms->GetCartesianPosition(i,0) << " " << atoms->GetCartesianPosition(i,1) << endl;
		}
		ASSERT(numNeighbors > 0);
		// get the nearest atom
		int nearest, nbIdx;
		double nearestDispSq = 1.e99;
		for(nbIdx=0;nbIdx<numNeighbors;++nbIdx)
		{
			if(displacementsSq[nbIdx] < nearestDispSq)
			{
				nearest = nbIdx;
				nearestDispSq = displacementsSq[nbIdx];
			}
		}
			neighborIndices.push_back(nearest);
			for(coord=0;coord<DIMENSION;++coord)
				actualNeighborDisplacements[vecIdx][coord] = displacements[nearest][coord];
	} // end loop over nearest neighbors (vecIdx)
	
	
	// now have the displacements to be added
	
	double deviation = 0.;
	ASSERT (numberNearestNeighbors%2 == 0);
	
	for(vecIdx = 0; vecIdx < numberNearestNeighbors/2; ++vecIdx)
	{
		for(coord=0;coord<DIMENSION;++coord)
		{
			double sum = actualNeighborDisplacements[vecIdx][coord] + actualNeighborDisplacements[vecIdx + numberNearestNeighbors/2][coord];
			deviation += sum * sum;
		}
	}
	
	return deviation;

}


