// ColorMethod.h: interface for the ColorMethod class.
//
//////////////////////////////////////////////////////////////////////

#ifndef COLORMETHOD_H
#define COLORMETHOD_H

#include "DMArray.h"
#include "Potential.h"
class ListOfAtoms;
class BravaisLatticeWithBasis;

// THIS FILE IS SWIGGED IN Observers.i
class ColorMethod
{
public:
	virtual double operator()(ListOfAtoms *atoms, int i)=0;
	virtual void Update(ListOfAtoms *atoms){}
};

class EnergyColorMethod : public ColorMethod
{
public:
	EnergyColorMethod(Potential *potential, double offset=0.)
		{this->potential = potential; this->offset = offset;}
    double operator()(ListOfAtoms *atoms, int i)
		{return potential->CalculateAtomicEnergy(atoms,i)-offset;}
	void SetOffset(double offset) {this->offset = offset;}
	double GetOffset() {return offset;}
	// Add an update function to calculate all the energies at once?
private:
	Potential *potential;
	double offset;
};

class StressColorMethod : public ColorMethod
{
public:
	StressColorMethod(Potential *potential)
		{this->potential = potential;this->component = 0;}
	void SetComponent(int comp){ASSERT(component < (DIMENSION+1)*DIMENSION/2); this->component = comp;}
		
	double operator()(ListOfAtoms *atoms, int i)
		{
			double stress[(DIMENSION+1)*DIMENSION/2];
			potential->CalculateStressTensor(atoms,i,stress);
			double trace = 0.;
			for(int coord=0;coord<DIMENSION;++coord) trace += stress[coord];
			// first DIMENSION elements of the stress vector are the diagonal elements of the stress tensor.
			return (stress[1]);
		}

    
private:
	int component;
	Potential *potential;
};


class ForceColorMethod : public ColorMethod
{
public:
	ForceColorMethod(Potential *potential);
	~ForceColorMethod();
	void SetComponent(int comp){ASSERT(component < DIMENSION); this->component = comp;}		
	double operator()(ListOfAtoms *atoms, int i)
		{
			return ((*forces)[i][component]);
		}   
	void Update(ListOfAtoms *atoms);
private:
	void Allocate(int nAtoms, const DMArray<double, DIMENSION> *array);
	int nAtoms;
	int component;
	DMArray<double,DIMENSION> *forces;
	Potential *potential;
};


class CoordinationColorMethod : public ColorMethod
{
public:
	CoordinationColorMethod();
	~CoordinationColorMethod();
	double operator()(ListOfAtoms *atoms, int i);

	//void Update(ListOfAtoms *atoms);
private:
	void AllocateForNeighborLocator(int maxNeighbors);
	int nAtoms;
	int maxNeighbors;
	int numberNearestNeighbors; 
	int *neighbors;
	double (*displacements)[DIMENSION];
	double *displacementsSq;
	double cutoffDist;
};


class CentroSymmetryColorMethod : public ColorMethod
{
public:
	CentroSymmetryColorMethod(BravaisLatticeWithBasis *undeformedLattice, int numberNearestNeighbors, DMArray<double, DIMENSION> *nearestNeighborVectors);
	~CentroSymmetryColorMethod();
	double operator()(ListOfAtoms *atoms, int i);

	//void Update(ListOfAtoms *atoms);
private:
	void AllocateForNeighborLocator(int maxNeighbors);
	int nAtoms;
	int maxNeighbors;
	int numberNearestNeighbors; 
	double **nearestNeighborVectors;
	double **actualNeighborDisplacements;
  	const BravaisLatticeWithBasis *undeformedLattice;
	int *neighbors;
	double (*displacements)[DIMENSION];
	double *displacementsSq;
};


#endif // !defined(COLORMETHOD_H)
