// StressIntensityObserver.cpp: implementation of the StressIntensityObserver class.
//
//////////////////////////////////////////////////////////////////////
#include "DMFirstHeader.h"
#include <fstream>
#include <map>
#include "DigitalMaterial.h"
#include "StressIntensityObserver.h"
#include "Potential.h"
#include "NeighborLocator.h"
#include "BravaisLatticeWithBasis.h"
#include "ListOfAtoms.h"
#include "NotchFieldDisplacer.h"
#include "TinyVector.h"
#include "DMArray.h"
#include "DMLastHeader.h"

using namespace std;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

StressIntensityObserver::StressIntensityObserver(Potential *potential, NeighborLocator *neighborLocator, BravaisLatticeWithBasis *lattice, NotchFieldDisplacer *complimentaryField, double pathCenter[DIMENSION], double pathRadius, double heightTolerance, const DMArray<double, DIMENSION> *undeformedPositions, int notifyLevel) : ListOfAtomsObserver(notifyLevel)
{
	this->potential = potential;
	ASSERT(potential);
	this->neighborLocator = neighborLocator;
	ASSERT(neighborLocator);
	this->lattice = lattice;
	ASSERT(lattice);
	this->complimentaryField = complimentaryField;
	ASSERT(complimentaryField);
	for(int coord=0;coord<DIMENSION;++coord)
		this->pathCenter[coord] = pathCenter[coord];
    this->pathRadius = pathRadius;	
	this->heightTolerance = heightTolerance;
	ASSERT(heightTolerance > 0.);
	this->undeformedPositions = undeformedPositions;
	ASSERT(undeformedPositions);

	H_integral = 0.;

	// the following is borrowed from Potential.h. It's a temporary means to an end
#if DIMENSION==2
	stresscomp[0][0] = 0; stresscomp[1][1] = 1; stresscomp[0][1] = 2; stresscomp[1][0] = 2;
#else // if DIMENSION==3
	stresscomp[0][0] = 0; stresscomp[1][1] = 1; stresscomp[2][2] = 2; 
	stresscomp[2][1] = stresscomp[1][2] = 3; stresscomp[2][0] = stresscomp[0][2] = 4; 
	stresscomp[1][0] = stresscomp[0][1] = 5;
#endif
}

StressIntensityObserver::~StressIntensityObserver()
{

}

void StressIntensityObserver::Update(ListOfAtoms *atoms)
{
	ASSERT(DIMENSION > 1);
	int coord;

	// for the given list of atoms, find a path over which to integrate, namely a list of atoms forming a path from one side of the notch to the other. The path will lie in a plane parallel to the xy plane, for now.

	// start by considering the circle of that radius, take points on it spaced by one half of the lattice spacing. For each point find the nearst atom to that point. If there's no atom, then the point is in the notch. Check if the atom has already been added to the list

	double spacing = lattice->GetLatticeSpacing();
	double deltaTheta = 0.5 * spacing/pathRadius;
	int numNeighbors, spacesLeftInArrays = atoms->GetNumber(), nearestAtom;
	int *neighbors = new int[spacesLeftInArrays];
	double (*displacements)[DIMENSION] = new double[spacesLeftInArrays][DIMENSION];
	double *displacementsSq = new double[spacesLeftInArrays];
	std::map<int, pair<int, TinyVector<double> > > contourPoints;
	std::set<int> atomNumbers;

	// assuming here that the notch is on the left, otherwise need to think more about how to know where to start the contour.

	fstream pathPoints("pathPoints.txt",ios::out);
	pathPoints.precision(10);

	// later can have the user input start and end values of theta, based on the particular notch geometry
	int atomIndex = 0;

	for(double theta=-PI;theta<PI;theta+=deltaTheta)
	{
		double point[DIMENSION];
		point[0] = pathCenter[0] + pathRadius * cos(theta);
		point[1] = pathCenter[1] + pathRadius * sin(theta);
		point[2] = pathCenter[2];

		spacesLeftInArrays = atoms->GetNumber();

		bool arraysNotTooFull = neighborLocator->Neighbors(point, neighbors, numNeighbors,displacements, displacementsSq, spacesLeftInArrays, 0.5*spacing);// -1 means use stored cutoff
		ASSERT(arraysNotTooFull);
		// Get the nearest atom to the given point on the circle
		// (if there are any within the cut and on the right plane)
		bool haveAnAtom = false;
		double nearestDispSq = 1.e99;
		for(int t=0;t<numNeighbors;++t)
			{
				if( fabs(displacements[t][2]) < heightTolerance && displacementsSq[t] < nearestDispSq)
				{
					haveAnAtom = true; // ie at least one atom
					nearestAtom = neighbors[t];
					nearestDispSq = displacementsSq[t];
				}
			}
		if( haveAnAtom )
		{
			if(atomNumbers.insert(nearestAtom).second) // ie, if it was indeed inserted, and not already there
			{
				double *position = (*atoms->GetCartesianPositions())[nearestAtom];
				pair<int, TinyVector<double> > atomInfo(nearestAtom, position);
				contourPoints.insert(pair<int, pair<int, TinyVector<double> > >(atomIndex, atomInfo));
				++atomIndex;
			}
		} // end of if haveAnAtom
	} // end of loop over angles finding atoms for the path

	std::map<int, pair<int,TinyVector<double> > >::iterator contourIter;
	if( contourPoints.size() == 0 )
		throw DMException("StressIntensityObserver","Update","Could not find any atoms close to the path");

	for(contourIter = contourPoints.begin();contourIter != contourPoints.end();contourIter++)	
	{
		const double *position;
		position = (*contourIter).second.second;
		pathPoints << position[0] << " " << position[1] << " " << position[2] << endl;
	}


	// need for each atom its position, its stress, its displacement. Iterate over contourPoints

	// calculate the integral

	double stress[(DIMENSION+1)*DIMENSION/2], compStress[(DIMENSION+1)*DIMENSION/2];
	double displacement[DIMENSION], compDisplacement[DIMENSION];
	H_integral = 0.;
	double lastPosition[DIMENSION], lastIntegrand[DIMENSION], tangentVector[DIMENSION];

	int count = 0;
	for(contourIter = contourPoints.begin();contourIter != contourPoints.end();contourIter++)
	{
		int atomNumber = (*contourIter).second.first;
		double *position, normalVector[DIMENSION], arcLength;
		position = (*contourIter).second.second;
		potential->CalculateStressTensor(atoms, atomNumber, stress);

		for(coord=0;coord<DIMENSION;++coord)
			displacement[coord] = position[coord] - (*undeformedPositions)[atomNumber][coord];

			
		// calculate complimentary fields
		for(coord=0;coord<DIMENSION;++coord)
			compDisplacement[coord] = 0.;
		complimentaryField->ElasticDisplacement(position, compDisplacement);
		complimentaryField->ElasticStress(position, compStress);
		

		// accumulate the integral. Calculate the part that depends only on fields at the current point		
		double integrand[DIMENSION];
		for(coord = 0; coord < DIMENSION; ++coord )
			integrand[coord]  = 0.;

		for(int alpha=0;alpha<DIMENSION;++alpha)
			for(int beta=0;beta<DIMENSION;++beta)
			{
				int stsIdx = stresscomp[alpha][beta];
				integrand[beta] += ( stress[stsIdx] * compDisplacement[alpha]
								- compStress[stsIdx] * displacement[alpha] );
			}

		// for the actual integral, need to consider the vectors joining 
		// successive atoms. Break the integral into contributions from each
		// vector, using the average field values from the two ends to make 
		// the integrand
		if(contourIter != contourPoints.begin()) // ie if not the first point
		{
			arcLength = 0.;		
			double dotProduct = 0.;
			for(coord=0;coord<DIMENSION;++coord)
			{
				tangentVector[coord] = position[coord] - lastPosition[coord];
				arcLength += tangentVector[coord] * tangentVector[coord];
				// the following is a good guess for the direction of normalVector
				normalVector[coord] = position[coord] - pathCenter[coord];
				dotProduct += normalVector[coord] * tangentVector[coord];
			}
			// need to make normalVector orthogonal to tangentVecotr and make it of unit length
			double invLengthNormal = 0.;
			for(coord=0;coord<DIMENSION;++coord)
			{
				normalVector[coord] -=  dotProduct * tangentVector[coord]/ arcLength; // at this stage arcLength is the square of the lenght of tangentVector
				invLengthNormal += normalVector[coord] * normalVector[coord];
			}
			
			arcLength = sqrt(arcLength);
			invLengthNormal = 1./sqrt(invLengthNormal);
			for(coord=0;coord<DIMENSION;++coord)
				normalVector[coord] *= invLengthNormal;
			

			// this is bascially the trapezoid rule
			for( coord = 0; coord < DIMENSION; ++coord )
				H_integral += 0.5 * (lastIntegrand[coord] + integrand[coord]) * normalVector[coord] * arcLength;
			
		} // end if (contourIter != contourPoints.begin())
		
		for( coord = 0; coord < DIMENSION; ++coord )
		{
			lastIntegrand[coord] = integrand[coord];
			lastPosition[coord] = position[coord];
		}
	} // end loop over atoms in the contour	
	
	delete [] neighbors;
	delete [] displacementsSq;
	delete [] displacements;
}


void StressIntensityObserver::SetPotential(Potential *potential)
{
	this->potential = potential;
}

void StressIntensityObserver::SetNeighborLocator(NeighborLocator *neighborLocator)
{
	this->neighborLocator = neighborLocator;	
}

void StressIntensityObserver::SetLattice(BravaisLatticeWithBasis *lattice)
{
	this->lattice = lattice;
}

void StressIntensityObserver::SetPathCenter(double pathCenter[DIMENSION])
{
	for(int coord=0;coord<DIMENSION;++coord)
		this->pathCenter[coord] = pathCenter[coord];
}


void StressIntensityObserver::SetPathRadius(double pathRadius)
{
	this->pathRadius = pathRadius;
}
