// EMTPotential.h: interface for the EMTPotential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef EMTPOTENTIAL_H
#define EMTPOTENTIAL_H

// Jakob Schiotz' FAST code, converted to the Digital Material Format
// We (Sethna, Cheong, and Bailey) have tried to leave the code in it's 
// original format as much as we could

// Currently we only support one species (n=1)


#include "DMFirstHeader.h"
#include <math.h>
#include <assert.h>
#include <string>
#include <stdio.h>
#include "Potential.h"
#include "DMArray.h"
#include "BoundaryConditions.h"

// XXX We'd like to avoid global #define constants. It's better to make them 
// constants for the class somehow (using the enum hack, or otherwise: see 
// "Thinking in C++" by Bruce Eckel, Prentice Hall, 1995, ISBN #0-13-917709-4
// Jakob?

#if DIMENSION == 3

#ifndef SWIG
#define BETA 1.809			// Geometric constant
#define MAXELEMENTS 1		// Max. number of different elements (Jakob 3)
#define ARRLIM 1500			// Size of array for pair force calculation
#define MAXEMTNAMELENG 8	// Max. length of element name
#define MAXTMP 12				// Max number of temporary arrays
#endif // SWIG

class EMTPotential : public Potential
{
// interface overridden from Potential (ajd)
public:
    EMTPotential(std::string element = "Cu");
	virtual ~EMTPotential();
	virtual void CalculateForces (ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateEnergy(ListOfAtoms *atoms);
	double CalculateEnergyAndForces (ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);
	void CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2]);
	void CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress);
	// Energy must be calculated first
	// Additional methods to return cutdist and cutlist, so that NeighborList
	// can be initialized

// interface overridden from Serializable
    void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}


// interface special to EMTPotential (ajd)
public:
	void CalculateSigmas(ListOfAtoms *atoms, const bool forcesOnly=false);
	double CalculateEnergyAfterSigmas(ListOfAtoms *atoms);	
	// Sigmas must be calculated first
	virtual void CalculateForcesAfterSigmas
		(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	// Additional methods to return cutdist and cutlist, so that NeighborList
	// can be initialized
	double GetCutoffList() { return cut.cutlist; }
    double GetLengthScale() const;
    double GetEnergyCurvatureScale() const;
protected:
    std::string element;
	const ListOfAtoms *atoms;             // The atoms we are working on
	void Allocate(const ListOfAtoms *atoms);
    int nAtoms;
    // Number of atoms (stored in ListOfAtoms: copied here for storage 
    // allocation)

	//! Density \a sigma1 in EMT potential (one element)
    DMArray<double> *sigma1;
	//! Density \a sigma1 in EMT potential (one element)
	DMArray<double> *sigma2;
	//! EMT: Cohesive Energy
    DMArray<double> *Ec;
	//! EMT: Atomic-sphere correction
    DMArray<double> *Eas;
	//! EMT: Neutral-sphere radius
    DMArray<double> *radius;
	//! EMT: Intermediate result (passed from Energy to Force!)
    DMArray<double> *dE0ds;
	DMArray<double> *dE0dsExtended;
  
    double *dTmp[MAXTMP];		// Temporary double precision storage
    float *fTmp[2*MAXTMP];		// do. single precision. *** dTmp, fTmp, and iTmp
    int *iTmp[2*MAXTMP];			// do. integer.			 *** overlayed in memory 

	// XX The emt_struct should be moved into the Type member for ListOfAtoms
    struct emt_struct;

    struct cut_struct
    {
		void setcut(emt_struct *e, int n);

		int shells[2];
		double cutslope, cutdist, cutlist;
    };

    struct emt_struct
    {
		emt_struct(const char *element);
		void CalculateGamma(cut_struct cut);
			
		double e0, seq, neq, V0, eta2, kappa, lambda, mass, invmass;
		double gamma1, gamma2;
		double pairA, pairD;
		int Z;
		char name[MAXEMTNAMELENG];
    };

    emt_struct *emt;
    cut_struct cut;
private:
	static const std::string className;

};

#endif // DIMENSION == 3

#endif // !defined(EMTPOTENTIAL_H)

