// MEAMPotential.cpp: implementation of the MEAMPotential class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"
#include <iostream>
#include "MEAMPotential.h"

#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"
#include "BoundaryConditions.h"

#include "DMLastHeader.h"

using namespace std;

#if DIMENSION == 3


static SerialBuilder<MEAMPotential> *_MEAMcreator = 
    new SerialBuilder<MEAMPotential>("MEAMPotential");

const string MEAMPotential::className = string("MEAMPotential");


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


MEAMPotential::MEAMPotential(MEAM_ParameterList *paramList) : TINY(1.0e-13)
{
	if( paramList )
	{
		rCut = paramList->rCut;
		SetCutoffDistance(rCut);
		deltaRcut = paramList->deltaRcut;
		scrnCmin = paramList->scrnCmin ;
		scrnCmax = paramList->scrnCmax;
		
		
		MEAM_E0 = paramList->E0;
		MEAM_invR0 = 1. / paramList->R0;
		MEAM_alpha = paramList->alpha;
		MEAM_A = paramList->A;
		for( int a = 0; a < 4; ++a )
		{
			MEAM_beta[a] = paramList->beta[a];
			MEAM_t[a] = paramList->t[a];
		}
		MEAM_Zi = paramList->Zi;
		referenceLattice = paramList->referenceLattice;
	}
	else
	{
		// defaults values are for Cu
		rCut = 4.0;
		SetCutoffDistance(rCut);
		deltaRcut = 0.1;
		scrnCmin = 2.0;
		scrnCmax = 2.8;
		
		
		MEAM_E0 = 3.540;
		MEAM_invR0 = 1. / 2.56;
		MEAM_alpha = 5.11;
		MEAM_A = 1.07;
		MEAM_beta[0] = 3.63;
		MEAM_beta[1] = 2.2;
		MEAM_beta[2] = 6.0;
		MEAM_beta[3] = 2.2;
		MEAM_t[0] = 1. ;
		MEAM_t[1] = 3.14;
		MEAM_t[2] = 2.49;
		MEAM_t[3] = 2.95;
		MEAM_Zi =  12;
	}


	nPairs = 0;
	displacementsSq = 0;
	displacements = 0;
	neighbors = 0;
	screeningFactors = 0;
	screeningDerivs = 0;
	pairDistance = 0;
	atomicElecDensity = 0;
	kx_scrnDer_over_Sijk = 0;
	kx_scrn_contrib = 0;
	forcePtr = 0;
	AllocateForNeighborLocator(50); // a rough estimate of the expected number of neighbors
	CalculateDensityScale();
}

MEAMPotential::~MEAMPotential()
{
	if(neighbors)
		FreeArrays();
}

void MEAMPotential::FreeArrays()
{
	delete [] neighbors;
	delete [] displacements;
	delete [] displacementsSq;
	delete [] screeningFactors;
	delete [] screeningDerivs;
	delete [] pairDistance;
	delete [] atomicElecDensity;
	delete [] kx_scrnDer_over_Sijk;
	delete [] kx_scrn_contrib;
	delete [] forcePtr;
}


void MEAMPotential::AllocateForNeighborLocator(int nPairs)
{
	// nPairs is playing the role of EMTPotential's arrayLim here
	this->nPairs = nPairs;

	if(neighbors) 
		FreeArrays();

	if(nPairs > 0)
	{
		neighbors       = new (std::nothrow) int[nPairs];
		displacements   = new (std::nothrow) double[nPairs][DIMENSION];
		displacementsSq = new (std::nothrow) double[nPairs];
		screeningFactors = new (std::nothrow) double[nPairs];
		screeningDerivs = new (std::nothrow) double[nPairs * (nPairs + 1)][DIMENSION];
		pairDistance = new (std::nothrow) double[nPairs];
		atomicElecDensity = new (std::nothrow) double[nPairs][4];
		kx_scrnDer_over_Sijk = new (std::nothrow) double[nPairs];
		kx_scrn_contrib = new (std::nothrow) bool[nPairs];
		forcePtr = new (std::nothrow) double*[nPairs];


		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		ASSERT(screeningFactors);
		ASSERT(pairDistance);
		ASSERT(atomicElecDensity);
		ASSERT(kx_scrnDer_over_Sijk);
		ASSERT(kx_scrn_contrib);
		ASSERT(forcePtr);

		if(0 == neighbors)
			throw NotEnoughMemory("MEAMPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", nPairs));
		if(0 == displacements)
			throw NotEnoughMemory("MEAMPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", nPairs*DIMENSION));
		if(0 == displacementsSq)
			throw NotEnoughMemory("MEAMPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d integers", nPairs));
		if(0 == screeningFactors)
			throw NotEnoughMemory("MEAMPotential",
								  "AllocateForNeighborLocator",
								  SprintString("Problem allocating %d doubles", nPairs));
		if(0 == pairDistance)
			throw NotEnoughMemory("MEAMPotential",
								  "AllocateForNeighborLocator",
								  SprintString("Problem allocating %d doubles", nPairs));
		
		if(0 == atomicElecDensity)
			throw NotEnoughMemory("MEAMPotential",
								  "AllocateForNeighborLocator",
								  SprintString("Problem allocating %d * 4 doubles", nPairs));
		if(0 == kx_scrnDer_over_Sijk)
			throw NotEnoughMemory("MEAMPotential",
								  "AllocateForNeighborLocator",
								  SprintString("Problem allocating %d doubles", nPairs));
		if(0 == kx_scrn_contrib)
			throw NotEnoughMemory("MEAMPotential",
								  "AllocateForNeighborLocator",
								  SprintString("Problem allocating %d doubles", nPairs));
		if(0 == forcePtr)
			throw NotEnoughMemory("MEAMPotential",
								  "AllocateForNeighborLocator",
								  SprintString("Problem allocating %d double*s", nPairs));
	}
	else
	{
		neighbors       = 0;
		displacements   = 0;
		displacementsSq = 0;
		screeningFactors = 0;
		pairDistance = 0;
		atomicElecDensity = 0;
		kx_scrnDer_over_Sijk = 0;
		kx_scrn_contrib = 0;
		forcePtr = 0;
	}
}


void MEAMPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	int i,coord,jx, kx, j, alpha, beta, gamma, p;
	double *forceI, dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom;
	double  invCoordination = 1./MEAM_Zi;

	forces->Zero();

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	int numberOfNeighbors, nAtoms = atoms->GetNumber();
	for( i = 0; i < nAtoms; ++i )
    {	
		forceI = (*forces)[i];

		// Get the neighbors of atom i
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->Neighbors(i, neighbors,
			numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,
				numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}

		// get screening functions, their derivatives, and atomic densities
		ScreeningAndAtomicDensities(numberOfNeighbors);
				
		// set up pointers to correct locations in force array for neighbors
		for( jx = 0; jx < numberOfNeighbors; ++jx )
		{
			j = neighbors[jx];
			if(j<nAtoms) forcePtr[jx] = (*forces)[j];
			else forcePtr[jx] = dummyForce;
		}
		

		// Do the pair part of the potential first.
		for( jx = 0; jx < numberOfNeighbors; ++jx )
		{
			if(screeningFactors[jx] > 0.0)
			{
				double forceFactorPair;
				double pairEnergy = CalculatePairInteraction(jx, &forceFactorPair);
				for (coord=0; coord<DIMENSION; coord++)
				{
					forceI[coord] += -displacements[jx][coord]*forceFactorPair
						- pairEnergy * screeningDerivs[jx * (numberOfNeighbors+1)][coord] / screeningFactors[jx];
					// where we added the contribution from screening factors
					// the contribution to the force on j will be added
					// when looping over all neighbors kx a few lines hence 
					forcePtr[jx][coord] += +displacements[jx][coord]*forceFactorPair;
					// because the screening factors are three-body terms the
					// other atoms involved get forces (so this isn't really
					// a true pair term).
					for( kx = 0; kx < numberOfNeighbors; ++kx )
						forcePtr[kx][coord] += -pairEnergy * screeningDerivs[jx * (numberOfNeighbors+1) + kx + 1][coord] / screeningFactors[jx];
				} // end coord loop
			} // end of if(screeningFactors[jx] > 0.0)
		} // end loop over jx
	
		// initialize density 
		double elecDensity = 0., partialDensitySq[4] = {0.,0.,0.,0.};
		for( jx = 0; jx < numberOfNeighbors; ++jx )
		{
			for( p = 0; p < 4; ++p )
				atomicElecDensity[jx][p] *= screeningFactors[jx];

			partialDensitySq[0] +=  atomicElecDensity[jx][0];
		}

		double density1factor[DIMENSION] = {0.,0.,0.,};		
		double density2term1factor[DIMENSION][DIMENSION] = {{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}};
		double density2term2 = 0.;
		double density3factor[DIMENSION][DIMENSION][DIMENSION] = {{{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}},{{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}},{{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}}};
		CalculatePartialDensities123(numberOfNeighbors, partialDensitySq, density1factor, density2term1factor, &density2term2, density3factor);



		double angDensitySq = 0.0;
		for( p = 1; p < 4; ++p )
			angDensitySq += MEAM_t[p] * partialDensitySq[p];
		
		// note that the Sq is inappropriate for the zeroth one, because
		// this was never squared, and is assumed to be unsquared 
		// by AddPartialDensities()
		double DFDPartialDensitySq[4], DDensityDAngularPartSq;
		elecDensity = AddPartialDensities(partialDensitySq[0], angDensitySq, DFDPartialDensitySq, &DDensityDAngularPartSq);
		
		double scaledDensity = elecDensity / zBar;
		//double scaledDensity = elecDensity * invCoordination;
		double DFDDensity = MEAM_A * MEAM_E0 * (1. + log(scaledDensity)) / zBar; // invCoordination is the derivative of scaledDensity wrt density
		//double DFDDensity = MEAM_A * MEAM_E0 * (1. + log(scaledDensity)) * invCoordination; // invCoordination is the derivative of scaledDensity wrt density

		// the zeroth derivative was filled in by AddPartialDensities()
		// The rest are got from the deriv wrt angular part sq
		for( p = 1; p < 4; ++p )
			DFDPartialDensitySq[p] = DDensityDAngularPartSq * MEAM_t[p];

		// multiply all four derivatives by DFDDensity  
		for( p = 0; p < 4; ++p )
			DFDPartialDensitySq[p] *= DFDDensity;
	   
		
		// now go through all the loops to get the actual forces
		for( jx = 0; jx < numberOfNeighbors; ++jx )		
			if(screeningFactors[jx] > 0.0)
			{
				double invPairDist = 1. / pairDistance[jx];
				
				// forces from density 0
				double forceFactor0 = DFDPartialDensitySq[0] * atomicElecDensity[jx][0] * MEAM_beta[0] * MEAM_invR0  * invPairDist;
				double forceFactor0b = DFDPartialDensitySq[0] * atomicElecDensity[jx][0] / screeningFactors[jx];

				for( coord = 0; coord < DIMENSION; ++coord )
				{
					forceI[coord] += - forceFactor0 * displacements[jx][coord]
						- forceFactor0b * screeningDerivs[jx * (numberOfNeighbors+1)][coord];
					// screening deriv term for atom j included in kx loop
					forcePtr[jx][coord] += forceFactor0 * displacements[jx][coord] ;
					
					for( kx = 0; kx < numberOfNeighbors; ++kx )
						forcePtr[kx][coord] += - forceFactor0b * screeningDerivs[jx * (numberOfNeighbors + 1) + kx + 1][coord];
				}	
				
				
				// forces from density 1
				for( alpha = 0; alpha < DIMENSION; ++alpha )
				{
					for( coord = 0; coord < DIMENSION; ++coord )
					{
						double forceFactor1 = -DFDPartialDensitySq[1] * 2 * density1factor[alpha] 
							* atomicElecDensity[jx][1]
							* ( ( alpha == coord ? 1. : 0.) * invPairDist
								- ( displacements[jx][alpha] * displacements[jx][coord] / displacementsSq[jx] )
								* (1/pairDistance[jx] + MEAM_beta[1] * MEAM_invR0)
								);
						forceI[coord] += -forceFactor1 
							- DFDPartialDensitySq[1] * 2 
							* density1factor[alpha]
							* displacements[jx][alpha]
							* atomicElecDensity[jx][1]
							* screeningDerivs[jx * (numberOfNeighbors+1)][coord] / (screeningFactors[jx] * pairDistance[jx]); 
						
						forcePtr[jx][coord] += forceFactor1;
						
						for( kx = 0; kx < numberOfNeighbors; ++kx )
							forcePtr[kx][coord] += - DFDPartialDensitySq[1]
								* 2 * density1factor[alpha] 
								* displacements[jx][alpha]
								* atomicElecDensity[jx][1]
								* screeningDerivs[jx * (numberOfNeighbors+1) + kx + 1][coord] / (screeningFactors[jx] * pairDistance[jx]); 
					} // end coord loop
				} // end loop over alpha 
			   	

				// forces from density 2, first term:
				for( alpha = 0; alpha < DIMENSION; ++alpha )
					for( beta = 0; beta < DIMENSION; ++beta )
					{
						// minus from the definition of force
						double forceFactor2 =  -DFDPartialDensitySq[2] * 2 * density2term1factor[alpha][beta] * atomicElecDensity[jx][2] / displacementsSq[jx];
						
						for( coord = 0; coord < DIMENSION; ++coord )
						{
							double forceFactor2a = forceFactor2 * (
								( alpha == coord ? displacements[jx][beta]  : 0.) +
								(  beta == coord ? displacements[jx][alpha] : 0.) 
								- displacements[jx][alpha] * displacements[jx][beta] * displacements[jx][coord] 
								* ( 2. / displacementsSq[jx] + MEAM_beta[2] * MEAM_invR0 * invPairDist)
								);
							
							forceI[coord] += -forceFactor2a + (forceFactor2
															   * displacements[jx][alpha] 
															   * displacements[jx][beta] 
															   * screeningDerivs[jx* (numberOfNeighbors+1)][coord]
															   )/ screeningFactors[jx];
							
							forcePtr[jx][coord] += forceFactor2a;
							
							for( kx = 0; kx < numberOfNeighbors; ++kx )
								forcePtr[kx][coord] += (forceFactor2
														* displacements[jx][alpha]
														* displacements[jx][beta]
														* screeningDerivs[jx * (numberOfNeighbors+1) + kx + 1][coord]) / screeningFactors[jx];
						}  // end coord loop
					} // end alpha-beta loop, density 2
		
				
				// forces from density 2, second term: one minus from the
				// definition of force, one from the sign of this term in
				// the energy, one from the derivative of the atomic density, 
				// giving the sign for the force on J

				double forceFactor2b = -DFDPartialDensitySq[2] * (2./3.) * density2term2 * atomicElecDensity[jx][2] * MEAM_beta[2] * MEAM_invR0 * invPairDist;
				for( coord = 0; coord < DIMENSION; ++coord )
				{
					forceI[coord] += - forceFactor2b * displacements[jx][coord]
						+ DFDPartialDensitySq[2] * (2./3.) * density2term2 
						* atomicElecDensity[jx][2]
						* screeningDerivs[jx * (numberOfNeighbors + 1)][coord] / screeningFactors[jx];
					
					forcePtr[jx][coord] += forceFactor2b * displacements[jx][coord];
					
					for( kx = 0; kx < numberOfNeighbors; ++kx )
						forcePtr[kx][coord] += 
							DFDPartialDensitySq[2] * (2./3.) * density2term2 
							* atomicElecDensity[jx][2] 
							* screeningDerivs[jx * (numberOfNeighbors + 1) + kx + 1][coord] / screeningFactors[jx];
				} // end coord loop

				
				// forces from density 3
				for( alpha = 0; alpha < DIMENSION; ++alpha )
					for( beta = 0; beta < DIMENSION; ++beta )
						for( gamma = 0; gamma < DIMENSION; ++gamma )
						{
							double forceFactor3 = -DFDPartialDensitySq[3] * 2 * density3factor[alpha][beta][gamma]  *  atomicElecDensity[jx][3] * invPairDist / displacementsSq[jx];
							for( coord = 0; coord < DIMENSION; ++coord )
							{
								double forceFactor3a = forceFactor3 * (
									( alpha == coord ? (displacements[jx][beta]  * displacements[jx][gamma]) : 0.) +
									(  beta == coord ? (displacements[jx][alpha] * displacements[jx][gamma]) : 0.) + 
									( gamma == coord ? (displacements[jx][alpha] * displacements[jx][beta])  : 0.) 
									- displacements[jx][alpha] * displacements[jx][beta] * displacements[jx][gamma] * displacements[jx][coord] 
									* ((3. / displacementsSq[jx]) + MEAM_beta[3] * MEAM_invR0 * invPairDist)
									);
								forceI[coord] += -forceFactor3a + forceFactor3
									* displacements[jx][alpha]
									* displacements[jx][beta]
									* displacements[jx][gamma]
									* screeningDerivs[jx * (numberOfNeighbors + 1)][coord] / screeningFactors[jx];
								
								forcePtr[jx][coord] +=  forceFactor3a;
								
								
								for( kx = 0; kx < numberOfNeighbors; ++kx )
									forcePtr[kx][coord] += forceFactor3
										* displacements[jx][alpha]
										* displacements[jx][beta]
										* displacements[jx][gamma]
										* screeningDerivs[jx * (numberOfNeighbors + 1) + kx + 1][coord] / screeningFactors[jx];
							} // end coord loop
						} // end alpha-beta-gamma loop, density 3
			} // end of if(screeningFactors[jx] > 0.0) and jx loop
	} // end of i loop
}


double MEAMPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	ASSERT(atoms);
	int nAtoms = atoms->GetNumber();
	double potentialEnergy = 0.0;
	for( int i = 0; i < nAtoms; ++i )
		potentialEnergy += CalculateAtomicEnergy(atoms, i);

	return potentialEnergy;
}


double MEAMPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{
	int jx,p;
	double potential = 0.0;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int numberOfNeighbors;
	int spacesLeftInArrays = nPairs;
	double dummy1, dummy2;


	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,
		numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,
			numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}    
	ASSERT(numberOfNeighbors <= nPairs);
	
	ScreeningAndAtomicDensities(numberOfNeighbors);

	// initialize density, partial densities 
	double elecDensity = 0., partialDensitySq[4] = {0.,0.,0.,0.};

	// Do the pair part of the potential first.
	double dummy;
	for( jx = 0; jx < numberOfNeighbors; ++jx )
	{
		if(screeningFactors[jx] > 0.0)
			potential += CalculatePairInteraction(jx, &dummy);
		// having used the bare atomic densities, now multiply by the screening
		// factor. We didn't before because the pair part got multiplied by 
		// this separately
	   	for( p = 0; p < 4; ++p )
			atomicElecDensity[jx][p] *= screeningFactors[jx];
		
		// sum 0th partial density here; rest in CalculatePartialDensities123
	   	partialDensitySq[0] +=  atomicElecDensity[jx][0];
	} // end loop over jx


	// calculate other partial densities; the following arrays are needed by CalculatePartialDensities123()
	double density1factor[DIMENSION] = {0.,0.,0.,}, density2term1factor[DIMENSION][DIMENSION] = {{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}}, density2term2 = 0., density3factor[DIMENSION][DIMENSION][DIMENSION] = {{{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}},{{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}},{{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}}};
	
	CalculatePartialDensities123(numberOfNeighbors, partialDensitySq, density1factor, density2term1factor, &density2term2, density3factor);

	double angDensitySq = 0.0;
	for( p = 1; p < 4; ++p ) angDensitySq += MEAM_t[p] * partialDensitySq[p];
	// note that 'Sq' is inappropriate for the zeroth one, because
	// this was never squared, and is assumed to be unsquared 
	// by AddPartialDensities()
	elecDensity = AddPartialDensities(partialDensitySq[0], angDensitySq, &dummy1, &dummy2);

	//double scaledDensity = (elecDensity / MEAM_Zi) + TINY;
	double scaledDensity = (elecDensity / zBar) + TINY;

	potential += MEAM_A * MEAM_E0 * scaledDensity * log(scaledDensity);
	return potential;
} // end CalculateAtomicEnergy

/*!

  This is a function that could easily be made to operate on arrays; thus
  there would be large arrays for the screening functions, the distances
  and the atomic densities (not necessarily the whole LOA at a time---
  it would be like is done for EMT)
  
*/

void MEAMPotential::ScreeningAndAtomicDensities(int numberOfNeighbors)
{
	int jx, kx, coord;
	// calculate distances
	for( jx = 0; jx < numberOfNeighbors; ++jx )
		pairDistance[jx] = sqrt(displacementsSq[jx]);

	int nScrnDerivs = numberOfNeighbors * (numberOfNeighbors + 1);
	for(int dIdx = 0; dIdx < nScrnDerivs; ++dIdx )
		for( coord = 0; coord < DIMENSION; ++coord )
			screeningDerivs[dIdx][coord] = 0.;


	// from which: screening factors
	for( jx = 0; jx < numberOfNeighbors; ++jx )
	{	
		// pair screening
		double twoBodyScrnDer_over_Scrn_Rij;
		screeningFactors[jx] = TwoBodyScreen(pairDistance[jx], &twoBodyScrnDer_over_Scrn_Rij);
		twoBodyScrnDer_over_Scrn_Rij /= ( (screeningFactors[jx] + TINY) * pairDistance[jx] ); // TINY is to avoid 0/0
		
		// angular (three-body) screening
		for( kx = 0; kx < numberOfNeighbors; ++kx)
		{

			kx_scrnDer_over_Sijk[jx] = 0.;
			kx_scrn_contrib[jx] = false;

			if(kx != jx)
			{
				double rkjSq = 0.;
				for ( int coord = 0; coord < DIMENSION; ++coord )
					rkjSq += (displacements[jx][coord] - displacements[kx][coord])
						* (displacements[jx][coord] - displacements[kx][coord]);

				double Xik = displacementsSq[kx] / displacementsSq[jx];
				double Xkj = rkjSq / displacementsSq[jx];
				double plus = Xik + Xkj;
				double minus = Xik - Xkj;
				double denom = (1. - minus * minus);

				
				if( denom > TINY)
				{
					// denom being positive means atom k
					// is between the two parallel lines which 
					// go through i and j and are perpendicular
					// to the line between i and j
					double C = (2 * plus - (minus * minus) - 1) / denom;
					double Sijk = ThreeBodyScreen(C, kx_scrnDer_over_Sijk + kx);
					screeningFactors[jx] *= Sijk;
					kx_scrnDer_over_Sijk[kx] /= (Sijk + TINY);
					kx_scrn_contrib[kx] = true;
				}
				else
					kx_scrn_contrib[kx] = false;

				// otherwise it's a right angle or greater: no screening
				// ie screeningFactor
				// is unity so needn't do anything				
				
				if (screeningFactors[jx] < TINY) break;
			}  // end if (jx != kx)

	
		} // end loop over kx
		// need separate loop to do derivatives, because the whole S_ij is 
		// needed for each term
		if(screeningFactors[jx] > TINY)
		{
			// start with terms from two body screening
			if(twoBodyScrnDer_over_Scrn_Rij != 0.)
				for( coord = 0; coord < DIMENSION; ++coord )
				{
					screeningDerivs[jx * (numberOfNeighbors+1)][coord] += -screeningFactors[jx] * twoBodyScrnDer_over_Scrn_Rij * displacements[jx][coord];
					screeningDerivs[jx * (numberOfNeighbors+1) + jx + 1][coord] += screeningFactors[jx] * twoBodyScrnDer_over_Scrn_Rij * displacements[jx][coord];
				}

			for( kx = 0; kx < numberOfNeighbors; ++kx)
			{
				if(kx_scrn_contrib[kx])
				{
					double dispjk[DIMENSION];
					double rkjSq = 0.;
					for ( coord = 0; coord < DIMENSION; ++coord )
					{
						dispjk[coord] = displacements[kx][coord] - displacements[jx][coord];
						rkjSq += dispjk[coord] * dispjk[coord];
					}		
					double Xik = displacementsSq[kx] / displacementsSq[jx];
					double Xkj = rkjSq / displacementsSq[jx];
					double plus = Xik + Xkj;
					double minus = Xik - Xkj;
					double denom = (1. - minus * minus);
					
					double dCdplus = 2. / denom;
					double dCdminus = - 2 * minus / denom  + 2 * minus * (2 * plus - minus * minus - 1) / (denom * denom);
					
					double dCdXik = dCdplus + dCdminus;
					double dCdXkj = dCdplus - dCdminus;
					
					// the following is not the full derivative of Sij wrt Xik
					// just one term from the sum over k
					double dSijdXik_Xik =  screeningFactors[jx] * kx_scrnDer_over_Sijk[kx] * dCdXik * Xik;
					double dSijdXkj_Xkj =  screeningFactors[jx] * kx_scrnDer_over_Sijk[kx] * dCdXkj * Xkj;
					
					for( coord = 0; coord < DIMENSION; ++coord )
					{
						// kth term derivative of Sij wrt position of atom i
						// the 0th position in the row is for i
						screeningDerivs[jx * (numberOfNeighbors+1)][coord] += 
							dSijdXik_Xik * 2 * ( displacements[jx][coord] / displacementsSq[jx] - displacements[kx][coord] / displacementsSq[kx] )
							+ dSijdXkj_Xkj * 2 * displacements[jx][coord] / displacementsSq[jx];
						// derivative of Sijk wrt position of atom j
						// because - is for i, must add one for the rest
						screeningDerivs[jx * (numberOfNeighbors+1) + jx + 1][coord]
							+= dSijdXik_Xik * -2 * displacements[jx][coord] / displacementsSq[jx]
							+ dSijdXkj_Xkj * -2 * (displacements[jx][coord] / displacementsSq[jx] + dispjk[coord] / rkjSq)
							;
						// derivative of Sijk wrt position of atom k
						screeningDerivs[jx * (numberOfNeighbors+1) + kx + 1][coord]
							+= dSijdXik_Xik * 2 * displacements[kx][coord] / displacementsSq[kx]
							+ dSijdXkj_Xkj * 2 * dispjk[coord] / rkjSq;
					}
						  
				} // end if kx_scrn_contrib[kx]
				
			} // end second loop (for derivatives) over kx
		
		} // end of if screeningFactors[jx] > TINY

	} // end loop over jx


	
	for( jx = 0; jx < numberOfNeighbors; ++jx )
		for(int q = 0; q < 4; ++q )
		{
			atomicElecDensity[jx][q] = exp(-MEAM_beta[q] * (pairDistance[jx] * MEAM_invR0 - 1.));
		}

} // end ScreeningAndAtomicDensities


/*!

  Another candidate for doing using large arrays. Would have arrays for the 
  partial densities as well as the density1factor etc

*/

void MEAMPotential::CalculatePartialDensities123(int numberOfNeighbors, double partialDensitySq[4], double density1factor[DIMENSION], double density2term1factor[DIMENSION][DIMENSION], double *density2term2, double density3factor[DIMENSION][DIMENSION][DIMENSION])
{
	int alpha, beta, gamma, jx;
	// density 1
	for( alpha = 0; alpha < DIMENSION; ++alpha )
		{
			for( jx = 0; jx < numberOfNeighbors; ++jx )
				density1factor[alpha] += displacements[jx][alpha] * atomicElecDensity[jx][1] / pairDistance[jx];
			
			partialDensitySq[1] += density1factor[alpha] * density1factor[alpha];
		}

	// density 2 has two terms. First term:
	for( alpha = 0; alpha < DIMENSION; ++alpha )
		for( beta = 0; beta < DIMENSION; ++beta )
		{
			for( jx = 0; jx < numberOfNeighbors; ++jx )
				density2term1factor[alpha][beta] += displacements[jx][alpha] * displacements[jx][beta] * atomicElecDensity[jx][2] / displacementsSq[jx];
			
			partialDensitySq[2] += density2term1factor[alpha][beta] * density2term1factor[alpha][beta];
		}

	// Second term:
	ASSERT(*density2term2 == 0.);
	for( jx = 0; jx < numberOfNeighbors; ++jx )
	{
		*density2term2 += atomicElecDensity[jx][2];
	}
	partialDensitySq[2] -= (*density2term2) * (*density2term2) / 3.;

	// density 3
	for( alpha = 0; alpha < DIMENSION; ++alpha )
		for( beta = 0; beta < DIMENSION; ++beta )
			for( gamma = 0; gamma < DIMENSION; ++gamma )
			{
				for( jx = 0; jx < numberOfNeighbors; ++jx )
					density3factor[alpha][beta][gamma] +=  displacements[jx][alpha] * displacements[jx][beta] * displacements[jx][gamma] * atomicElecDensity[jx][3] / (displacementsSq[jx] * pairDistance[jx]);
				
				partialDensitySq[3] += density3factor[alpha][beta][gamma] * density3factor[alpha][beta][gamma];
			}
} // end CalculatePartialDensities123


/*!

  Could also be done using arrays. At this point, we'd be using up a lot of 
  memory!


*/

double MEAMPotential::CalculatePairInteraction(int jx, double *forceFactorPair)
{
	double invCoordination = 1./MEAM_Zi;
	
	// contribution from reference structure energies
	double aStar = MEAM_alpha * (pairDistance[jx] * MEAM_invR0 - 1.); 
	double refCohesiveEnergy = - MEAM_E0 * (1. + aStar) * exp(-aStar);
	double refCohesiveEnergyDeriv = MEAM_E0 * aStar * exp(-aStar) * MEAM_alpha * MEAM_invR0;
	
	// background density and its derivs in reference lattice
	double backgroundDensityZero = MEAM_Zi * atomicElecDensity[jx][0];
	double backgroundDensityZeroDeriv = - backgroundDensityZero * MEAM_beta[0] * MEAM_invR0;
	double backgroundAngularPartSq = 0.;
	double angularPartSqDeriv = 0.;
	
	if( referenceLattice == "diamond" )
	{
		backgroundAngularPartSq = (32./9.) * MEAM_t[3]
			* atomicElecDensity[jx][3] * atomicElecDensity[jx][3];
		angularPartSqDeriv = -2 * backgroundAngularPartSq * MEAM_beta[3] * MEAM_invR0;
	}
	
	else if( referenceLattice == "hcp" )
	{
		backgroundAngularPartSq = (1./3.) * MEAM_t[3] * atomicElecDensity[jx][3] * atomicElecDensity[jx][3];
		angularPartSqDeriv = -2 * backgroundAngularPartSq * MEAM_beta[3] * MEAM_invR0;
	}
	
	else if( referenceLattice == "dimer" )
	{
		backgroundAngularPartSq = 
			MEAM_t[1] * atomicElecDensity[jx][1] * atomicElecDensity[jx][1] 
			+ (2./3.) * MEAM_t[2] * atomicElecDensity[jx][2] * atomicElecDensity[jx][2]
			+ MEAM_t[3] * atomicElecDensity[jx][3] * atomicElecDensity[jx][3];
		
		angularPartSqDeriv = 
			-2 *  ( 
				MEAM_t[1] * atomicElecDensity[jx][1] * atomicElecDensity[jx][1] * MEAM_beta[1] 
				+ (2./3.) * MEAM_t[2] * atomicElecDensity[jx][2] * atomicElecDensity[jx][2] * MEAM_beta[2]
				+  MEAM_t[3] * atomicElecDensity[jx][3] * atomicElecDensity[jx][3] * MEAM_beta[3] 
				) * MEAM_invR0;
	}
	
	// otherwise it's assumed to be 
	// fcc or bcc in which case there's no angular contribution
	
	// combine the base part and the angular part
	double derivWRTZeroPart, derivWRTAngularPartSq;
	double backgroundDensity = AddPartialDensities(backgroundDensityZero, backgroundAngularPartSq, &derivWRTZeroPart, &derivWRTAngularPartSq);
	
	double backgroundDensityDeriv = derivWRTZeroPart * backgroundDensityZeroDeriv + derivWRTAngularPartSq * angularPartSqDeriv;
	
	//double scaledDensity = backgroundDensity * invCoordination;
	double scaledDensity = backgroundDensity / zBar;
	
	// embedding energy in reference lattice
	// deriv here is wrt backgroundDensity not scaledDensity
	double refEmbeddingEnergy = MEAM_A * MEAM_E0 * scaledDensity * log(scaledDensity);
	//double refEmbeddingEnergyDeriv = MEAM_A * MEAM_E0 * (1.0 + log(scaledDensity)) * backgroundDensityDeriv * invCoordination;
	double refEmbeddingEnergyDeriv = MEAM_A * MEAM_E0 * (1.0 + log(scaledDensity)) * backgroundDensityDeriv / zBar;

	
	double pairPotential = (refCohesiveEnergy - refEmbeddingEnergy) * screeningFactors[jx] * invCoordination;
	
	*forceFactorPair = - (refCohesiveEnergyDeriv - refEmbeddingEnergyDeriv)   * screeningFactors[jx] * invCoordination / pairDistance[jx];			

	return pairPotential;

}// end CalculatePairInteraction



///////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
///////////////////////////////////////////////////////////////////////////////

void MEAMPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);

	writer->PutDouble(MEAM_E0, "MEAM_E0");
	writer->PutDouble(MEAM_invR0, "MEAM_invR0");
	writer->PutDouble(MEAM_alpha, "MEAM_alpha");
	writer->PutDouble(MEAM_A, "MEAM_A");
	// variable names in netCDF must be alphanumeric plus "_"
	writer->PutDouble(MEAM_beta[0], "MEAM_beta_0");
	writer->PutDouble(MEAM_beta[1], "MEAM_beta_1");
	writer->PutDouble(MEAM_beta[2], "MEAM_beta_2");
	writer->PutDouble(MEAM_beta[3], "MEAM_beta_3");
	writer->PutDouble(MEAM_t[0], "MEAM_t_0");
	writer->PutDouble(MEAM_t[1], "MEAM_t_1");
	writer->PutDouble(MEAM_t[2], "MEAM_t_2");
	writer->PutDouble(MEAM_t[3], "MEAM_t_3");
	writer->PutInt(MEAM_Zi, "MEAM_Zi");
	
	
}

void MEAMPotential::Load(DMReader *reader)
{
	Potential::Load(reader);
	// We save the parameters in case somebody 
	// changes them in the future, and in the future after that wants to 
	// re examine old data. The function CheckPotentialParameter
	// throws an error if the double retrieved via the reader using the string
	// passed is not equal to the double passed

	CheckPotentialParameter(reader, MEAM_E0, "MEAM_E0");
	CheckPotentialParameter(reader, MEAM_invR0, "MEAM_invR0");
	CheckPotentialParameter(reader, MEAM_alpha, "MEAM_alpha");
	CheckPotentialParameter(reader, MEAM_A, "MEAM_A");
	CheckPotentialParameter(reader, MEAM_beta[0], "MEAM_beta[0]");
	CheckPotentialParameter(reader, MEAM_beta[1], "MEAM_beta[1]");
	CheckPotentialParameter(reader, MEAM_beta[2], "MEAM_beta[2]");
	CheckPotentialParameter(reader, MEAM_beta[3], "MEAM_beta[3]");
	CheckPotentialParameter(reader, MEAM_t[0], "MEAM_t[0]");
	CheckPotentialParameter(reader, MEAM_t[1], "MEAM_t[1]");
	CheckPotentialParameter(reader, MEAM_t[2], "MEAM_t[2]");
	CheckPotentialParameter(reader, MEAM_t[3], "MEAM_t[3]");
	CheckPotentialParameter(reader, MEAM_Zi, "MEAM_Zi");
}

#endif // DIMENISON == 3
