// StillingerWeberPotential.cpp: implementation of the
// StillingerWeberPotential class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"

#include "StillingerWeberPotential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"

//#define BRITTLE_FACTOR 2

#include "DMLastHeader.h"

using namespace std;


static SerialBuilder<StillingerWeberPotential> *_SWcreator = 
    new SerialBuilder<StillingerWeberPotential>("StillingerWeberPotential");

const string StillingerWeberPotential::className = string("StillingerWeberPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


StillingerWeberPotential::StillingerWeberPotential(double BRITTLE_FACTOR, string unitString) :SW_A(7.049556277),
													  SW_B(0.6022245584), 
													  SW_lambda(BRITTLE_FACTOR * 21.),
													  SW_gamma(1.2),
													  SW_a(1.8)
{
	SetCutoffDistance(SW_a);

	nAtoms = 0;

	neighbors = NULL;
	displacements = NULL;
	displacementsSq = NULL;
	if(unitString == "ev_angstroms1")
	{
		// this follows the original SW paper
		units = EV_ANGSTROMS1;
		lengthMultiplier = 2.0951; // the number of Angstroms per natural length unit.
		energyMultiplier = 2.167238; // = 3.4723/1.60217733: apart from factors of 10^-19 which cancel, the numerator is the number of Joules per natural energy unit and the denominator the number of Joules per eV. Thus overall, multiplying by this number converts energies in natural units to eV. The 3.4723 comes from the SW paper.
	}
	else if(unitString == "ev_angstroms2")
	{
		units = EV_ANGSTROMS2;
		lengthMultiplier = 2.0951;
		energyMultiplier = 2.315; 
		// this energy scaling is used by many people. See for example Balamane et al, Phys Rev B 46, 2250 (1992)
	}
	else
	{
		units = NATURAL;
		lengthMultiplier = 1.;
		energyMultiplier = 1.;
	}
}

StillingerWeberPotential::~StillingerWeberPotential()
{
	if(neighbors)
	{
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}
}

void StillingerWeberPotential::Allocate(int nAtoms)
{
   	// nAtoms is playing the role of EMTPotential's arrayLim here
	this->nAtoms = nAtoms;	
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nAtoms > 0)
	{
		neighbors       = new (std::nothrow) int[nAtoms];
		displacements   = new (std::nothrow) double[nAtoms][DIMENSION];
		displacementsSq = new (std::nothrow) double[nAtoms];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("StillingerWeberPotential",
								  "Allocate",
								  "Allocating nAtoms doubles");
		if(0 == displacements)
			throw NotEnoughMemory("StillingerWeberPotential",
								  "Allocate",
								  "Allocating nAtoms*DIMENSION doubles");
		if(0 == displacementsSq)
			throw NotEnoughMemory("StillingerWeberPotential",
								  "Allocate",
								  "Allocating nAtoms doubles");
	}
	else
	{
		neighbors       = NULL;
		displacements   = NULL;
		displacementsSq = NULL;
	}
	
}


/*! 
  By using Neighbors rather than half neighbors we can consider for a given
  i atom, all the triples where i is the 'middle' atom---given all neighbors 
  of i, loop over pairs of that set, and only compute the term where i is the 
  middle atom. This way you don't need to worry about counting the triple
  twice---it'll only be counted once with i in the middle.

*/

void StillingerWeberPotential::CalculateForces(ListOfAtoms *atoms,DMArray<double, DIMENSION> *forces)
{
	ASSERT(atoms && forces);

	int i,coord,jx,kx,j,k;
	int numberNeighbors;
	bool arrayNotTooFull;
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ, *forceK;
	double dummyForce[DIMENSION]; // where we send the force 
	// contribution to a ghost atom
	const double oneThird = 1./3.;
 
	for(i=0;i<nAtoms;i++)
    {	
		spacesLeftInArrays = nAtoms;
		arrayNotTooFull	= neighborLocator->Neighbors(i,neighbors,numberNeighbors, 
			displacements,displacementsSq,spacesLeftInArrays);
		ASSERT(arrayNotTooFull);


		// rescale according to lengthMultiplier
		if( units != NATURAL )
		{
			double l_scale = 1./lengthMultiplier;
			cblas_dscal(numberNeighbors*DIMENSION,l_scale, (double *) displacements, 1);
			cblas_dscal(numberNeighbors,l_scale*l_scale, displacementsSq, 1);
		}
		
		forceI = (*forces)[i];
		for(jx=0;jx<numberNeighbors;jx++)
		{
			j=neighbors[jx];
			if(j<nAtoms) forceJ = (*forces)[j];
			else forceJ = dummyForce;

			if(j<i)
			{
				double r2inv = 1.0/displacementsSq[jx];
				double r4inv = r2inv*r2inv;
				double r6inv = r2inv*r4inv;
				double r1 = sqrt(displacementsSq[jx]);
				
				double forceR = (4.*SW_A*SW_B*r6inv + SW_A*(SW_B*r4inv - 1) /
								 (r1*(r1-SW_a)*(r1-SW_a))) 
					* exp(1./(r1-SW_a));
				
				
				// Note that the displacements given by Neighbors are of particles j wrt particles i
				for (coord=0; coord<DIMENSION; coord++)
				{
					forceI[coord] += -displacements[jx][coord]*forceR;
					forceJ[coord] += +displacements[jx][coord]*forceR;
				}
			}	// endif for j<i don't want to double count. Normally avoided by the use of HalfNeighbors rather than Neighbors.
		
			// 3 body contributions
			for(kx=0;kx<jx;++kx)	
			{
				k=neighbors[kx];
				if(k<nAtoms) forceK = (*forces)[k];
				else forceK = dummyForce;
				// calculate the angle
				double cosTheta = 0;
				for(coord=0;coord<DIMENSION;++coord)
					cosTheta += displacements[jx][coord] *
						displacements[kx][coord];
				double r_ij = sqrt(displacementsSq[jx]);
				double r_ik = sqrt(displacementsSq[kx]);
				cosTheta /= (r_ij * r_ik);
				
				// forces. First some useful expressions and derivatives of h (the name given to the three body term in SW paper):
				double expFactor = SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)));
				double trigFactor = cosTheta + oneThird;
				double dh_drij_over_rij = - expFactor*trigFactor*trigFactor*SW_gamma/((r_ij-SW_a)*(r_ij-SW_a))/r_ij;
				double dh_drik_over_rik = - expFactor*trigFactor*trigFactor*SW_gamma/((r_ik-SW_a)*(r_ik-SW_a))/r_ik;
 				double dh_dCosTheta = 2*expFactor*trigFactor;

				double dCosThetadr_i[DIMENSION],dCosThetadr_j[DIMENSION],dCosThetadr_k[DIMENSION];
				
				for(coord=0; coord<DIMENSION; coord++)
				{
					dCosThetadr_i[coord] = cosTheta*(displacements[jx][coord]/displacementsSq[jx] + displacements[kx][coord]/displacementsSq[kx]) - (displacements[jx][coord]+displacements[kx][coord])/(r_ij*r_ik);
					dCosThetadr_j[coord] = displacements[kx][coord]/(r_ij*r_ik) - cosTheta*displacements[jx][coord]/displacementsSq[jx];
					dCosThetadr_k[coord] = displacements[jx][coord]/(r_ij*r_ik) - cosTheta*displacements[kx][coord]/displacementsSq[kx];

					forceI[coord] -= dh_drij_over_rij * -displacements[jx][coord] + dh_drik_over_rik * -displacements[kx][coord] + dh_dCosTheta * dCosThetadr_i[coord];
					forceJ[coord] -= dh_drij_over_rij * displacements[jx][coord] + dh_dCosTheta * dCosThetadr_j[coord];
					forceK[coord] -= dh_drik_over_rik * displacements[kx][coord] + dh_dCosTheta * dCosThetadr_k[coord];
				}

			} // finished with atom k, given by index kx
		
		}	// Finished with atom j, given by index jx

	} // end of i loop

 	// rescale according to lengthMultiplier
	if( units != NATURAL )
	{
		(*forces) *= energyMultiplier/lengthMultiplier;
	}

}


void StillingerWeberPotential::CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) 
// the force array isn't initialized in this function
{
	// rescale according to lengthMultiplier
	if( units != NATURAL )
	{
		throw DMException("StillingerWeberPotential","CalculateForcesBetweenTypes","non-natural units not yet implemented for this function");
	}

	int i,j,coord,jx,kx,k;
	int numNeighbors;
	bool arrayNotTooFull;
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ,*forceK;
	double dummyForce[DIMENSION];
	const double oneThird = 1./3.;
 
	int nAtoms1 = atoms->GetLeaf(type1)->GetNumber();
	for(i=0;i<nAtoms1;i++)
    {	
		spacesLeftInArrays = nAtoms;
		arrayNotTooFull	= neighborLocator->Neighbors(i, type1, type2,neighbors,numNeighbors, 
											displacements,displacementsSq,spacesLeftInArrays);
		ASSERT(arrayNotTooFull);

		forceI = (*(forces->GetLeaf(type1)))[i];
		for(jx=0;jx<numNeighbors;jx++)
		{
			j=neighbors[jx];
			forceJ = (*forces)[j];

			if(j<i)
			{
				double r2inv = 1.0/displacementsSq[jx];
				double r4inv = r2inv*r2inv;
				double r6inv = r2inv*r4inv;
				double r1 = sqrt(displacementsSq[jx]);
				
				double forceR = (4.*SW_A*SW_B*r6inv + SW_A*(SW_B*r4inv - 1) /
								 (r1*(r1-SW_a)*(r1-SW_a))) 
					* exp(1./(r1-SW_a));
				
				
				// Note that the displacements given by Neighbors are of particles j wrt particles i
				for (coord=0; coord<DIMENSION; coord++)
				{
					forceI[coord] += -displacements[jx][coord]*forceR;
					forceJ[coord] += +displacements[jx][coord]*forceR;
				}
			}	// endif for j<i don't want to double count. Normally avoided by the use of HalfNeighbors rather than Neighbors.
		
			// 3 body contributions
			for(kx=0;kx<jx;++kx)	
			{
				k=neighbors[kx];
				if(k<nAtoms) forceK = (*forces)[k];
				else forceK = dummyForce;
				// calculate the angle
				double cosTheta = 0;
				for(coord=0;coord<DIMENSION;++coord)
					cosTheta += displacements[jx][coord] *
						displacements[kx][coord];
				double r_ij = sqrt(displacementsSq[jx]);
				double r_ik = sqrt(displacementsSq[kx]);
				cosTheta /= (r_ij * r_ik);
				
				// forces. First some useful expressions and derivatives of h (the name given to the three body term in SW paper):
				double expFactor = SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)));
				double trigFactor = cosTheta + oneThird;
				double dh_drij_over_rij = - expFactor*trigFactor*trigFactor*SW_gamma/((r_ij-SW_a)*(r_ij-SW_a))/r_ij;
				double dh_drik_over_rik = - expFactor*trigFactor*trigFactor*SW_gamma/((r_ik-SW_a)*(r_ik-SW_a))/r_ik;
 				double dh_dCosTheta = 2*expFactor*trigFactor;

				double dCosThetadr_i[DIMENSION],dCosThetadr_j[DIMENSION],dCosThetadr_k[DIMENSION];
				
				for(coord=0; coord<DIMENSION; coord++)
				{
					dCosThetadr_i[coord] = cosTheta*(displacements[jx][coord]/displacementsSq[jx] + displacements[kx][coord]/displacementsSq[kx]) - (displacements[jx][coord]+displacements[kx][coord])/(r_ij*r_ik);
					dCosThetadr_j[coord] = displacements[kx][coord]/(r_ij*r_ik) - cosTheta*displacements[jx][coord]/displacementsSq[jx];
					dCosThetadr_k[coord] = displacements[jx][coord]/(r_ij*r_ik) - cosTheta*displacements[kx][coord]/displacementsSq[kx];

					forceI[coord] -= dh_drij_over_rij * -displacements[jx][coord] + dh_drik_over_rik * -displacements[kx][coord] + dh_dCosTheta * dCosThetadr_i[coord];
					forceJ[coord] -= dh_drij_over_rij * displacements[jx][coord] + dh_dCosTheta * dCosThetadr_j[coord];
					forceK[coord] -= dh_drik_over_rik * displacements[kx][coord] + dh_dCosTheta * dCosThetadr_k[coord];
				}

			} // finished with atom k, given by index kx
			
		}	// Finished with atom j, given by index jx
	} // end of i loop
}



/*! 
  By using Neighbors rather than half neighbors we can consider for a given
  i atom, all the triples where i is the 'middle' atom---given all neighbors 
  of i, loop over pairs of that set, and only compute the term where i is the 
  middle atom. This way you don't need to worry about counting the triple
  twice---it'll only be counted once with i in the middle.

*/

double StillingerWeberPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	ASSERT(atoms);

	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potentialEnergy = 0.0;
	int numberNeighbors;
	bool arrayNotTooFull;
	int i,j,jx,k,kx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	const double oneThird = 1./3.;


	for(i=0;i<nAtoms;i++)
    {
		spacesLeftInArrays = nAtoms;
		arrayNotTooFull	= neighborLocator->Neighbors(i,neighbors,numberNeighbors, 
											displacements,displacementsSq,spacesLeftInArrays);
		ASSERT(arrayNotTooFull);

		// rescale according to lengthMultiplier
		if( units != NATURAL )
		{
			double l_scale = 1./lengthMultiplier;
			cblas_dscal(numberNeighbors*DIMENSION,l_scale, (double *) displacements, 1);
			cblas_dscal(numberNeighbors,l_scale*l_scale, displacementsSq, 1);
		}



		for(jx=0;jx<numberNeighbors;++jx)
		{
			j=neighbors[jx];
			// because we got the neighbors and not just the half neighbors, we need the following if to avoid double counting the pair part.
			if(j<i) // avoid double counting
			{
				double r2inv = 1.0/displacementsSq[jx];
				double r4inv = r2inv*r2inv;
				double r1 = sqrt(displacementsSq[jx]);
				double pairEnergy = SW_A*(SW_B*r4inv - 1)*exp(1./(r1-SW_a));
				if(j >= nAtoms) // means this pair is between a real and a ghost atom
					pairEnergy /= 2.;
				potentialEnergy += pairEnergy;
			} // completes if(j<i)

			// 3 body contributions
			for(kx=0;kx<jx;++kx)	
			{
				k=neighbors[kx];
				// calculate the angle
				double cosTheta = 0;
				for(int coord=0;coord<DIMENSION;++coord)
					cosTheta += displacements[jx][coord]*displacements[kx][coord];
				double r_ij = sqrt(displacementsSq[jx]);
				double r_ik = sqrt(displacementsSq[kx]);
				cosTheta /= (r_ij * r_ik);

				potentialEnergy += SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)) )*(cosTheta+oneThird)*(cosTheta+oneThird);
	
			} // finished with atom k, given by index kx
			
		}	// Finished with atom j, given by index jx
    }
	return energyMultiplier * potentialEnergy;
}

double StillingerWeberPotential::CalculateEnergyAndForces(ListOfAtoms *atoms,DMArray<double, DIMENSION> *forces)
{
	ASSERT(atoms && forces);

	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potentialEnergy = 0.0;
	int i,jx,kx,j,k,coord;
	int numberNeighbors;
	bool arrayNotTooFull;

	// zero the forces:
	forces->Zero();

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	int spacesLeftInArrays;
	const double oneThird = 1./3.;
	double *forceI, *forceJ, *forceK;
	double dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom
	
	for(i=0;i<nAtoms;i++)
	{
		spacesLeftInArrays = nAtoms;
		arrayNotTooFull	= neighborLocator->Neighbors(i,neighbors,numberNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		ASSERT(arrayNotTooFull);


		// rescale according to lengthMultiplier
		if( units != NATURAL )
		{
			double l_scale = 1./lengthMultiplier;
			cblas_dscal(numberNeighbors*DIMENSION,l_scale, (double *) displacements, 1);
			cblas_dscal(numberNeighbors,l_scale*l_scale, displacementsSq, 1);
		}



    	forceI = (*forces)[i];
		for(jx=0;jx<numberNeighbors;jx++)
		{
			j=neighbors[jx];
			if(j<nAtoms) forceJ = (*forces)[j];
			else forceJ = dummyForce;
			
			if(j<i)
			{
				double r1 = sqrt(displacementsSq[jx]);	
				double r2inv = 1.0/ displacementsSq[jx];
				double r4inv = r2inv*r2inv;
				double r6inv = r2inv*r4inv;
				
				// pair energy
				double pairEnergy = SW_A*(SW_B*r4inv - 1)*exp(1./(r1-SW_a));
				if(neighbors[jx] >= nAtoms) // means this pair is between a real and a ghost atom
					pairEnergy /= 2;
				potentialEnergy += pairEnergy;				
				
				// pair forces
				double forceR = (4.*SW_A*SW_B*r6inv + SW_A*(SW_B*r4inv - 1)/(r1*(r1-SW_a)*(r1-SW_a)))*exp(1./(r1-SW_a));
				for (coord=0; coord<DIMENSION; coord++)
				{
					(*forces)[i][coord] += -displacements[jx][coord]*forceR;
					(*forces)[j][coord] += +displacements[jx][coord]*forceR;
				}
			}
			
			// 3 body contributions
			for(kx=0;kx<jx;++kx)	
			{
				k=neighbors[kx];
				if(k<nAtoms) forceK = (*forces)[k];
				else forceK = dummyForce;
				// calculate the angle
				double cosTheta = 0;
				for(coord=0;coord<DIMENSION;++coord)
					cosTheta += displacements[jx][coord]*displacements[kx][coord];
				double r_ij = sqrt(displacementsSq[jx]);
				double r_ik = sqrt(displacementsSq[kx]);
				cosTheta /= (r_ij * r_ik);
				
				// Useful expressions and derivatives of h (the name given to the three body term in SW paper):
				double expFactor = SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)));
				double trigFactor = cosTheta + oneThird;
				
				// 3 body energy
				potentialEnergy += SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)) )*(cosTheta+oneThird)*(cosTheta+oneThird);

				// 3 body forces
				double dh_drij_over_rij = - expFactor*trigFactor*trigFactor*SW_gamma/((r_ij-SW_a)*(r_ij-SW_a))/r_ij;
				double dh_drik_over_rik = - expFactor*trigFactor*trigFactor*SW_gamma/((r_ik-SW_a)*(r_ik-SW_a))/r_ik;
 				double dh_dCosTheta = 2*expFactor*trigFactor;

				double dCosThetadr_i[DIMENSION],dCosThetadr_j[DIMENSION],dCosThetadr_k[DIMENSION];
				
				for(coord=0; coord<DIMENSION; coord++)
				{
					dCosThetadr_i[coord] = cosTheta*(displacements[jx][coord]/displacementsSq[jx] + displacements[kx][coord]/displacementsSq[kx]) - (displacements[jx][coord]+displacements[kx][coord])/(r_ij*r_ik);
					dCosThetadr_j[coord] = displacements[kx][coord]/(r_ij*r_ik) - cosTheta*displacements[jx][coord]/displacementsSq[jx];
					dCosThetadr_k[coord] = displacements[jx][coord]/(r_ij*r_ik) - cosTheta*displacements[kx][coord]/displacementsSq[kx];

					forceI[coord] -= dh_drij_over_rij * -displacements[jx][coord] + dh_drik_over_rik * -displacements[kx][coord] + dh_dCosTheta * dCosThetadr_i[coord];
					forceJ[coord] -= dh_drij_over_rij * displacements[jx][coord] + dh_dCosTheta * dCosThetadr_j[coord];
					forceK[coord] -= dh_drik_over_rik * displacements[kx][coord] + dh_dCosTheta * dCosThetadr_k[coord];
				}

			} // finished with atom k, given by index kx

		}	// Finished with atom j, given by index jx

	} // end of i loop
 
	if( units != NATURAL )
	{
		(*forces) *= energyMultiplier/lengthMultiplier;
	}
	return energyMultiplier * potentialEnergy;
}

/*! 
Here the energy of an atom is half the energy in each of its bonds.

*/

double StillingerWeberPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{  

	int j,k,kx;
	double cosTheta,r_ij,r_ik,pairEnergy,r1;
	const double oneThird = 1./3.;
    if(i >= atoms->GetNumber() || i<0)
		throw IndexError("StillingerWeberPotential","CalculateAtomicEnergy",
						 "No such atom.");
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potentialEnergy = 0.0, r2inv,r4inv;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int spacesLeftInArrays, numberOfNeighbors;
	spacesLeftInArrays = nAtoms;
	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,numberOfNeighbors, 
					       displacements,displacementsSq,spacesLeftInArrays);
	ASSERT(arrayNotTooFull);


	// rescale according to lengthMultiplier
	if( units != NATURAL )
	{
		double l_scale = 1./lengthMultiplier;
		cblas_dscal(numberOfNeighbors*DIMENSION,l_scale, (double *) displacements, 1);
		cblas_dscal(numberOfNeighbors,l_scale*l_scale, displacementsSq, 1);
	}
	

	for(int jx=0;jx<numberOfNeighbors;jx++)
	{
	   
		j=neighbors[jx];
		r2inv = 1.0/displacementsSq[jx];
		r4inv = r2inv*r2inv;
		r1 = sqrt(displacementsSq[jx]);
		pairEnergy = SW_A*(SW_B*r4inv - 1)*exp(1./(r1-SW_a));
		potentialEnergy += pairEnergy/2; // since we have split energy of bonds among the atoms.
		// 3 body contributions: we can associate the contribution from the term with i in the middle to that atom
		for(kx=0;kx<jx;++kx)	
		{
			k=neighbors[kx];
			// calculate the angle
			cosTheta = 0;
			for(int coord=0;coord<DIMENSION;++coord)
				cosTheta += displacements[jx][coord]*displacements[kx][coord];
			r_ij = sqrt(displacementsSq[jx]);
			r_ik = sqrt(displacementsSq[kx]);
			cosTheta /= r_ij * r_ik;
			potentialEnergy += SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)) )*(cosTheta+oneThird)*(cosTheta+oneThird);	
		} // finished with atom k, given by index kx
			
	}	// Finished with atom j, given by index jx 
		   
			 
	return energyMultiplier * potentialEnergy;   

}

void StillingerWeberPotential::CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[DIMENSION*(DIMENSION+1)/2])
{
	for( int p = 0; p < (DIMENSION+1)*DIMENSION/2; ++p )
		stress[p] = 0.;

	StressTensorKernel(atoms, i, stress);

	// Add the dynamic part and scale with the atomic volume
	double latticeConst = (4./sqrt(3.)) * GetLengthScale();
	double invvol = 8./(latticeConst * latticeConst * latticeConst); // 2.59 is equilibrium lattice constant. 8 atoms per unit cell in a diamond lattice. Assuming D=3 here.
	double mass = atoms->GetMass();
	const double *veloCurrent = (*(atoms->GetCartesianVelocities()))[i] ;
	for(int alpha=0; alpha<DIMENSION; ++alpha)
		for(int beta=alpha; beta<DIMENSION; ++beta)
		{
			int ii = stresscomp[alpha][beta];
			stress[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
			stress[ii] *= invvol;
		}
}

void StillingerWeberPotential::CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress)
{
	int i, nAtoms = atoms->GetNumber();
	stress->Zero();
	
	// the main work is done by StressTensorKernel which deals with one atom at a time
	DMArrayIterator<double,(DIMENSION+1)*DIMENSION/2> stressIter(stress); 
	for( i = 0, stressIter.Init(); i < nAtoms; ++i, stressIter.Next() )
	{
		double *current = stressIter.Current();
		StressTensorKernel(atoms,i, current);
	}


	// Add the dynamic part and scale with the atomic volume
	double latticeConst = (4./sqrt(3.)) * GetLengthScale();
	double invvol = 8./(latticeConst * latticeConst * latticeConst); // assuming diamond lattice (8 atoms per unit cell) and DIMENSION = 3 of course.
	double mass = atoms->GetMass();
	double stressMultiplier = energyMultiplier;
	for(int coord = 0; coord < DIMENSION; ++coord)
		stressMultiplier /= lengthMultiplier;
		
	DMArrayIterator<double, DIMENSION> velIter(atoms->GetCartesianVelocities());
	for( velIter.Init(), stressIter.Init(); !velIter.IsDone(); velIter.Next(), stressIter.Next() )
	{
		const double *veloCurrent = velIter.Current();
		double *stressCurrent = stressIter.Current();
		for(int alpha=0; alpha<DIMENSION; ++alpha)
			for(int beta=alpha; beta<DIMENSION; ++beta)
			{
				int ii = stresscomp[alpha][beta];
				stressCurrent[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
				stressCurrent[ii] *= invvol * stressMultiplier;
			}
		
	}
}



void StillingerWeberPotential::StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2])
{
	int jx,kx,ii,coord,alpha,beta;
	const double oneThird = 1./3.;
	
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int spacesLeftInArrays, numberOfNeighbors;
	spacesLeftInArrays = nAtoms;

	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,numberOfNeighbors, 
													  displacements,displacementsSq,spacesLeftInArrays);
	ASSERT(arrayNotTooFull);

	// rescale according to lengthMultiplier
	if( units != NATURAL )
	{
		double l_scale = 1./lengthMultiplier;
		cblas_dscal(numberOfNeighbors*DIMENSION,l_scale, (double *) displacements, 1);
		cblas_dscal(numberOfNeighbors,l_scale*l_scale, displacementsSq, 1);
	}

	
	for(jx=0;jx<numberOfNeighbors;jx++)
	{
		// pair contributions
		double r2inv = 1.0/displacementsSq[jx];
		double r4inv = r2inv*r2inv;
		double r6inv = r2inv*r4inv;
		double r1 = sqrt(displacementsSq[jx]);
		double forceR = (4.*SW_A*SW_B*r6inv + SW_A*(SW_B*r4inv - 1) /
						 (r1*(r1-SW_a)*(r1-SW_a))) 
			* exp(1./(r1-SW_a));
		
		
		for(alpha=0; alpha<DIMENSION; ++alpha)
			for(beta=alpha; beta<DIMENSION; ++beta)
			{
				ii = stresscomp[alpha][beta];
				
				stress[ii] += - 0.5*forceR*displacements[jx][alpha]*displacements[jx][beta];
			}

		// three-body contributions
		for(kx=0;kx<jx;++kx)	
		{
			// calculate the angle
			double cosTheta = 0;
			for(coord=0;coord<DIMENSION;++coord)
				cosTheta += displacements[jx][coord] *
					displacements[kx][coord];
			double r_ij = sqrt(displacementsSq[jx]);
			double r_ik = sqrt(displacementsSq[kx]);
			cosTheta /= (r_ij * r_ik);
			
			double expFactor = SW_lambda*exp(SW_gamma*(1./(r_ij-SW_a)+1./(r_ik-SW_a)));
			double trigFactor = cosTheta + oneThird;

			double dh_dr_ij[DIMENSION],dh_dr_ik[DIMENSION];
			for(coord=0;coord<DIMENSION;++coord)
			{
				dh_dr_ij[coord] = expFactor*trigFactor*trigFactor*(SW_gamma*displacements[jx][coord]/(r_ij*(r_ij-SW_a)*(r_ij-SW_a)))
					+ expFactor * 2 * trigFactor * (displacements[kx][coord]/(r_ij*r_ik) - cosTheta*displacements[jx][coord]/(r_ij*r_ij));

				dh_dr_ik[coord] = expFactor*trigFactor*trigFactor*(SW_gamma*displacements[kx][coord]/(r_ik*(r_ik-SW_a)*(r_ik-SW_a)))
					+ expFactor * 2 * trigFactor * (displacements[jx][coord]/(r_ij*r_ik) - cosTheta*displacements[kx][coord]/(r_ik*r_ik));
			}


			for(alpha=0; alpha<DIMENSION; ++alpha)
				for(beta=alpha; beta<DIMENSION; ++beta)
				{
					ii = stresscomp[alpha][beta];
					
					stress[ii] += dh_dr_ij[alpha]*displacements[jx][beta] + dh_dr_ik[alpha]*displacements[kx][beta];
				}

		}
	}

}


void StillingerWeberPotential::CalculateHessianMatrix(ListOfAtoms *atoms, double *matrix)
{
	ASSERT(atoms);
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());

	int i,j,jx,k,kx, coord, coord1, coord2,size = nAtoms*DIMENSION,spacesLeftInArrays;
	int numberNeighbors, rowIndex, colIndex;
	const double oneThird = 1./3.; 
	double SW_a2 = SW_a * SW_a;
	// zero the matrix
	memset(matrix,0,size * size * sizeof(double));
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);	

	for(i=0;i<nAtoms;i++)
    {
		spacesLeftInArrays = nAtoms;
		bool arrayNotTooFull	= neighborLocator->Neighbors(i,neighbors,numberNeighbors, 
													 displacements,displacementsSq,spacesLeftInArrays);
		ASSERT(arrayNotTooFull);


		// rescale according to lengthMultiplier
		if( units != NATURAL )
		{
			double l_scale = 1./lengthMultiplier;
			cblas_dscal(numberNeighbors*DIMENSION,l_scale, (double *) displacements, 1);
			cblas_dscal(numberNeighbors,l_scale*l_scale, displacementsSq, 1);
		}
		
		// pair terms first...do derivatives of force on j and of force on i with respect to postion of i..should also be able to do both with respect to j
		for(jx=0;jx<numberNeighbors;++jx)
		{
			j = neighbors[jx];		   
			if(j < i)
			{
				double r2 = displacementsSq[jx];
				double r1 = sqrt(r2);				
				double r3 = r2 * r1;
				double r4 = r2 * r2;
				double r5 = r4*r1;
				double r6 = r4*r2;
				double r7 = r6*r1;
				double r2inv = 1.0/r2;
				double r4inv = r2inv * r2inv;
				double r6inv = r2inv * r4inv;
				double r8inv = r4inv * r4inv;
				double a_minus_r1Inv2 = 1./((SW_a - r1)*(SW_a - r1));
				double a_minus_r1Inv4 = a_minus_r1Inv2 * a_minus_r1Inv2;
				double derivatives[DIMENSION][DIMENSION];
				double expFactor = SW_A * exp(1./(r1-SW_a));
				// derivatives[coord1][coord2] is the derivative of coord1 of the force on j due to i with respect to coord 2 of the position of i

				double commonPart = - expFactor * a_minus_r1Inv4 * r8inv * 
							(
								2. * (r7 - SW_a * r6) + 
								r6*(1. - 2*SW_a + r1) + 
								r4 * (SW_a2 * r1 - 24 * SW_B) - 
								r2 * SW_B * (1. - 2 * SW_a + 144. * SW_a2) + 
								2. * r3 * SW_B * (48. * SW_a - 1.) +
								6. * SW_a * r1 * SW_B * (16. * SW_a2 + 3. * r1) - 
								3. * SW_B * (8. * SW_a2 * SW_a2 + 3. * SW_a2 * r1 + 3. * r3)
								);


				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = commonPart * displacements[jx][coord1] * displacements[jx][coord2]; 
				
				
				double diagonalCorrection =  -expFactor * r6inv * a_minus_r1Inv2 * 
					(
						-r5 - 8 * SW_a * r1 * SW_B + 
						4 * r2 * SW_B + SW_B * (4*SW_a2 + r1)
						);
				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * j + coord2; 
						colIndex = DIMENSION * i + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
						rowIndex = DIMENSION * i + coord2; 
						colIndex = DIMENSION * i + coord1; 
						matrix[rowIndex + size * colIndex] += -derivatives[coord1][coord2];
						rowIndex = DIMENSION * i + coord2; 
						colIndex = DIMENSION * j + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
						rowIndex = DIMENSION * j + coord2; 
						colIndex = DIMENSION * j + coord1; 
						matrix[rowIndex + size * colIndex] += -derivatives[coord1][coord2];

					} // end coord1, coord2 loop
			} // if i<j
			
			// triple terms...same trick as in force calculator should work: we consider just the terms for which i is the middle atom but have to consider derivatives with respect to i and each other atom in the triple, of the forces on each of three atoms
			
			for(kx=0;kx<jx;++kx)	
			{
				k=neighbors[kx];
				double r_ij = sqrt(displacementsSq[jx]);
				double r_ik = sqrt(displacementsSq[kx]);
				double r_ij2 = r_ij * r_ij;
				double r_ij3 = r_ij2 * r_ij;
				double r_ij4 = r_ij3 * r_ij;
				double r_ik2 = r_ik * r_ik;
				double r_ik3 = r_ik2 * r_ik;
				double r_ik4 = r_ik3 * r_ik;
				// calculate the angle
				double cosTheta = 0;
				for(coord=0;coord<DIMENSION;++coord)
					cosTheta += displacements[jx][coord] *
						displacements[kx][coord];			
				cosTheta /= (r_ij * r_ik);
				double trigFactor = cosTheta + oneThird;
				double r_ij_m_a = r_ij - SW_a;
				double r_ik_m_a = r_ik - SW_a;
				double r_ij_m_a2 = r_ij_m_a *  r_ij_m_a;
				double r_ik_m_a2 = r_ik_m_a *  r_ik_m_a;
				double r_ij_m_a3 = r_ij_m_a2 *  r_ij_m_a;
				double r_ik_m_a3 = r_ik_m_a2 *  r_ik_m_a;
				double SW_a2 = SW_a * SW_a;
				double SW_a3 = SW_a2 * SW_a;
				double SW_a4 = SW_a3 * SW_a;
				double SW_a_gamma = SW_a * SW_gamma;
				double expFactor = SW_lambda*exp(SW_gamma*(1./(r_ij_m_a)+1./(r_ik_m_a)));
				

				double dispJJ[DIMENSION][DIMENSION], dispKK[DIMENSION][DIMENSION],dispJK[DIMENSION][DIMENSION];
				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						dispJJ[coord1][coord2] = displacements[jx][coord1] * displacements[jx][coord2];
						dispKK[coord1][coord2] = displacements[kx][coord1] * displacements[kx][coord2];
						dispJK[coord1][coord2] = displacements[jx][coord1] * displacements[kx][coord2];
					}
				double jFactor[DIMENSION], kFactor[DIMENSION];
				for(coord1=0;coord1<DIMENSION;++coord1)
				{
					jFactor[coord1] = displacements[jx][coord1]/(r_ij * r_ij_m_a2);
					kFactor[coord1] = displacements[kx][coord1]/(r_ik * r_ik_m_a2);
				}
				double derivatives[DIMENSION][DIMENSION];
				// derivatives[coord1][coord2] is the derivative with respect to coord1 of the position of whichever atom, of coord2 of the force on whichever atom 



				// There are three atoms involved here. We need the deriviatives of the forces on each one with respect to the positions of each one, so there are nin derivatives, not counting different spatial components (that's taken care of by the coord loops) dfidxi, dfidxj, dfidkx, dfjdxi etc. However dfidxj = dfjdxi since these are the same second derivatives of the potential, so there are six distinct formulas, which I have worked out in mathematica. The longests is dfidxi.
				
				// *************************** dfidxi *********************

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = expFactor * 
							(
								-SW_gamma * trigFactor * trigFactor * (
									jFactor[coord1] * jFactor[coord2] *  ( 2 * r_ij_m_a + r_ij_m_a2/r_ij)   +  kFactor[coord1] * kFactor[coord2] *  ( 2 * r_ik_m_a + r_ik_m_a2/r_ik)
									) - 
								2 * (
									cosTheta * (r_ik2 * displacements[jx][coord1] + r_ij2 * displacements[kx][coord1]) - r_ij * r_ik * (displacements[jx][coord1] + displacements[kx][coord1])
									) * 
								(
									cosTheta * (r_ik2 * displacements[jx][coord2] + r_ij2 * displacements[kx][coord2]) - r_ij * r_ik * (displacements[jx][coord2] + displacements[kx][coord2])
									) /(r_ij4 * r_ik4) + 
								2 * trigFactor * (
									-(
										3 * cosTheta * (r_ik4 * dispJJ[coord1][coord2] +  r_ij4 * dispKK[coord1][coord2]) + 
										
										cosTheta * r_ij2 * r_ik2 * (dispJK[coord2][coord1] + dispJK[coord1][coord2]) - 
										r_ik * r_ij3 * (dispJK[coord2][coord1] + 2 * dispKK[coord1][coord2] + dispJK[coord1][coord2]) - 
										r_ij * r_ik3 * (dispJK[coord2][coord1] + 2 * dispJJ[coord1][coord2] + dispJK[coord1][coord2])
										)/(r_ij4 * r_ik4) - 
									(
										(cosTheta * r_ik2 * displacements[jx][coord1] + cosTheta * r_ij2 * displacements[kx][coord1] - r_ij * r_ik * (displacements[jx][coord1] + displacements[kx][coord1])) * (jFactor[coord2] + kFactor[coord2]) + 
										(cosTheta * r_ik2 * displacements[jx][coord2] + cosTheta * r_ij2 * displacements[kx][coord2] - r_ij * r_ik * (displacements[jx][coord2] + displacements[kx][coord2])) * (jFactor[coord1] + kFactor[coord1])										

										) * SW_gamma/(r_ij2 * r_ik2) 
									) - 
								trigFactor * trigFactor * SW_gamma * SW_gamma * (jFactor[coord1] + kFactor[coord1]) * (jFactor[coord2] + kFactor[coord2])
									 
								
								);
				
				// add extra diagonal piece
				double diagonalCorrection = trigFactor * expFactor * 
					(
						r_ij * r_ik * (
							r_ij3 * SW_gamma + r_ik_m_a2 * (-12 * SW_a2 + r_ik * SW_gamma) - 2 * r_ij2 * (6 * r_ik2 - 12 * r_ik * SW_a + 6 * SW_a2 + SW_a_gamma) + r_ij * SW_a * (24 * r_ik2 - 48 * r_ik * SW_a + 24 * SW_a2 + SW_a_gamma)
							) + 
						3 * cosTheta * (
							2 * r_ik2 * r_ik_m_a2 * SW_a2 - r_ij * r_ik2 * r_ik_m_a2 * (4 * SW_a - SW_gamma) + (r_ij4 - 2 * r_ij3 * SW_a) * (2 * r_ik2 - 4 * r_ik * SW_a + 2 * SW_a2 + r_ik * SW_gamma ) + r_ij2 * (2 * r_ik4 - 4 * r_ik3 * SW_a + 4 * r_ik2 * SW_a2 - 4 * r_ik * SW_a3 + 2 * SW_a4 + r_ik * SW_a2 * SW_gamma)
							)
						
						)/(3 * r_ij2 * r_ik2 * r_ij_m_a2 * r_ik_m_a2); 
				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;
				
				
				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * i + coord2; 
						colIndex = DIMENSION * i + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
					}


				// *************************** dfjdxj *********************

	
				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = expFactor * 
							(
								-2 * cosTheta * cosTheta * dispJJ[coord1][coord2] + 
								r_ij * trigFactor * (-SW_gamma * trigFactor * dispJJ[coord1][coord2]/r_ij_m_a2 + 2 * (dispJK[coord2][coord1] + dispJK[coord1][coord2])/r_ik) + 
								r_ij2 * (
									SW_gamma * trigFactor * trigFactor * dispJJ[coord1][coord2] * (-2/r_ij_m_a  - SW_gamma /r_ij_m_a2 ) + 
									2 * SW_gamma * trigFactor * (dispJK[coord2][coord1] + dispJK[coord1][coord2])/r_ik
									 - 2 * dispKK[coord1][coord2] * r_ij_m_a2/r_ik2
									)/r_ij_m_a2 + 
								2 * cosTheta * ((-3. -2 * r_ij * SW_gamma / r_ij_m_a2)*trigFactor * dispJJ[coord1][coord2] + r_ij * (dispJK[coord2][coord1] + dispJK[coord1][coord2])/r_ik)

								)/ r_ij4;
				
				// add extra diagonal piece

				diagonalCorrection = trigFactor * expFactor * (2*cosTheta * r_ij_m_a2 + r_ij * SW_gamma * trigFactor) / (r_ij2 * r_ij_m_a2);

				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * j + coord2; 
						colIndex = DIMENSION * j + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
					}

				// ********************** dfkdxk *****************************

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = expFactor * 
							(
								-2 * cosTheta * cosTheta * dispKK[coord1][coord2] + 
								r_ik * trigFactor * (-SW_gamma * trigFactor * dispKK[coord1][coord2]/r_ik_m_a2 + 2 * (dispJK[coord1][coord2] + dispJK[coord2][coord1])/r_ij) + 
								r_ik2 * (
									SW_gamma * trigFactor * trigFactor * dispKK[coord1][coord2] * (-2/r_ik_m_a  - SW_gamma /r_ik_m_a2 ) + 
									2 * SW_gamma * trigFactor * (dispJK[coord1][coord2] + dispJK[coord2][coord1])/r_ij
									- 2 * dispJJ[coord1][coord2] * r_ik_m_a2/r_ij2
									)/r_ik_m_a2 + 
								2 * cosTheta * ((-3. -2 * r_ik * SW_gamma / r_ik_m_a2)*trigFactor * dispKK[coord1][coord2] + r_ik * (dispJK[coord1][coord2] + dispJK[coord2][coord1])/r_ij)
								
								)/ r_ik4;


				// add extra diagonal piece
				diagonalCorrection = trigFactor * expFactor * (2*cosTheta * r_ik_m_a2 + r_ik * SW_gamma * trigFactor) / (r_ik2 * r_ik_m_a2);


				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * k + coord2; 
						colIndex = DIMENSION * k + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
					}
				

				// ********************* dfidxj and dfjdxi ********************

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = expFactor * 
							(
								(3 * r_ij - SW_a) * SW_gamma * trigFactor * trigFactor * dispJJ[coord1][coord2] / (r_ij3 * r_ij_m_a3) + 
								2 * (cosTheta * r_ik * displacements[jx][coord1] - r_ij * displacements[kx][coord1]) * (cosTheta * r_ik2 * displacements[jx][coord2] + cosTheta * r_ij2 * displacements[kx][coord2] - r_ij * r_ik * (displacements[jx][coord2] + displacements[kx][coord2])) / (r_ij4 * r_ik3) - 
								2 * trigFactor * (-3 * cosTheta * r_ik4 * dispJJ[coord1][coord2] - cosTheta * r_ij2 * r_ik2 * dispJK[coord1][coord2] + r_ij3 * r_ik * dispKK[coord1][coord2] + r_ij * r_ik3 * (dispJK[coord2][coord1] + dispJJ[coord1][coord2] + dispJK[coord1][coord2])) / (r_ij4 * r_ik4) + 
								2 * SW_gamma * trigFactor * jFactor[coord1] * (cosTheta * (r_ik2 * displacements[jx][coord2] + r_ij2 * displacements[kx][coord2]) - r_ij * r_ik * (displacements[jx][coord2] + displacements[kx][coord2]) ) / (r_ij2 * r_ik2) + 
								SW_gamma * SW_gamma * trigFactor * trigFactor * jFactor[coord1] * (jFactor[coord2] + kFactor[coord2]) + 
								2 * SW_gamma * trigFactor * (cosTheta * r_ik * displacements[jx][coord1] - r_ij * displacements[kx][coord1]) * (jFactor[coord2] + kFactor[coord2]) / (r_ij2 * r_ik)
							);
				
				// add extra diagonal piece
				diagonalCorrection = -trigFactor * expFactor * (
					2 * (-r_ij3 + cosTheta * r_ik * SW_a2) +
					2 * r_ij2 * (cosTheta * r_ik + 2 * SW_a) + 
					r_ij * (-4 * cosTheta * r_ik * SW_a - 2 * SW_a2 + r_ik * SW_gamma * trigFactor)
					)/(r_ij2 * r_ik * r_ij_m_a2);

				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;

				// derivatives[coord1][coord2] has derivative of the coord2 component of the force on atom i wrt coord1 of the position of atom j 

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * i + coord2; 
						colIndex = DIMENSION * j + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
						rowIndex = DIMENSION * j + coord1; 
						colIndex = DIMENSION * i + coord2; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
					}

				
				// ******************** dfidxk and dfkdxi ********************
				
				
				
				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = expFactor *
							(
								(3 * r_ik - SW_a) * SW_gamma * trigFactor * trigFactor * dispKK[coord1][coord2] / (r_ik3 * r_ik_m_a3) + 
								2*(cosTheta * r_ij * displacements[kx][coord1] - r_ik * displacements[jx][coord1]) * (cosTheta * r_ik2 * displacements[jx][coord2] + cosTheta * r_ij2 * displacements[kx][coord2] - r_ij * r_ik * (displacements[jx][coord2] + displacements[kx][coord2])) / (r_ij3 * r_ik4) - 
								2 * trigFactor * (-3 * cosTheta * r_ij4 * dispKK[coord1][coord2] - cosTheta * r_ij2 * r_ik2 * dispJK[coord2][coord1] + r_ik3 * r_ij * dispJJ[coord1][coord2] + r_ik * r_ij3 * (dispJK[coord2][coord1] + dispKK[coord1][coord2] + dispJK[coord1][coord2])) / (r_ij4 * r_ik4) + 
								2 * SW_gamma * trigFactor * kFactor[coord1] * (cosTheta * (r_ik2 * displacements[jx][coord2] + r_ij2 * displacements[kx][coord2]) - r_ij * r_ik * (displacements[jx][coord2] + displacements[kx][coord2]) ) / (r_ij2 * r_ik2) + 
								SW_gamma * SW_gamma * trigFactor * trigFactor * kFactor[coord1] * (jFactor[coord2] + kFactor[coord2]) + 
								2 * SW_gamma * trigFactor * (cosTheta * r_ij * displacements[kx][coord1] - r_ik * displacements[jx][coord1]) * (jFactor[coord2] + kFactor[coord2]) / (r_ij * r_ik2)
								);

				// add extra diagonal piece				
				diagonalCorrection = -trigFactor * expFactor * (
					2 * (-r_ik3 + cosTheta * r_ij * SW_a2) +
					2 * r_ik2 * (cosTheta * r_ij + 2 * SW_a) + 
					r_ik * (-4 * cosTheta * r_ij * SW_a - 2 * SW_a2 + r_ij * SW_gamma * trigFactor)
					)/(r_ik2 * r_ij * r_ik_m_a2);


				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * i + coord2; 
						colIndex = DIMENSION * k + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
						rowIndex = DIMENSION * k + coord1; 
						colIndex = DIMENSION * i + coord2; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
					}

				// ******************** dfjdxk and dfkdxj *****************

				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
						derivatives[coord1][coord2] = -expFactor * 
							(
								2 * r_ij2 * r_ik2 * (cosTheta * cosTheta  * dispJK[coord2][coord1] +  dispJK[coord1][coord2]) - 
								2 * r_ij3 * r_ik * trigFactor * dispKK[coord1][coord2] + 
								r_ij4 * r_ik4 * SW_gamma * SW_gamma * trigFactor * trigFactor * jFactor[coord2] * kFactor[coord1] + 
								2 * cosTheta * r_ij * r_ik * (
									-r_ik2 * dispJJ[coord1][coord2] - r_ij2 * dispKK[coord1][coord2] + r_ij * r_ik * trigFactor * displacements[kx][coord1] * (displacements[jx][coord2] + r_ij2 * SW_gamma * jFactor[coord2]) + r_ij * r_ik3 * SW_gamma * trigFactor * displacements[jx][coord2] * kFactor[coord1]) - 
								2 * r_ik3 * trigFactor * (r_ij * dispJJ[coord1][coord2] + r_ij3 * SW_gamma * (displacements[jx][coord1] * jFactor[coord2] + displacements[kx][coord2] * kFactor[coord1]))
							
							) / (r_ij4 * r_ik4); 
				
				// add extra diagonal piece
				diagonalCorrection = -2 * trigFactor * expFactor/(r_ij * r_ik);


				for(coord1=0;coord1<DIMENSION;++coord1)
					derivatives[coord1][coord1] += diagonalCorrection;


				for(coord1=0;coord1<DIMENSION;++coord1)
					for(coord2=0;coord2<DIMENSION;++coord2)
					{
						// differentiating rowIndex wrt colIndex
						rowIndex = DIMENSION * j + coord2; 
						colIndex = DIMENSION * k + coord1; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
						rowIndex = DIMENSION * k + coord1; 
						colIndex = DIMENSION * j + coord2; 
						matrix[rowIndex + size * colIndex] += derivatives[coord1][coord2];
					}


			} // finished with atom k, given by index kx
		
		}	// Finished with atom j, given by index jx
	} // end of i loop


	// rescale according to energyMultiplier/lengthMultiplier^2
	if( units != NATURAL )
	{
		double l_scale = energyMultiplier/(lengthMultiplier * lengthMultiplier);
		cblas_dscal(size*size,l_scale, matrix, 1);
	}
	
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void StillingerWeberPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);

	writer->PutDouble(SW_A, "SW_A");
	writer->PutDouble(SW_B, "SW_B");
	writer->PutDouble(SW_lambda, "SW_lambda");
	writer->PutDouble(SW_gamma, "SW_gamma");
	writer->PutDouble(SW_a, "SW_a"); // same as cutoffDistance


}

void StillingerWeberPotential::Load(DMReader *reader)
{
	Potential::Load(reader);

	CheckPotentialParameter(reader, SW_A, "SW_A");
	CheckPotentialParameter(reader, SW_B, "SW_B");
	CheckPotentialParameter(reader, SW_lambda, "SW_lambda");
	CheckPotentialParameter(reader, SW_gamma, "SW_gamma");
	CheckPotentialParameter(reader, SW_a, "SW_a");
}
