// IsotropicMultipoleField.cpp: implementation of the IsotropicMultipoleField class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <fstream>
#include "DigitalMaterial.h"

#include "IsotropicMultipoleField.h"
#include "DMArray.h"

#include "DMLastHeader.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IsotropicMultipoleField::IsotropicMultipoleField(int maximumOrder, const double direction[DIMENSION], double center[DIMENSION],double nu):MultipoleField(maximumOrder,direction,center)
{
	this->nu = nu;
	nu_factor = 1./(2*(1.-nu));

	cutoff = 0.;
	
}

IsotropicMultipoleField::~IsotropicMultipoleField()
{


}

void IsotropicMultipoleField::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
	int n,coord,coord2;
	double displacementLocalBasis[2];
	for(int p=0;p<2;++p)
		displacementLocalBasis[p] = 0.;
	
	CalculateGeometry(position);

	double r_minus_n = 1.,cosNTheta=cosTheta,sinNTheta=sinTheta,cosN1Theta,sinN1Theta;

	if(rPerp > cutoff)
		for(n=1;n<=maximumOrder;++n)
		{
			r_minus_n *= rPerpInverse;
			cosN1Theta = cos((n+1.)*theta);
			sinN1Theta = sin((n+1.)*theta);
			
			displacementLocalBasis[0] += r_minus_n *
				(0.5*nu_factor*sinTheta *
				 (coefficients[n-1][0] * sinN1Theta -
				  coefficients[n-1][1] * cosN1Theta)

				 -(coefficients[n-1][0] * cosNTheta +
				   coefficients[n-1][1] * sinNTheta)/(2.*n) 

				 - coefficients[n-1][2] * sinNTheta +
				 coefficients[n-1][3] * cosNTheta);
			
			displacementLocalBasis[1] += r_minus_n *
				(-0.5*nu_factor*sinTheta*(coefficients[n-1][0]*cosN1Theta + coefficients[n-1][1]*sinN1Theta)

				 + (coefficients[n-1][0]*sinNTheta -
					coefficients[n-1][1] * cosNTheta)*0.5*nu_factor*(1-2.*nu)/n

				 + coefficients[n-1][2]*cosNTheta +
				 coefficients[n-1][3]*sinNTheta);
			
			cosNTheta = cosN1Theta;
			sinNTheta = sinN1Theta;
			
		} // end of n loop
	
	for(coord=0;coord<DIMENSION;++coord)
		for(coord2=0;coord2<2;++coord2)// there are only two orientingVectors
			displacement[coord] += orientingVector[coord2][coord]*displacementLocalBasis[coord2];

}



void IsotropicMultipoleField::GradientWRTParameters(const double position[DIMENSION], double (*displacementGradient)[DIMENSION])
{
	int n,coord;

	for(coord=0;coord<DIMENSION;++coord)
	{
		for(n=0;n<4*maximumOrder;++n)
			displacementGradient[n][coord] = 0.;    
	}

	
	CalculateGeometry(position);

	double r_minus_n = 1.,cosNTheta=cosTheta,sinNTheta=sinTheta,cosN1Theta=cos(2*theta),sinN1Theta=sin(2*theta);
	
	if(rPerp > cutoff)
		for(n=1;n<=maximumOrder;++n)
		{
			r_minus_n *= rPerpInverse;
			cosN1Theta = cos((n+1.)*theta);
			sinN1Theta = sin((n+1.)*theta);

			for(coord=0;coord<DIMENSION;++coord)
			{
				displacementGradient[4*(n-1)][coord] += r_minus_n *
					( orientingVector[0][coord] *
					 (0.5*nu_factor*sinTheta*sinN1Theta  - cosNTheta/(2.*n))
					 + orientingVector[1][coord] * 
					 (0.5*nu_factor* (-sinTheta*cosN1Theta +
									  sinNTheta*(1-2.*nu)/n)) );
				
				displacementGradient[4*(n-1) +1][coord] += r_minus_n *
					(orientingVector[0][coord]*(-0.5*nu_factor*sinTheta*cosN1Theta - sinNTheta/(2.*n))
					 + orientingVector[1][coord] *
					 (-0.5*nu_factor*(sinTheta*sinN1Theta +
									  cosNTheta*(1-2.*nu)/n)));

				displacementGradient[4*(n-1) +2][coord] += r_minus_n *
					(orientingVector[0][coord] * -sinNTheta +
					 orientingVector[1][coord]*cosNTheta);
				
				displacementGradient[4*(n-1) +3][coord] += r_minus_n *
					(orientingVector[0][coord]*cosNTheta +
					 orientingVector[1][coord]*sinNTheta);
			} // end of coord loop
			
			cosNTheta = cosN1Theta;
			sinNTheta = sinN1Theta;	
		} // end of n loop

}

void IsotropicMultipoleField::GradientWRTPosition(const double position[DIMENSION],double gradientWRTPosition[DIMENSION][DIMENSION])
{
	int n,coord,coord2;
	double gradientLocalBasis[2][2];
	for(coord=0;coord<2;++coord)
		for(coord2=0;coord2<2;++coord2)
			gradientLocalBasis[coord][coord2] = 0.;
	
	CalculateGeometry(position);
	
	double r_minus_n = rPerpInverse,r_minus_nP1,cosNTheta=cosTheta,sinNTheta=sinTheta,cosN1Theta=cos(2*theta),sinN1Theta=sin(2*theta),cosN2Theta,sinN2Theta; 
    
	if (rPerp > cutoff) 
		for(n=1;n<=maximumOrder;++n)
		{
			r_minus_nP1 = r_minus_n * rPerpInverse;
			cosN2Theta = cos((n+2.)*theta);
			sinN2Theta = sin((n+2.)*theta);
			
			gradientLocalBasis[0][0] += -r_minus_nP1 * 
				( 0.5*(n+1)*nu_factor*sinTheta *
				  (coefficients[n-1][0] * sinN2Theta - coefficients[n-1][1] * cosN2Theta) 

				- 0.5*(coefficients[n-1][0] * cosN1Theta +
					  coefficients[n-1][1] * sinN1Theta) 

				- n * (coefficients[n-1][2] * sinN1Theta - 
					   coefficients[n-1][3] * cosN1Theta));
			
			gradientLocalBasis[0][1] += -r_minus_nP1 *
				(-0.5*nu_factor*
				 (coefficients[n-1][0]*sinN1Theta -  coefficients[n-1][1]*cosN1Theta) -
				 0.5*(n+1)*nu_factor*sinTheta*(coefficients[n-1][0]*cosN2Theta + coefficients[n-1][1]*sinN2Theta) - 
				 
				 0.5*(coefficients[n-1][0]*sinN1Theta -
					  coefficients[n-1][1]*cosN1Theta) + 

				 n*(coefficients[n-1][2]*cosN1Theta +
					coefficients[n-1][3]*sinN1Theta));
			
			gradientLocalBasis[1][0] += -r_minus_nP1 *
				(-0.5*(n+1)*nu_factor*sinTheta *
				 (coefficients[n-1][0]*cosN2Theta +
				  coefficients[n-1][1]*sinN2Theta) +
				 
				 0.5*(1-2.*nu)*nu_factor*(coefficients[n-1][0]*sinN1Theta -
										  coefficients[n-1][1]*cosN1Theta) +

				 n*(coefficients[n-1][2]*cosN1Theta +
					coefficients[n-1][3]*sinN1Theta));
			
			gradientLocalBasis[1][1] += -r_minus_nP1 *
				(0.5*nu_factor*(coefficients[n-1][0]*cosN1Theta +
								coefficients[n-1][1]*sinN1Theta) 
				 
				- 0.5*(n+1)*nu_factor*sinTheta *
				 (coefficients[n-1][0]*sinN2Theta -
				  coefficients[n-1][1]*cosN2Theta)

				 - 0.5*(1-2.*nu)*nu_factor *
				 (coefficients[n-1][0]*cosN1Theta +
				  coefficients[n-1][1]*sinN1Theta)

				 + n*(coefficients[n-1][2]*sinN1Theta -
					  coefficients[n-1][3]*cosN1Theta));
			
			
			cosNTheta = cosN1Theta;
			sinNTheta = sinN1Theta;	
			cosN1Theta = cosN2Theta;
			sinN1Theta = sinN2Theta;
			r_minus_n = r_minus_nP1;
			
	  	} // end of n loop

	for(int a=0;a<DIMENSION;++a)
		for(int b=0;b<DIMENSION;++b)
			for(int c=0;c<2;++c)
				for(int d=0;d<2;++d)
				{
					gradientWRTPosition[a][b] += gradientLocalBasis[c][d]*orientingVector[c][a]*orientingVector[d][b];
				}
	
}
