// NotchFieldDisplacer.h: interface for the NotchFieldDisplacer class.
//
//////////////////////////////////////////////////////////////////////

#ifndef NOTCHFIELDMAKER_H
#define NOTCHFIELDMAKER_H

#include "DMFirstHeader.h"
#include <math.h>
#include "DigitalMaterial.h"
#include <assert.h>
#include <complex>

#include "Quasi2DElasticField.h"
#include "ListOfAtoms.h"

//! This class implements continuum displacements corresponding to a Notch, for given mode I stress intensity factor

/*!
  The continuum solution is from the papers of Suwito, Dunn, Labossiere and
  others (e.g. J. Appl. Phys. 83, No 7, 3574, (1998)) and is based on the
  Stroh formalism for anisotropic elasticity in cases of two dimensional 
  deformations (see T. C. T. Ting, Anisotropic Elasticity: Theory and 
  Applications, OUP, NY, 1996)

  The class is derived from Quasi2DElasticField, and thus one passes a
  direction and center. Orienting vectors (basis vectors within the plane
  normal to direction) are chosen or may be set (via Quasi2DElasticField
  functions). Also passed in is the stress intensity factor and the notch angle
  knowns as alpha (sometimes beta) in the papers. This is not the same angle 
  is the notch angle in NotchMaker (though it is simply related to it). Finally
  one must pass in the Stroh eigenvectors and eigenvalues. Perhaps one could 
  pass in the elastic constants and then calculate the eigenvectors and 
  eigenvalues but this seems unnecessary.

  For now this assumes that DIMENSION is 3. The formalism assumes so and would
  have to be adapted for purely 2D problems.

  need to put in the orienting vectors--right now it only works if the 
  notch is parallel to the z-axis


  \author nbailey
  \date 7/3/2001
  \ingroup moleculardynamics
*/


class NotchFieldDisplacer : public Quasi2DElasticField
{
public:
	NotchFieldDisplacer(const double direction[DIMENSION], double center[DIMENSION], DMArray<double, DIMENSION> *strohVectorsReal, DMArray<double, DIMENSION> *strohVectorsImag, double strohEigenValuesReal[DIMENSION], double strohEigenValuesImag[DIMENSION],double notchAlphaAngle, double lambdaExponent[2], double stressIntensity[2]);
	virtual ~NotchFieldDisplacer();


	

	// Gets and Sets for member variables
	void SetStressIntensity( double stressIntensity[2] );
	double GetStressIntensity(int a) { return stressIntensity[a]; }

	void SetDirection(const double direction[DIMENSION]); 
	const double *GetDirection() const {return (const double *)direction;}
	void SetCenter(const double center[DIMENSION]); 
	const double *GetCenter() const {return (const double *)center;}
	void ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION]);
	void ElasticStress(const double position[DIMENSION], double stress[(DIMENSION+1)*DIMENSION/2]);

#ifndef SWIG

	void GradientWRTPosition(const double position[DIMENSION],  double gradientWRTPosition[DIMENSION][DIMENSION]);


protected:
	virtual void UpdateMemberVariables();
	std::complex<double> **strohVectors;
	std::complex<double> strohEigenvalues[DIMENSION];
	double notchAlphaAngle;
	double lambdaExponent[2];
	double stressIntensity[2];

	std::complex<double> XiMinusAlpha[DIMENSION];
	std::complex<double> bKqFactor[2][DIMENSION];
	static const double TINY;

#endif //SWIG
};

#endif // !defined(NOTCHFIELDMAKER_H)
